      program CardioELSE

!     Copyright (C) CRS4, 2005-2011
!     Loc. Piscina Manna, Edificio 1 - 09010 Pula (CA) - ITALY 
!     Phone: +39 070 92501 Fax: +39 070 9250216 Email: maggio@crs4.it

!     This file is part of CardioELSE.

!     CardioELSE is free software: you can redistribute it and/or modify it
!     under the terms of the GNU Lesser General Public License as published
!     by the Free Software Foundation, either version 2.1 of the License, or
!     (at your option) any later version.

!     CardioELSE is distributed in the hope that it will be useful, but WITHOUT
!     ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
!     FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
!     License for more details. The offer of CardioELSE under the terms of the
!     License is subject to the License being interpreted in accordance with
!     Italian Law and subject to any action against CRS4
!     being under the jurisdiction of the Italian Courts.

!     You should have received a copy of the GNU Lesser General Public License
!     along with CardioELSE. If not, see <http://www.gnu.org/licenses/>.

!     ---

!     author: Fabio Maggio, Bioinformatics Group, CRS4 - May 2011
!     contributions: Bruno Carpentieri, Gianmauro Cuccuru, Giorgio Fotia,
!                    James Southern
!     http://www.crs4.it/bioinformatics

!     CardioELSE has been developed within the Predict Project
!     http://www.vph-predict.eu/

!     CardioELSE is a 3D spectral element solver based on all-hexahedra grids 
!     for the bidomain equation with the LR1 membrane model [1]

!     / c_m v_t - div (M_i grad u_i) + i_ion = I^i_app      in Om x (0,T)
!     |-c_m v_t - div (M_e grad u_e) - i_ion = I^e_app      in Om x (0,T)
!     | n^T M_i,e grad u_i,e nu = psi_i,e                   on Gamma_N x (0,T)
!    < 
!     | u_i,e = psi_i,e                                     on Gamma_D x (0,T)
!     | v(t=t_0)   = v_0                                    in Om
!     \ + LR1 model                                         in Om x (0,T)

!     with M_i,e 3x3 symmetric tensor, nu outward unit normal to Gamma_N

!     LR1 membrane model: the gate variables w, such that 
!     dw/dt = alpha (1-w) - beta*w
!     are integrated  using an Euler implicit scheme:
!     [1/dt + alpha + beta] w^(n+1) = w^n/dt + alpha
!     other variables in the LR1 model (concentration) are handled explicitly

!     time-marching is performed via the semi-implicit scheme shown in [2]:
!     [c_m M/dt+K_i   -c_m M/dt  ] [u_i^(n+1)]   [F_i(v^n,w^(n+1),I^i_app^(n+1)]
!     [                          ] [         ] = [                             ]
!     [ -c_m M/dt    c_m M/dt+K_e] [u_e^(n+1)]   [F_i(v^n,w^(n+1),I^i_app^(n+1)]

!     where K_i,e  = integral_Om((grad N)^T M_i,e grad N d_Om)
!     and M={M_pq} = integral_Om(N_p N_q d_Om)

!     The global matrix, with size (2n)x(2n), is stored in CSR format, suitable
!     for the use of the preconditioner/solver library PETSc

!     [1] Garny A, Noble D, Hunter PJ & Kohl P. Cellular Open Resource (COR): 
!         current status and future directions. Phil Trans R Soc A 367: 
!         1885-1906 (2009).
!     [2] Colli Franzone P, Pavarino L, Savare' G, "Computational 
!         electrocardiology: mathematical and numerical modeling" in 
!         Quarteroni A, Formaggia L, Veneziani A (eds.), "Complex systems in 
!         biomedicine", Springer-Verlag, Milano 2006

!     For a synthetic description of the code and quick guidelines, refer to

!     [3] F. Maggio, "A SEM cardiac simulator based on the bidomain 
!         problem coupled with the Luo-Rudy I membrane model", CRS4 Technical 
!         Report, to appear

!  units of measurement: -------------------------------------------------------
!  length           [cm]
!  times            [ms]
!  Chi              [1/cm]                 (membrane area per volume unit)
!  C_m              [microF/cm^2]          (membrane capacitance per area unit)
!  u_*, v           [mV]                   (electrical potential)
!  d(u_*)/dt,...    [mV/ms=V/s]            (potential time-derivative)
!  sigma_*          [mS/cm=0.001/(Ohm*cm)] (conductivity tensors)
!  i_ion            [microA/cm^2]          (ionic current per area unit)
!  I_ion            [microA/cm^3]          (ionic current per volume unit)
!  psi_*            [microA/cm^2]          (Neumann BC)
!  Chi C_M M dV/dt  [microA]               (term in the variational form)
!  K U_*            [microA]               (term in the variational form)
!  Chi M i_ion      [microA]               (term in the variational form)
!  Int_Gamma psi_*  [microA]               (term in the variational form)
!     --------------------------------------------------------------------------

      implicit none

#include "finclude/petscsys.h"
#include "finclude/petscvec.h"
#include "finclude/petscmat.h"
#include "finclude/petscksp.h"
#include "finclude/petscpc.h"

      integer(4) :: nnod, nnod_macro, ns, nn, nn2, nn3, nelem, nzero, nrig, n_BC
      integer(4) :: DOF_no_Dir_i, DOF_no_Dir_e, ndir_i, nneu_i, ndir_e, nneu_e
      integer(4) :: nts, nrec, narg, iargc, nnz_tot, fa, irec
      integer(4), allocatable :: bindx(:), index_rec(:), faces(:,:), itersnap(:)
      integer(4), allocatable :: point_no_Dir(:), row_ptr(:), col_idx(:)
      integer(4) :: i, i0, i1, i2, i3, i4, j1, k1, imax, ncol_max, nsnapshots
      integer(4), allocatable :: con(:,:), con_spx(:,:), mat(:), bw(:)
      integer(4), allocatable, dimension(:,:) :: con_dir_i, con_dir_e,         &
                                                 con_neu_i, con_neu_e, con_BC
      integer(4) :: ie, iii, j, k, err, algebraic_solver
      real(8) :: zero=0.d0, x1, x2, y1, y2, z1, z2, niter_ave, tsnap, dt_out
      real(8), allocatable :: xxx_macro(:), yyy_macro(:), zzz_macro(:)
      real(8), allocatable :: x_spx(:), y_spx(:), z_spx(:), val(:), val_CSR(:)
      real(8), allocatable :: effe(:), effe_Dir(:), effe_Neu(:)
      real(8), allocatable :: xs(:), ys(:), zs(:), jac(:), uuu(:), vvv(:)
      real(8), allocatable :: omega(:), omega2_neu_i(:,:), omega2_neu_e(:,:)
      real(8) :: ttt, t0, tfin, deltat_ODE, deltat_PDE, term
      real(8) :: inizio, fine, inizio_tl, fine_tl, cg_time1, cg_time2
      real(8) :: inizio_read, fine_read, inizio_prepro, fine_prepro
      real(8) :: amp_Dir_i, amp_Dir_e, amp_Neu_i, amp_Neu_e
      real(8) :: amp_I_app_i, amp_I_app_e,dx_ave,dy_ave,dz_ave
!     LR1 model:
      real(8) :: stim_start, stim_end, stim_duration, stim_period, eps=1.d-12
      real(8) :: Chi, C_m, D_i, D_e
      real(8), external :: ggg_i, ggg_e, hhh
      character*512 :: fichier_histo, fichier_model
      character*512 :: fichier_SE, curr_amp
      character :: ans*1, version*80, riga512*512, riga10*10, code_BC*2
!     character*2, allocatable :: code_BC(:)
      logical :: write_it_hist=.false.,exists    
      logical :: write_spectral_grid_abaqus=.false., singular_system, esiste   
      logical :: Dir_not_empty, Neu_not_empty     
      logical :: write_vtk=.false., write_xdmf=.false.,write_histogram=.false.

      PetscErrorCode   ierr
      PetscMPIInt nprocs, myid, master

      common / domain_box / x1, x2, y1, y2, z1, z2
      common / boundary / Dir_not_empty, Neu_not_empty
      common / Dirichlet_stimulus / amp_Dir_i, amp_Dir_e
      common / Neumann_stimulus / amp_Neu_i, amp_Neu_e
      common / applied_Currents / amp_I_app_i, amp_I_app_e
      common / stim_parameters / stim_start, stim_end, stim_duration,          &
               stim_period
      common / membrane / Chi, C_m, D_i, D_e

      call PetscInitialize(PETSC_NULL_CHARACTER,ierr)
      call MPI_Comm_size(PETSC_COMM_WORLD,nprocs,ierr)
      call MPI_Comm_rank(PETSC_COMM_WORLD,myid,ierr)
      master=0

      version='Version 1.0 (May 2011)'
      call CPU_time(inizio)
      if (myid==master) then
        write(*,*); write(*,698); write(*,699); write(*,700)
        write(*,800); write(*,801); write(*,*)
      endif
698   format(t1,"CardioELSE: a 3D spectral element solver for the bidomain equations")
699   format(t1,"with the Luo-Rudy I membrane model")
700   format(t1,"Bioinformatics Group, CRS4 - May 2011 - http://www.crs4.it/bioinformatics")
800   format(t1,"This cardiac solver has been developed within the Predict Project")
801   format(t1,"http://www.vph-predict.eu/")

      open(23,file='lin_solv.input')
      k=0
      loop_0: do
        read(23,'(a)') riga512
        loop_1: do j=1,512
          if (riga512(j:j).ne.' ') exit loop_1
        enddo loop_1
        if (riga512(j:j).ne.'!') then
          backspace(23)
          exit loop_0
        endif
      enddo loop_0
      read(23,*) algebraic_solver
      close(23)

      call CPU_time(inizio_read)
      open(23,file='CardioELSE.input')
      read(23,'(a)') fichier_model
      read(23,*) ns
      call clean_file(fichier_model)
      open(24,file=fichier_model)
      read(24,'(a)') fichier_SE
      call clean_file(fichier_SE)
      inquire (file=fichier_SE, exist=esiste)
      if (.not.esiste) then
        if (myid==master) then
          write(*,*)
          write(*,*) 'Error! Grid file "'//fichier_SE(1:len_trim(fichier_SE))//&
                     '" does not exist!'
          write(*,*)
        endif
        stop
      endif            
      inquire (file='membrane.input', exist=exists)
      if (.not.exists) then
        if (myid==master)                                                      &
          write(*,*) "Error! file 'membrane.input' does not exist!"
        stop
      endif
      read(24,'(a)') riga512
      i=index(riga512,'!')
      if (i>1) then
        i=i-1
      else
        i=len_trim(riga512)
      endif
      read(riga512(1:i),*,iostat=iii) t0, tfin, deltat_ODE, deltat_PDE, dt_out
      if (iii.ne.0) then
        read(riga512,*) t0, tfin, deltat_ODE, deltat_PDE
        dt_out=0.d0
      endif

      if (deltat_ODE>deltat_PDE)                                               &
          stop "Error! deltat_ODE should be smaller than deltat_PDE!"
      term=dabs(dfloat(idint(deltat_PDE/deltat_ODE))-deltat_PDE/deltat_ODE)    &
          /(deltat_PDE/deltat_ODE)
      if (term>eps) stop "Error! deltat_PDE should be a multiple of deltat_ODE!"
      read(24,*) stim_start, stim_end, stim_duration, stim_period
      read(24,'(a)') riga512
      call read_code_BC(riga512,code_BC)
      read(24,*) amp_Dir_i, amp_Dir_e
      read(24,*) amp_Neu_i, amp_Neu_e
      read(24,*) amp_I_app_i, amp_I_app_e
      read(24,'(a512)') curr_amp
      call curr_amp_location(curr_amp)
      read(24,*) nrec
      if (nrec>0) then
        allocate (xs(nrec),ys(nrec),zs(nrec))
        do i=1,nrec
          read(24,*) xs(i), ys(i), zs(i)
        enddo
      endif

      read(24,*)nsnapshots
      if (nsnapshots>0) then
        allocate (itersnap(0:nsnapshots-1)); itersnap=0
        do i=0,nsnapshots-1
          read(24,*)tsnap
          itersnap(i)=int(tsnap/deltat_PDE)
        enddo
      else if (nsnapshots==0.and.dt_out>0.d0) then
        term=t0
        loop_count_itersnap: do
          term=term+dt_out
          if (term>tfin) exit loop_count_itersnap
          nsnapshots=nsnapshots+1
        enddo loop_count_itersnap
        allocate (itersnap(0:nsnapshots-1)); itersnap=0
        tsnap=t0
        do i=0,nsnapshots-1
          tsnap=tsnap+dt_out
          itersnap(i)=int(tsnap/deltat_PDE)
        enddo
      endif
      close(24)

      read(23,'(a10)') riga10
      if (index(riga10,'vtk')>0.or.index(riga10,'VTK')>0.or.                   &
          index(riga10,'Vtk')>0) then
        write_vtk=.true.
      else if (index(riga10,'xdm')>0.or.index(riga10,'XDM')>0.or.              &
               index(riga10,'Xdm')>0) then
        write_vtk=.true.
!       write_xdmf=.true.
        if (myid==master) then
          write(*,"(t1,'Warning! XDMF format not available for this version.')")
          write(*,"(t1,'Output files will be saved in VTK format.')")
          write(*,*)
        endif
      else
        stop "ERROR! Unknown graphic files format (should be 'vtk' or 'xdmf')"
      endif
      read(23,*) write_it_hist
      read(23,*) write_spectral_grid_abaqus
      read(23,*) write_histogram
      close(23)

      open(23,file='membrane.input')
      read(23,*) Chi
      read(23,*) C_m
      read(23,*) D_i
      read(23,*) D_e
      close(23)

      if (myid==master) write(*,701) fichier_SE(1:len_trim(fichier_SE))
701   format(t1,"reading the gridfile '",a,"' ... ",$)
!     the grid parameters are read:
      call read_grid_parameters(fichier_SE,nnod_macro,nelem)
!     nnod_macro=# "macro" gridpoints; nelem=# hexahedra
!     ndir_i=# quadrilaterals on Dirichlet boundary for DOF u_i
!     ndir_e=# quadrilaterals on Dirichlet boundary for DOF u_e
!     nneu_i=# quadrilaterals on Neumann boundary for DOF u_i
!     nneu_e=# quadrilaterals on Neumann boundary for DOF u_ei

      Dir_not_empty=.false.; Neu_not_empty=.false.
      if ((code_BC(1:1)=='D'.or.code_BC(1:1)=='d').and.                        &
          (amp_Dir_i.ne.zero)) Dir_not_empty=.true.
      if ((code_BC(2:2)=='D'.or.code_BC(2:2)=='d').and.                        &
          (amp_Dir_e.ne.zero)) Dir_not_empty=.true.
      if ((code_BC(1:1)=='N'.or.code_BC(1:1)=='n').and.                        &
          (amp_Neu_i.ne.zero)) Neu_not_empty=.true.
      if ((code_BC(2:2)=='N'.or.code_BC(2:2)=='n').and.                        &
          (amp_Neu_e.ne.zero)) Neu_not_empty=.true.
      singular_system=.true.
      if (Dir_not_empty) singular_system=.false.

!     gridpoints and connectivities are read
      allocate(xxx_macro(nnod_macro),yyy_macro(nnod_macro),                    &
               zzz_macro(nnod_macro))
      allocate (con(8,nelem),mat(nelem))
      call read_grid(fichier_SE,nnod_macro,nelem,xxx_macro,yyy_macro,zzz_macro,&
                     con,mat)

!     boundary faces in the source regions are counted and identified
      call count_boundary_faces(nnod_macro,nelem,con,xxx_macro,yyy_macro,      &
                                zzz_macro,code_BC,n_BC)
      allocate (con_BC(2,n_BC))
      call list_boundary_faces(nnod_macro,nelem,n_BC,con,xxx_macro,yyy_macro,  &
                               zzz_macro,code_BC,con_BC)
      ndir_i=0; ndir_e=0; nneu_i=0; nneu_e=0
      if (code_BC(1:1)=='D'.or.code_BC(1:1)=='d') ndir_i=n_BC
      if (code_BC(2:2)=='D'.or.code_BC(2:2)=='d') ndir_e=n_BC
      if (code_BC(1:1)=='N'.or.code_BC(1:1)=='n') nneu_i=n_BC
      if (code_BC(2:2)=='N'.or.code_BC(2:2)=='n') nneu_e=n_BC
      allocate (con_dir_i(2,ndir_i),con_dir_e(2,ndir_e))
      allocate (con_neu_i(2,nneu_i),con_neu_e(2,nneu_e))
      if (code_BC(1:1)=='D'.or.code_BC(1:1)=='d') con_dir_i=con_BC
      if (code_BC(2:2)=='D'.or.code_BC(2:2)=='d') con_dir_e=con_BC
      if (code_BC(1:1)=='N'.or.code_BC(1:1)=='n') con_neu_i=con_BC
      if (code_BC(2:2)=='N'.or.code_BC(2:2)=='n') con_neu_e=con_BC

      call CPU_time(fine_read)          
      if (myid==master) then
        write(*,702); write(*,*)
        call eval_grid_size(nnod_macro,nelem,xxx_macro,yyy_macro,zzz_macro,con,&
                            dx_ave,dy_ave,dz_ave)
        write(riga512,'("average dx,dy,dz = ",g20.3,",",g20.3,",",g20.3)')    &
              dx_ave,dy_ave,dz_ave
        call clean_multiple_blank_string(512,riga512)
        write(*,'(t1,a)') riga512(1:len_trim(riga512))
      endif
702   format("done.")
      call MPI_Bcast(ns, 1, MPI_Integer, master, PETSC_COMM_WORLD, ierr)

      call CPU_time(inizio_prepro)
!     spectral connectivities generation
!     (see description of subroutine "make_spectral_connectivities")
!     ---------------------------------------------------------------
      nnod=nnod_macro           ! nnod is the # spectral grid-points
      nn=ns+1                   ! ns=spectral degree; nn=# grid-points per edge
      nn2=nn*nn                 ! nn2=# grid-points per face
      nn3=nn2*nn                ! nn3=# grid-points per hexahedron
      allocate (con_spx(nn3,nelem))
      con_spx=0
      call make_spectral_connectivities(nnod_macro,nn,nelem,con,               &
                                        con_spx,nnod,ncol_max,err)

!     spectral grid and Jacobian generation
!     (see description of subroutine "make_spectral_grid")
!     ---------------------------------------------------------------     
      allocate (x_spx(nnod), y_spx(nnod), z_spx(nnod), jac(nelem))
      x_spx=zero; y_spx=zero; z_spx=zero; jac=zero
      call make_spectral_grid(nn,nnod,nelem,nnod_macro,xxx_macro,yyy_macro,    &
                              zzz_macro,con,con_spx,x_spx,y_spx,z_spx)

!     evaluates min and max cartesian components of each gridpoint
      call evaluate_box(nnod,x_spx,y_spx,z_spx,x1,x2,y1,y2,z1,z2)
      if (myid==master) then
        write(riga512,'("X range: ",g20.3," to ",g20.3," ( range extension: ", &
                        g20.3,")")') x1, x2, x2-x1
        call clean_multiple_blank_string(512,riga512)
        write(*,'(t1,a)') riga512(1:len_trim(riga512))
        write(riga512,'("Y range: ",g20.3," to ",g20.3," ( range extension: ", &
                       g20.3,")")') y1, y2, y2-y1
        call clean_multiple_blank_string(512,riga512)
        write(*,'(t1,a)') riga512(1:len_trim(riga512))
        write(riga512,'("Z range: ",g20.3," to ",g20.3," ( range extension: ", &
                        g20.3,")")') z1, z2, z2-z1
        call clean_multiple_blank_string(512,riga512)
        write(*,'(t1,a)') riga512(1:len_trim(riga512))
        write(*,*)
      endif

!     if (write_spectral_grid_abaqus) then
!       open(78,file="spectral_grid.inp")
!       write(78,'(t1,"*HEADING")')
!       write(78,'(t1,"CardioELSE")')
!       write(78,'(t1,"*NODE")')
!       do i=1,nnod
!         write(78,781) i,x_spx(i),y_spx(i),z_spx(i)
!       enddo
781     format(t1,i6,3(",",2x,g12.6))
!       write(78,'(t1,"*ELEMENT, TYPE=C3D8R, ELSET=EB1")')
!       do ie=1,nelem
!         write(78,782) ie,con_spx(1,ie),con_spx(nn,ie),                       &
!                          con_spx(nn2,ie),con_spx(nn2-nn+1,ie),               &
!                          con_spx(nn3-nn2+1,ie),con_spx(nn3-nn2+nn,ie),       &
!                          con_spx(nn3,ie),con_spx(nn3-nn+1,ie)
!       enddo
782     format(t1,i6,8(",",2x,i6))
!       close(78)
!     endif

!     evaluates position of receivers, moved to the nearest gridpoint
      allocate (index_rec(nrec)); index_rec=0
      call evaluate_index_rec(nnod,nrec,x_spx,y_spx,z_spx,xs,ys,zs,index_rec)

!     generate the pointer point_no_Dir to non-Dirichlet DOFs:
!     point_no_Dir(j)=0 ---> j is a Dirichlet DOF 
      allocate (point_no_Dir(2*nnod)); point_no_Dir=0
      call evaluate_DOF_Dir(nn,nelem,nnod,ndir_i,ndir_e,con_dir_i,con_dir_e,   &
                            con_spx,DOF_no_Dir_i,DOF_no_Dir_e,point_no_Dir)
      nrig=DOF_no_Dir_i+DOF_no_Dir_e

!     evaluate Jacobian and normals for surface integrals (needed for BCs)
      allocate(effe_Neu(nrig)); effe_Neu=zero
      if (nneu_i+nneu_e>0) then 
        call make_effe_Neu(nn,nnod,nrig,nelem,con_spx,nneu_i,nneu_e,con_neu_i, &
                           con_neu_e,x_spx,y_spx,z_spx,point_no_Dir,effe_Neu)
      endif

!     matrix generation
      allocate (bw(2*nnod)); bw=0
      call make_bw(nn,nnod,nelem,con_spx,point_no_Dir,bw,nnz_tot)
!     write parameters on the screen
      if (myid==master) then
        call write_grid_parameters(ns,fichier_SE,nnod_macro,nnod,DOF_no_Dir_i, &
                                   DOF_no_Dir_e,nelem,ndir_i,ndir_e,nneu_i,    &
                                   nneu_e,nnz_tot,t0,tfin,deltat_ODE,          &
                                   deltat_PDE,write_it_hist)
        write(*,*)
      endif

      allocate (row_ptr(0:nrig),col_idx(0:nnz_tot-1)); row_ptr=-1; col_idx=-1
      call make_CSR_indices(nn,nnod,nelem,nrig,nnz_tot,con_spx,point_no_Dir,   &
                            bw,row_ptr,col_idx)
      allocate(val_CSR(0:nnz_tot-1), omega(nnod), effe_Dir(nrig))
      val_CSR=zero; omega=zero; effe_Dir=zero
      call make_global_matrix_CSR(nn,nnod,nelem,nrig,x_spx,y_spx,z_spx,        &
                                  con_spx,deltat_PDE,nnz_tot,row_ptr,col_idx,  &
                                  point_no_Dir,val_CSR,omega,effe_Dir)
      call check_zero_in_stiffness_mat_CSR(nnz_tot,val_CSR,nzero)
      if (nzero>0) then
        if (myid==master) write(*,*) "the CSR stiffness matrix has ",nzero,    &
                                     " null elements"
      endif

     if (write_histogram) then    ! makes non-zero-entries-per-row histogram
       if (myid==master) then
         fichier_histo="non_zero_histogram.d"
         call make_histogram_CSR(nrig,nnz_tot,row_ptr,fichier_histo)
        endif
      endif
!     ..........................................................................

!     uuu(1:DOF_no_Dir_i) = u_i  ;  uuu(DOF_no_Dir_i+1:nrig) = u_e
!     where u_i and u_e, unknowns of the PDE, are respectively:
!     the internal potential u_int and the external potential u_ext
      allocate (uuu(nrig)); uuu=zero

      call CPU_time(fine_prepro)

      call CPU_time(inizio_tl)

      if (algebraic_solver==0) then
        if (myid==master) write(*,706)
706     format(t1,"Time loop has started (Conjugate Gradient + diagonal preconditioner)...")

        call time_loop_CSR(myid,master,nn,nnod,nelem,nrig,nnz_tot,nrec,x_spx,  &
                           y_spx,z_spx,con_spx,index_rec,t0,tfin,ttt,          &
                           deltat_ODE,deltat_PDE,omega,effe_Dir,point_no_Dir,  &
                           effe_Neu,row_ptr,col_idx,val_CSR,uuu,Dir_not_empty, &
                           Neu_not_empty,write_it_hist,write_vtk,write_xdmf,   &
                           itersnap,nsnapshots,niter_ave,cg_time1,cg_time2)  

      else if (algebraic_solver==1) then
        if (myid==master) write(*,707)
707     format(t1,"Time loop has started (PETSc)...")

        call time_loop_PETSc(myid, nprocs, master,                             &
                             nn,nnod,nelem,nrig,nnz_tot,nrec,x_spx,y_spx,z_spx,&
                             con_spx,index_rec,t0,tfin,ttt,deltat_ODE,         &
                             deltat_PDE,omega,effe_Dir,point_no_Dir,effe_Neu,  &
                             row_ptr,col_idx,val_CSR,uuu,Dir_not_empty,        &
                             Neu_not_empty,write_it_hist,write_vtk,write_xdmf, &
                             itersnap,nsnapshots,niter_ave,cg_time1,cg_time2)

      endif

      call CPU_time(fine_tl)
      if (myid==master) write(*,*)

!     if needed, numerical solution uuu is extended to uuu(nnn)
      if (Dir_not_empty) then
        allocate (effe(DOF_no_Dir_i+DOF_no_Dir_e)); effe=zero
        effe=uuu
        deallocate(uuu)
        allocate(uuu(2*nnod)); uuu=zero
        do i=1,nnod
          iii=point_no_Dir(i)
          if (iii>0) then
            uuu(i)=effe(iii)
          else
            uuu(i)=ggg_i(x_spx(i),y_spx(i),z_spx(i))*hhh(ttt)
          endif
        enddo
        do i=1,nnod
          iii=point_no_Dir(i+nnod)
          if (iii>0) then
            uuu(i+nnod)=effe(iii)
          else
            uuu(i+nnod)=ggg_e(x_spx(i),y_spx(i),z_spx(i))*hhh(ttt)
          endif
        enddo
        deallocate(effe)       
      endif

      if (myid==master) then
        inquire (file='execution_times.d',EXIST=exists)
        if (exists) then
          open(33,file='execution_times.d',position='append')
        else
          open(33,file='execution_times.d')
          write(33,330)
          write(33,*)
        endif
        write(33,331) nn-1, fine_read-inizio_read, fine_prepro-inizio_prepro,  &
                      fine_tl-inizio_tl, niter_ave
        close(33)
      endif
330   format(t1,'#     p     ,   read (s)   ,preprocess (s), time-loop (s), <# &
iter.>')
331   format(t1,i10,'  ,  ',3(g10.4,'  ,  '),g8.2)

      call CPU_time(fine)

      call PetscFinalize(ierr)

      end program CardioELSE


!     --------------------------------------------------------------------------


      subroutine write_grid_parameters(ns,fichier,nnod,nnod_spx,DOF_no_Dir_i,  &
                                       DOF_no_Dir_e,nelem,ndir_i,ndir_e,nneu_i,&
                                       nneu_e,nnz,t0,tfin,dt_ODE,dt_PDE,       &
                                       write_it_hist)

      implicit none
      integer(4) :: i, j
      integer(4) :: ns, nnod, nnod_spx, DOF_no_Dir_i, DOF_no_Dir_e, nelem
      integer(4) :: ndir_i, ndir_e, nneu_i, nneu_e, nnz
      real(8) :: t0,tfin,dt_ODE,dt_PDE
      character(512) :: fichier,stringa
      logical :: write_it_hist

      write(stringa,'(t1,"spectral degree: ",t40,i10)') ns
      write(*,'(a)') stringa(1:len_trim(stringa))
      write(stringa,'(t1,"# macro-gridpoints: ",t40,i10)') nnod
      write(*,'(a)') stringa(1:len_trim(stringa))
      write(stringa,'(t1,"# spectral gridpoints: ",t40,i10)') nnod_spx
      write(*,'(a)') stringa(1:len_trim(stringa))
      write(stringa,'(t1,"# non-Dirichlet spectral DOF : ",t40,i10)')   &
            DOF_no_Dir_i+DOF_no_Dir_e
      write(*,'(a)') stringa(1:len_trim(stringa))
      write(stringa,'(t1,"# nnz: ",t30,i20)') nnz
      write(*,'(a)') stringa(1:len_trim(stringa))
      write(stringa,'(t1,"# elements: ",t40,i10)') nelem
      write(*,'(a)') stringa(1:len_trim(stringa))
      write(stringa,'(t1,"# Dirichlet BC elements: ",t40,i10)') ndir_i+ndir_e
      write(*,'(a)') stringa(1:len_trim(stringa))               
      write(stringa,'(t1,"# Neumann BC elements: ",t40,i10)') nneu_i+nneu_e
      write(*,'(a)') stringa(1:len_trim(stringa))    
      write(stringa,'(t1,"t_0, t_fin, dt_ODE, dt_PDE (all [ms])   =   ",       &
                      4(g8.2,1x))') t0, tfin, dt_ODE, dt_PDE
      write(*,'(a)') stringa(1:len_trim(stringa))
      
      if (write_it_hist) then
        open(23,file='iterations_info.output')
        write(stringa,'(t1,"spectral degree: ",t40,i10)') ns
        write(23,'(a)') stringa(1:len_trim(stringa))
        write(stringa,'(t1,"# macro-gridpoints: ",t40,i10)') nnod
        write(23,'(a)') stringa(1:len_trim(stringa))
        write(stringa,'(t1,"# spectral gridpoints: ",t40,i10)') nnod_spx
        write(23,'(a)') stringa(1:len_trim(stringa))
        write(stringa,'(t1,"# non-Dirichlet spectral DOF : ",t40,i10)')   &
              DOF_no_Dir_i+DOF_no_Dir_e
        write(23,'(a)') stringa(1:len_trim(stringa))
        write(stringa,'(t1,"# nnz: ",t30,i20)') nnz
        write(23,'(a)') stringa(1:len_trim(stringa))
        write(stringa,'(t1,"# elements: ",t40,i10)') nelem
        write(23,'(a)') stringa(1:len_trim(stringa))
        write(stringa,'(t1,"# Dirichlet BC elements: ",t40,i10)') ndir_i+ndir_e
        write(23,'(a)') stringa(1:len_trim(stringa))
        write(stringa,'(t1,"# Neumann BC elements: ",t40,i10)') nneu_i+nneu_e
        write(23,'(a)') stringa(1:len_trim(stringa))
        write(stringa,'(t1,"t_0, t_fin, dt_ODE, dt_PDE (all [ms])   =   ",     &
                        4(g8.2,1x))') t0, tfin, dt_ODE, dt_PDE
        write(23,'(a)') stringa(1:len_trim(stringa))
        close(23)
      endif

      return

      end subroutine write_grid_parameters


!     --------------------------------------------------------------------------


      subroutine evaluate_box(nnod,xxx,yyy,zzz,x1,x2,y1,y2,z1,z2)

!     min and max cartesian components of each gridpoint are evaluated,

      implicit none
      
      integer(4) :: i, nnod
      real(8) :: x1, x2, y1, y2, z1, z2, xxx(nnod), yyy(nnod), zzz(nnod)

      x1=1.d90; x2=-1.d90; y1=1.d90; y2=-1.d90; z1=1.d90; z2=-1.d90
      do i=1,nnod
        if (xxx(i)<x1) x1=xxx(i); if (xxx(i)>x2) x2=xxx(i)
        if (yyy(i)<y1) y1=yyy(i); if (yyy(i)>y2) y2=yyy(i)
        if (zzz(i)<z1) z1=zzz(i); if (zzz(i)>z2) z2=zzz(i)
      enddo

      return

      end subroutine evaluate_box


!     --------------------------------------------------------------------------


      subroutine make_global_matrix_CSR(nn,nnod,nelem,nrig,xxx,yyy,zzz,con,    &
                                        deltat,nnz_tot,row_ptr,col_idx,        &
                                        point_no_Dir,val,omega_tot,effe_Dir)

!     the stiffness matrix is evaluated
!     integrals in the reference hexahedron are evaluated numerically

      implicit none


      integer(4) :: ie,h,i,j,l,m,n,p,q,r,s,t,iii,i1,j1,k1,nn,nnz_tot,nnod,nelem
      integer(4) :: a, b, c, h_tl, h_tr, h_bl, h_br, nn2, ppp, qqq, nrig
      integer(4) :: con(nn**3,nelem), row_ptr(0:nrig), col_idx(0:nnz_tot-1)
      integer(4) :: point_no_Dir(2*nnod)
      real(8) :: val(0:nnz_tot-1), j_det(nn,nn,nn), integrale
      real(8) :: zero=0.d0, term, dede(0:3,3), deltat, effe_Dir(nrig)
      real(8) :: xxx(nnod), yyy(nnod), zzz(nnod), ct(nn), ww(nn), dd(nn,nn)
      real(8) :: serv_i, serv_e, cost
      real(8) :: aaa(nn,nn,nn,3,3), alfa_i(nn,nn,nn,3,3), alfa_e(nn,nn,nn,3,3)
      real(8), dimension(nelem) :: alfa1,beta1,gamma1,alfa2,beta2,gamma2,alfa3
      real(8), dimension(nelem) :: beta3,gamma3,lambda1,mu1,nu1,sigma1,lambda2
      real(8), dimension(nelem) :: mu2,nu2,sigma2,lambda3,mu3,nu3,sigma3
      real(8) :: omega_tot(nnod), omega0(nn,nn,nn), omega(nn,nn,nn) 
      real(8), dimension(nn,nn) :: dexdex,dexdey,dexdez,deydex,deydey,deydez,  &
                                   dezdex,dezdey,dezdez
      real(8), external :: mcap_i, mcap_e, c_m_per_vol, ggg_i, ggg_e
      logical :: trovato=.false., in_stencil=.false.

      nn2=nn*nn
      ct=zero; ww=zero; dd=zero
      call lgl(nn,ct,ww,dd)
      do c=1,nn
        do b=1,nn
          do a=1,nn
            omega0(a,b,c)=ww(a)*ww(b)*ww(c)
          enddo
        enddo
      enddo
      val=zero; omega_tot=zero; effe_dir=zero

      call make_mapping_parameters(nn,nnod,nelem,xxx,yyy,zzz,con,              &
           alfa1,beta1,gamma1,alfa2,beta2,gamma2,alfa3,beta3,gamma3,           &
           lambda1,mu1,nu1,sigma1,lambda2,mu2,nu2,sigma2,lambda3,mu3,nu3,sigma3)

      do ie=1,nelem

        do k1=1,nn
          do i1=1,nn
            dexdey(i1,k1)=beta1(ie)+lambda1(ie)*ct(i1)+nu1(ie)*ct(k1)+         &
                          sigma1(ie)*ct(i1)*ct(k1)
            deydey(i1,k1)=beta2(ie)+lambda2(ie)*ct(i1)+nu2(ie)*ct(k1)+         &
                          sigma2(ie)*ct(i1)*ct(k1)
            dezdey(i1,k1)=beta3(ie)+lambda3(ie)*ct(i1)+nu3(ie)*ct(k1)+         &
                          sigma3(ie)*ct(i1)*ct(k1)
          enddo
          do j1=1,nn
            dexdex(j1,k1)=alfa1(ie)+lambda1(ie)*ct(j1)+mu1(ie)*ct(k1)+         &
                          sigma1(ie)*ct(j1)*ct(k1)
            deydex(j1,k1)=alfa2(ie)+lambda2(ie)*ct(j1)+mu2(ie)*ct(k1)+         &
                          sigma2(ie)*ct(j1)*ct(k1)
            dezdex(j1,k1)=alfa3(ie)+lambda3(ie)*ct(j1)+mu3(ie)*ct(k1)+         &
                          sigma3(ie)*ct(j1)*ct(k1)
          enddo
        enddo
        do j1=1,nn
          do i1=1,nn
            dexdez(i1,j1)=gamma1(ie)+mu1(ie)*ct(i1)+nu1(ie)*ct(j1)+            &
                          sigma1(ie)*ct(i1)*ct(j1)
            deydez(i1,j1)=gamma2(ie)+mu2(ie)*ct(i1)+nu2(ie)*ct(j1)+            &
                          sigma2(ie)*ct(i1)*ct(j1)
            dezdez(i1,j1)=gamma3(ie)+mu3(ie)*ct(i1)+nu3(ie)*ct(j1)+            &
                          sigma3(ie)*ct(i1)*ct(j1)
          enddo
        enddo

        do k1=1,nn
          do j1=1,nn
            do i1=1,nn

              j_det(i1,j1,k1)=dexdex(j1,k1)*(deydey(i1,k1)*dezdez(i1,j1)-      &
                              deydez(i1,j1)*dezdey(i1,k1)) +                   &
                              deydex(j1,k1)*(dezdey(i1,k1)*dexdez(i1,j1)-      &
                              dexdey(i1,k1)*dezdez(i1,j1)) +                   &
                              dezdex(j1,k1)*(dexdey(i1,k1)*deydez(i1,j1)-      &
                              deydey(i1,k1)*dexdez(i1,j1))

              omega(i1,j1,k1)=omega0(i1,j1,k1)*dabs(j_det(i1,j1,k1))
              i=con((k1-1)*nn2+(j1-1)*nn+i1,ie)
              omega_tot(i)=omega_tot(i)+omega(i1,j1,k1) 
            enddo
          enddo
        enddo

        do k1=1,nn
          do j1=1,nn
            do i1=1,nn
              aaa(i1,j1,k1,1,1)=-(dezdey(i1,k1)*deydez(i1,j1)-                 &
                                   deydey(i1,k1)*dezdez(i1,j1))/j_det(i1,j1,k1)
              aaa(i1,j1,k1,1,2)=-(deydex(j1,k1)*dezdez(i1,j1)-                 &
                                   dezdex(j1,k1)*deydez(i1,j1))/j_det(i1,j1,k1)
              aaa(i1,j1,k1,1,3)=-(dezdex(j1,k1)*deydey(i1,k1)-                 &
                                   deydex(j1,k1)*dezdey(i1,k1))/j_det(i1,j1,k1)

              aaa(i1,j1,k1,2,1)=+(dezdey(i1,k1)*dexdez(i1,j1)-                 &
                                   dexdey(i1,k1)*dezdez(i1,j1))/j_det(i1,j1,k1)
              aaa(i1,j1,k1,2,2)=+(dexdex(j1,k1)*dezdez(i1,j1)-                 &
                                   dezdex(j1,k1)*dexdez(i1,j1))/j_det(i1,j1,k1)
              aaa(i1,j1,k1,2,3)=+(dezdex(j1,k1)*dexdey(i1,k1)-                 &
                                   dexdex(j1,k1)*dezdey(i1,k1))/j_det(i1,j1,k1)
                                                                              
              aaa(i1,j1,k1,3,1)=-(deydey(i1,k1)*dexdez(i1,j1)-                 &
                                  dexdey(i1,k1)*deydez(i1,j1))/j_det(i1,j1,k1)
              aaa(i1,j1,k1,3,2)=-(dexdex(j1,k1)*deydez(i1,j1)-                 &
                                  deydex(j1,k1)*dexdez(i1,j1))/j_det(i1,j1,k1)
              aaa(i1,j1,k1,3,3)=-(deydex(j1,k1)*dexdey(i1,k1)-                 &
                                  dexdex(j1,k1)*deydey(i1,k1))/j_det(i1,j1,k1)
            enddo
          enddo
        enddo    
    
        alfa_i=zero; alfa_e=zero
        do k1=1,nn
          do j1=1,nn
            do i1=1,nn
              iii=con((k1-1)*nn2+(j1-1)*nn+i1,ie)
              do l=1,3
                do m=1,3
                  do r=1,3
                    do s=1,3
                      alfa_i(i1,j1,k1,l,m)=alfa_i(i1,j1,k1,l,m)+               &
                           mcap_i(r,s,xxx(iii),yyy(iii),zzz(iii))*             &
                           aaa(i1,j1,k1,r,l)*aaa(i1,j1,k1,s,m)
                      alfa_e(i1,j1,k1,l,m)=alfa_e(i1,j1,k1,l,m)+               &
                           mcap_e(r,s,xxx(iii),yyy(iii),zzz(iii))*             &
                           aaa(i1,j1,k1,r,l)*aaa(i1,j1,k1,s,m)
                    enddo
                  enddo
                enddo
              enddo  
            enddo
          enddo
        enddo                    
                                                                       
!       1st block of DOF_no_Dir_i rows
        do n=1,nn
          do m=1,nn
            do l=1,nn
              iii=(n-1)*nn2+(m-1)*nn+l
              p=con(iii,ie)
              ppp=point_no_Dir(p)
              if (ppp>0) then

              do t=1,nn
                do s=1,nn
                  do r=1,nn

                    in_stencil=.false.
                    serv_i=zero

                    if (m==s.and.n==t) then
                      in_stencil=.true.
                      do a=1,nn
                        term=dd(a,l)*dd(a,r)*omega(a,m,n)
                        serv_i=serv_i+alfa_i(a,m,n,1,1)*term
                      enddo
                    endif
                    if (n==t) then
                      in_stencil=.true.
                      term=dd(m,s)*dd(r,l)*omega(r,m,n)
                      serv_i=serv_i+alfa_i(r,m,n,1,2)*term
                    endif
                    if (m==s) then
                      in_stencil=.true.
                      term=dd(n,t)*dd(r,l)*omega(r,m,n)
                      serv_i=serv_i+alfa_i(r,m,n,1,3)*term
                    endif

                    if (l==r.and.n==t) then
                      in_stencil=.true.
                      do b=1,nn
                        term=dd(b,m)*dd(b,s)*omega(l,b,n)
                        serv_i=serv_i+alfa_i(l,b,n,2,2)*term
                      enddo
                    endif
                    if (n==t) then
                      in_stencil=.true.
                      term=dd(s,m)*dd(l,r)*omega(l,s,n)
                      serv_i=serv_i+alfa_i(l,s,n,2,1)*term
                    endif
                    if (l==r) then
                      in_stencil=.true.
                      term=dd(s,m)*dd(n,t)*omega(l,s,n)
                      serv_i=serv_i+alfa_i(l,s,n,2,3)*term
                    endif
                    
                    if (l==r.and.m==s) then
                      in_stencil=.true.
                      do c=1,nn
                        term=dd(c,n)*dd(c,t)*omega(l,m,c)
                        serv_i=serv_i+alfa_i(l,m,c,3,3)*term
                      enddo
                    endif
                    if (m==s) then
                      in_stencil=.true.
                      term=dd(t,n)*dd(l,r)*omega(l,m,t)
                      serv_i=serv_i+alfa_i(l,m,t,3,1)*term
                    endif
                    if (l==r) then
                      in_stencil=.true.
                      term=dd(t,n)*dd(m,s)*omega(l,m,t)
                      serv_i=serv_i+alfa_i(l,m,t,3,2)*term
                    endif                                              

                    if (in_stencil) then

                      iii=(t-1)*nn2+(s-1)*nn+r
                      q=con(iii,ie)
                      qqq=point_no_Dir(q)

                      if (qqq>0) then      
                                
                        trovato=.false.
                        loop_cerca_pos_top: do h=row_ptr(ppp-1),row_ptr(ppp)-1
                          if (col_idx(h)==qqq-1) then
                            h_tl=h
                            trovato=.true.
                            exit loop_cerca_pos_top
                          endif
                        enddo loop_cerca_pos_top 
                        if (.not.trovato) then
                          write(*,*)
                          write(*,*)  "ERROR!  h_tl NOT FOUND!"
                          write(*,*)
                          stop
                        endif
                        val(h_tl)=val(h_tl)+serv_i   
                                      
                      else        ! if q IS a Dirichlet node

                        effe_Dir(ppp)=effe_Dir(ppp)                            &
                                     -serv_i*ggg_i(xxx(q),yyy(q),zzz(q))

                      endif        ! if q is or is not a Dirichlet node

                    endif        ! if (in_stencil)  

                  enddo            ! this is
                enddo            ! the loop
              enddo                ! over q

              endif                ! if p is not a Dirichlet node

            enddo                ! this is
          enddo                    ! the loop
        enddo                    ! over p

!       2nd block of DOF_no_Dir_e rows
        do n=1,nn
          do m=1,nn
            do l=1,nn
              iii=(n-1)*nn2+(m-1)*nn+l
              p=con(iii,ie)
              ppp=point_no_Dir(p+nnod)
              if (ppp>0) then

              do t=1,nn
                do s=1,nn
                  do r=1,nn

                    in_stencil=.false.
                    serv_e=zero

                    if (m==s.and.n==t) then
                      in_stencil=.true.
                      do a=1,nn
                        term=dd(a,l)*dd(a,r)*omega(a,m,n)
                        serv_e=serv_e+alfa_e(a,m,n,1,1)*term
                      enddo
                    endif
                    if (n==t) then
                      in_stencil=.true.
                      term=dd(m,s)*dd(r,l)*omega(r,m,n)
                      serv_e=serv_e+alfa_e(r,m,n,1,2)*term
                    endif
                    if (m==s) then
                      in_stencil=.true.
                      term=dd(n,t)*dd(r,l)*omega(r,m,n)
                      serv_e=serv_e+alfa_e(r,m,n,1,3)*term
                    endif

                    if (l==r.and.n==t) then
                      in_stencil=.true.
                      do b=1,nn
                        term=dd(b,m)*dd(b,s)*omega(l,b,n)
                        serv_e=serv_e+alfa_e(l,b,n,2,2)*term
                      enddo
                    endif
                    if (n==t) then
                      in_stencil=.true.
                      term=dd(s,m)*dd(l,r)*omega(l,s,n)
                      serv_e=serv_e+alfa_e(l,s,n,2,1)*term
                    endif
                    if (l==r) then
                      in_stencil=.true.
                      term=dd(s,m)*dd(n,t)*omega(l,s,n)
                      serv_e=serv_e+alfa_e(l,s,n,2,3)*term
                    endif
                    
                    if (l==r.and.m==s) then
                      in_stencil=.true.
                      do c=1,nn
                        term=dd(c,n)*dd(c,t)*omega(l,m,c)
                        serv_e=serv_e+alfa_e(l,m,c,3,3)*term
                      enddo
                    endif
                    if (m==s) then
                      in_stencil=.true.
                      term=dd(t,n)*dd(l,r)*omega(l,m,t)
                      serv_e=serv_e+alfa_e(l,m,t,3,1)*term
                    endif
                    if (l==r) then
                      in_stencil=.true.
                      term=dd(t,n)*dd(m,s)*omega(l,m,t)
                      serv_e=serv_e+alfa_e(l,m,t,3,2)*term
                    endif                                              

                    if (in_stencil) then

                      iii=(t-1)*nn2+(s-1)*nn+r
                      q=con(iii,ie)
                      qqq=point_no_Dir(q+nnod)

                      if (qqq>0) then
                   
                        trovato=.false.
                        loop_cerca_pos_bot: do h=row_ptr(ppp-1),row_ptr(ppp)-1                               
                          if (col_idx(h)==qqq-1) then
                            h_br=h
                            trovato=.true.
                            exit loop_cerca_pos_bot
                          endif
                        enddo loop_cerca_pos_bot 
                        if (.not.trovato) then
                          stop "ERROR!  h_br NOT FOUND!"
                        endif
                        val(h_br)=val(h_br)+serv_e
                    
                      else        ! if q IS a Dirichlet node

                        effe_Dir(ppp)=effe_Dir(ppp)                            &
                                     -serv_e*ggg_e(xxx(q),yyy(q),zzz(q))

                      endif        ! if q is or is not a Dirichlet node

                    endif        ! if (in_stencil)  
 
                  enddo            ! this is
                enddo            ! the loop
              enddo                ! over q

              endif                ! if p is not a Dirichlet node

            enddo                ! this is
          enddo                    ! the loop
        enddo                    ! over p
  
      enddo                        ! this is the loop over hexahedra

!     1st block of DOF_no_Dir_i rows: diagonals
      do p=1,nnod
        ppp=point_no_Dir(p)
        if (ppp>0) then
          term=c_m_per_vol(xxx(p),yyy(p),zzz(p))*omega_tot(p)/deltat
          trovato=.false.
          loop_cerca_diag_1: do qqq=row_ptr(ppp-1),row_ptr(ppp)-1
            if (col_idx(qqq)==ppp-1) then
              trovato=.true.
              exit loop_cerca_diag_1
            endif
          enddo loop_cerca_diag_1
          if (.not.trovato) then
            write(*,*) 'Error! Diagonal element not found!'
            stop
          endif
          val(qqq)=val(qqq)+term
!         secondary diagonal:
          qqq=point_no_Dir(p+nnod)
          if (qqq>0) then
            h_tr=row_ptr(ppp)-1
            if (col_idx(h_tr)==qqq-1) then
              val(h_tr)=-term
            else 
              write(*,*) 'Error! Secondary diagonal element not found!'
              stop
            endif
          else
            effe_Dir(ppp)=effe_Dir(ppp)+term*ggg_e(xxx(p),yyy(p),zzz(p))
          endif
        endif
      enddo

!     2nd block of DOF_no_Dir_e rows: diagonals
      do p=1,nnod
        ppp=point_no_Dir(p+nnod)
        if (ppp>0) then
          term=c_m_per_vol(xxx(p),yyy(p),zzz(p))*omega_tot(p)/deltat
          trovato=.false.
          loop_cerca_diag_2: do qqq=row_ptr(ppp-1),row_ptr(ppp)-1
            if (col_idx(qqq)==ppp-1) then
              trovato=.true.
              exit loop_cerca_diag_2
            endif
          enddo loop_cerca_diag_2
          if (.not.trovato) then
            write(*,*) 'Error! Diagonal element not found!'
            stop
          endif
          val(qqq)=val(qqq)+term
!         secondary diagonal:
          qqq=point_no_Dir(p)
          if (qqq>0) then
            h_bl=row_ptr(ppp-1)
            if (col_idx(h_bl)==qqq-1) then
              val(h_bl)=-term
            else 
              write(*,*) 'Error! Secondary diagonal element not found!'
              stop
            endif
          else
            effe_Dir(ppp)=effe_Dir(ppp)+term*ggg_i(xxx(p),yyy(p),zzz(p))
          endif
        endif
      enddo

      return

      end subroutine make_global_matrix_CSR


!     --------------------------------------------------------------------------


      real(8) function mcap_i(i,j,xxx,yyy,zzz)
                       
!     conductivity tensor M_i

      implicit none

      integer(4) :: i,j
      real(8) :: xxx,yyy,zzz
      real(8) :: Chi, C_m, D_i, D_e
      common / membrane / Chi, C_m, D_i, D_e

      if (i==1) then
        if (j==1) then
          mcap_i=1.d0
        else if (j==2) then
          mcap_i=0.0d0
        else if (j==3) then
          mcap_i=0.0d0
        endif
      else if (i==2) then
        if (j==1) then
          mcap_i=0.0d0
        else if (j==2) then
          mcap_i=1.d0
        else if (j==3) then
          mcap_i=0.0d0
        endif
      else if (i==3) then
        if (j==1) then
          mcap_i=0.0d0
        else if (j==2) then
          mcap_i=0.0d0
        else if (j==3) then
          mcap_i=1.d0
        endif
      endif

      mcap_i=D_i*mcap_i

      return

      end function mcap_i


!     --------------------------------------------------------------------------


      real(8) function mcap_e(i,j,xxx,yyy,zzz)
                       
!     conductivity tensor M_e

      implicit none

      integer(4) :: i,j
      real(8) :: xxx,yyy,zzz
      real(8) :: Chi, C_m, D_i, D_e
      common / membrane / Chi, C_m, D_i, D_e

      if (i==1) then
        if (j==1) then
          mcap_e=1.d0
        else if (j==2) then
          mcap_e=0.0d0
        else if (j==3) then
          mcap_e=0.0d0
        endif
      else if (i==2) then
        if (j==1) then
          mcap_e=0.0d0
        else if (j==2) then
          mcap_e=1.d0
        else if (j==3) then
          mcap_e=0.0d0
        endif
      else if (i==3) then
        if (j==1) then
          mcap_e=0.0d0
        else if (j==2) then
          mcap_e=0.0d0
        else if (j==3) then
          mcap_e=1.d0
        endif
      endif

      mcap_e=D_e*mcap_e

      return

      end function mcap_e


!     --------------------------------------------------------------------------


      subroutine read_grid_parameters(fichier,nnod,nelem)

!     The grid parameters (#nodes, #elements, etc) are read 

      implicit none

      integer(4) :: ic, nnod, nelem
      character :: fichier*512, riga*512, linea*1

      open(23,file=fichier)
      riga=' '; nnod=0; nelem=0
99    continue
      do while (index(riga,'*NODE')+index(riga,'*ELEMENT')==0) 
        read(23,'(a)',end=102) riga
      enddo
      ic=0; linea=' '
      do
        read(23,'(a1)',end=101) linea
        if (linea=='*') goto 101
        ic=ic+1
      enddo
101   continue
      backspace(23)

      if (index(riga,'*NODE')>0) then
        nnod=nnod+ic
      else if(index(riga,'*ELEMENT')>0) then
        nelem=nelem+ic
      endif
      read(23,'(a)',end=102) riga
      goto 99
102   continue
      close(23)

      return

      end subroutine read_grid_parameters


!     --------------------------------------------------------------------------


      subroutine read_grid(fichier,nnod,nelem,xxx,yyy,zzz,con,mat)


!     The grid is read (nodes, elements, etc)

      implicit none
      integer(4) :: i,j,ic,iblock,ie,i1,i2,i3,i4,i5,i6,i7,i8
      integer(4) :: nnod,nelem,con(8,nelem)
      integer(4) :: mat(nelem)
      real(8) :: xxx(nnod), yyy(nnod), zzz(nnod)
      character*512 :: fichier, riga, linea

      open(23,file=fichier)
      riga=' '
99    continue
      do while (index(riga,'*NODE')+index(riga,'*ELEMENT')==0) 
        read(23,'(a)',end=102) riga
      enddo
      ic=0

      if (index(riga,'*NODE')>0) then
        do
          read(23,'(a)',end=102) linea
          if (linea==''.or.index(linea,'*')>0) goto 101
          read(linea,*) i,xxx(i),yyy(i),zzz(i)
          if (i>nnod) then
            write(*,*)
            write(*,*) "Error! Non consecutive gridpoints' indices!"
            write(*,*) "Run the 'compress' command in Cubit"
            write(*,*)
            stop
          endif 
        enddo
      else if(index(riga,'*ELEMENT')>0) then
        if (index(riga,'TYPE=C3D8R')>0) then
          ic=index(riga,'ELSET=EB')
          loop_cerca_fine_riga: do i=ic+1,len_trim(riga)
            if (riga(i:i)==''.or.riga(i:i)==char(9)) exit loop_cerca_fine_riga
          enddo loop_cerca_fine_riga
          read(riga(ic+8:i-1),*) iblock
          do 
            read(23,'(a)',end=102) linea
            if (linea==''.or.index(linea,'*')>0) goto 101
            read(linea,*) ie, (con(j,ie), j=1,8)
            if (ie>nelem) then
              write(*,*)
              write(*,*) "Error! Non consecutive elements' indices!"
              write(*,*) "If using Cubit, run the 'compress' command"
              write(*,*)
              stop
            endif 
            mat(ie)=iblock
          enddo
        else
          stop "Error! Non-hexahedral elements!"
        endif
      endif
101   continue
      backspace(23)
      read(23,'(a)',end=102) riga
      goto 99
102   continue
      close(23)

      return

      end subroutine read_grid


!     --------------------------------------------------------------------------


      subroutine draw_vtk(fichier,titolo,nn,nnod,nelem,xxx,yyy,zzz,con,uuu)

!     draw the array uuu in vtk format for visualization with Mayavi

      implicit none
      integer(4) :: nn, nn2, nnod, nelem, con(nn**3,nelem), i, j, k, ie, iii
      integer(4) :: i0, i1, i2, i3, i4, i5, i6,i7
      real(8) :: xxx(nnod), yyy(nnod), zzz(nnod), uuu(nnod)
      character :: fichier*512, titolo*512

      nn2=nn*nn
      open(53,file=fichier)
      write(53,530)
530   format(t1,"# vtk DataFile Version 2.0") 
      write(53,531) titolo(1:len_trim(titolo))
531   format(t1,a)
      write(53,532)
532   format(t1,"ASCII")
      write(53,533)
533   format(t1,"DATASET UNSTRUCTURED_GRID")    
      write(53,534) nnod
534   format(t1,"POINTS ",i10," float")   
      do i=1,nnod
        if (dabs(xxx(i))<1.d-30) xxx(i)=0.d0
        if (dabs(yyy(i))<1.d-30) yyy(i)=0.d0
        if (dabs(zzz(i))<1.d-30) zzz(i)=0.d0
        write(53,535) xxx(i),yyy(i),zzz(i)
      enddo
535   format(t1,g12.6,2(2x,g12.6))  
      write(53,536) nelem*(nn-1)**3, 9*nelem*(nn-1)**3
536   format(t1,"CELLS ",i10,2x,i10)     
      do ie=1,nelem
        do k=2,nn
          do j=2,nn
            do i=2,nn
              i0=con((k-2)*nn2+(j-2)*nn+i,ie)-1
              i1=con((k-2)*nn2+(j-1)*nn+i,ie)-1
              i2=con((k-2)*nn2+(j-1)*nn+i-1,ie)-1
              i3=con((k-2)*nn2+(j-2)*nn+i-1,ie)-1
              i4=con((k-1)*nn2+(j-2)*nn+i,ie)-1
                i5=con((k-1)*nn2+(j-1)*nn+i,ie)-1
              i6=con((k-1)*nn2+(j-1)*nn+i-1,ie)-1
              i7=con((k-1)*nn2+(j-2)*nn+i-1,ie)-1
              write(53,537) i0, i1, i2, i3, i4, i5, i6, i7
            enddo 
          enddo
        enddo
      enddo
537   format(t1,"8 ",8(1x,i8))  
      write(53,538) nelem*(nn-1)**3
538   format(t1,"CELL_TYPES ",i10) 
      do i=1,nelem*(nn-1)**3
        write(53,539) 
      enddo
539   format(t1,"12")
      write(53,540) nnod
540   format(t1,"POINT_DATA ",i10) 
      write(53,541) 
541   format(t1,"SCALARS cardiac_potential float 1")
      write(53,542) 
542   format(t1,"LOOKUP_TABLE default")
      do i=1,nnod
        if (dabs(uuu(i))<1.d-16) uuu(i)=0.d0
      enddo
      do i=1,nnod
        write(53,543) uuu(i)
      enddo
543   format(t1,g12.6)
      close(53)

      return
      
      end subroutine draw_vtk


!     --------------------------------------------------------------------------


      subroutine mio_shell(n,a)

!     integer array a, with dim=n, is ordered in ascending way

      implicit none
      integer(4) :: n,i,j,inc,v,a(n)
      inc=1
1     inc=3*inc+1
      if(inc.le.n)goto 1
2     continue
      inc=inc/3
        do i=inc+1,n
          v=a(i)
          j=i
3         if(a(j-inc).gt.v)then
            a(j)=a(j-inc)
            j=j-inc
            if(j.le.inc)goto 4
          goto 3
          endif
4         a(j)=v
        enddo
      if(inc.gt.1)goto 2
      return

      end subroutine mio_shell


!     --------------------------------------------------------------------------


      subroutine make_bw(nn,nnod,nelem,con_spx,point_no_Dir,bw,nnz_tot)

!     bandwith of the stiffness matrix

      implicit none

      integer(4) :: nn, nn2, nnod, nelem, nnz_tot, bw_max, stencil
      integer(4) :: i, j, k, ie, iii, l, ie_ot, i1, j1, k1
      integer(4) :: iii1, iii2, jjj, p, q, nserv, ic, ic1, ic2
      integer(4) :: con_spx(nn**3,nelem), bw(2*nnod), point_no_Dir(2*nnod)
      integer(4), allocatable :: banda(:), posizione(:), con_dual(:,:)
      integer(4), allocatable :: serv1(:), serv2(:)
             
      nn2=nn*nn
!     stencil=3*nn2-3*nn
      stencil=3*nn2-3

      allocate (banda(nnod)); banda=0
      do ie=1,nelem
        do k=1,nn
          do j=1,nn
            do i=1,nn
              iii=con_spx((k-1)*nn2+(j-1)*nn+i,ie)
              banda(iii)=banda(iii)+1
            enddo
          enddo
        enddo
      enddo
      bw_max=maxval(banda)

      allocate (con_dual(bw_max,nnod)); con_dual=0
      allocate (posizione(nnod)); posizione=0      
      do ie=1,nelem
        do k=1,nn
          do j=1,nn
            do i=1,nn
              iii=con_spx((k-1)*nn2+(j-1)*nn+i,ie)
              if (point_no_Dir(iii)>0.or.point_no_Dir(iii+nnod)>0) then
                posizione(iii)=posizione(iii)+1
                if (posizione(iii)>bw_max)                                     &
                  stop "Error! Insufficient allocation in make_spectral_connect&
ivities!"
                con_dual(posizione(iii),iii)=ie
              endif
            enddo
          enddo
        enddo
      enddo
      deallocate(posizione)

      bw=0
      do ie=1,nelem
        do k=1,nn
          do j=1,nn
            do i=1,nn
              iii=con_spx((k-1)*nn2+(j-1)*nn+i,ie)
              
              if (point_no_Dir(iii)>0.or.point_no_Dir(iii+nnod)>0) then
                if (bw(iii)==0) then
                  nserv=banda(iii)*stencil

                  if (point_no_Dir(iii)>0.and.point_no_Dir(iii+nnod)>0) then

                    allocate(serv1(nserv),serv2(nserv)); serv1=0; serv2=0                            
                    ic1=0; ic2=0
                    do l=1,banda(iii)

                      ie_ot=con_dual(l,iii)
                      loop_cerca_node_equiv_1: do k1=1,nn
                        do j1=1,nn
                          do i1=1,nn
                            iii1=con_spx((k1-1)*nn2+(j1-1)*nn+i1,ie_ot)
                            if (iii1==iii) exit loop_cerca_node_equiv_1
                          enddo
                         enddo
                      enddo loop_cerca_node_equiv_1

                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((k1-1)*nn2+(q-1)*nn+p,ie_ot)
                          if (jjj.ne.iii)  then
                            iii1=point_no_Dir(jjj); iii2=point_no_Dir(jjj+nnod)
                            if (iii1>0)  then
                              ic1=ic1+1
                              serv1(ic1)=iii1
                            endif
                            if (iii2>0)  then
                              ic2=ic2+1
                              serv2(ic2)=iii2
                            endif
                          endif
                        enddo
                      enddo
                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((q-1)*nn2+(j1-1)*nn+p,ie_ot)
                          if (jjj.ne.iii)  then
                            iii1=point_no_Dir(jjj); iii2=point_no_Dir(jjj+nnod)
                            if (iii1>0)  then
                              ic1=ic1+1
                              serv1(ic1)=iii1
                            endif
                            if (iii2>0)  then
                              ic2=ic2+1
                              serv2(ic2)=iii2
                            endif
                          endif
                        enddo
                      enddo
                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((q-1)*nn2+(p-1)*nn+i1,ie_ot)
                          if (jjj.ne.iii)  then
                            iii1=point_no_Dir(jjj); iii2=point_no_Dir(jjj+nnod)
                            if (iii1>0)  then
                              ic1=ic1+1
                              serv1(ic1)=iii1
                            endif
                            if (iii2>0)  then
                              ic2=ic2+1
                              serv2(ic2)=iii2
                            endif
                          endif
                        enddo
                      enddo

                    enddo    ! loop over elements adjacent to node iii

                    call mio_shell(nserv,serv1)
                    call mio_shell(nserv,serv2)

                    loop_cerca_non_zero_11: do l=1,nserv
                      if (serv1(l).ne.0) exit loop_cerca_non_zero_11
                    enddo loop_cerca_non_zero_11
                    ic1=1
                    do jjj=l+1,nserv
                      if (serv1(jjj).ne.serv1(jjj-1)) ic1=ic1+1
                    enddo
                    loop_cerca_non_zero_21: do l=1,nserv
                      if (serv2(l).ne.0) exit loop_cerca_non_zero_21
                    enddo loop_cerca_non_zero_21
                    ic2=1
                    do jjj=l+1,nserv
                      if (serv2(jjj).ne.serv2(jjj-1)) ic2=ic2+1
                    enddo
                    bw(iii)=ic1; bw(iii+nnod)=ic2
                    deallocate (serv1, serv2)

                  else if (point_no_Dir(iii+nnod)==0) then

                    allocate(serv1(nserv)); serv1=0                            
                    ic1=0
                    do l=1,banda(iii)

                      ie_ot=con_dual(l,iii)
                      loop_cerca_node_equiv_2: do k1=1,nn
                        do j1=1,nn
                          do i1=1,nn
                            iii1=con_spx((k1-1)*nn2+(j1-1)*nn+i1,ie_ot)
                            if (iii1==iii) exit loop_cerca_node_equiv_2
                          enddo
                         enddo
                      enddo loop_cerca_node_equiv_2

                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((k1-1)*nn2+(q-1)*nn+p,ie_ot)
                          if (jjj.ne.iii)  then
                            iii1=point_no_Dir(jjj)
                            if (iii1>0)  then
                              ic1=ic1+1
                              serv1(ic1)=iii1
                            endif
                          endif
                        enddo
                      enddo
                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((q-1)*nn2+(j1-1)*nn+p,ie_ot)
                          if (jjj.ne.iii)  then
                            iii1=point_no_Dir(jjj)
                            if (iii1>0)  then
                              ic1=ic1+1
                              serv1(ic1)=iii1
                            endif
                          endif
                        enddo
                      enddo
                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((q-1)*nn2+(p-1)*nn+i1,ie_ot)
                          if (jjj.ne.iii)  then
                            iii1=point_no_Dir(jjj)
                            if (iii1>0)  then
                              ic1=ic1+1
                              serv1(ic1)=iii1
                            endif
                          endif
                        enddo
                      enddo

                    enddo    ! loop over elements adjacent to node iii

                    call mio_shell(nserv,serv1)

                    loop_cerca_non_zero_12: do l=1,nserv
                      if (serv1(l).ne.0) exit loop_cerca_non_zero_12
                    enddo loop_cerca_non_zero_12
                    ic1=1
                    do jjj=l+1,nserv
                      if (serv1(jjj).ne.serv1(jjj-1)) ic1=ic1+1
                    enddo
                    bw(iii)=ic1
                    deallocate (serv1)

                  else if (point_no_Dir(iii)==0) then

                    allocate(serv2(nserv)); serv2=0                            
                    ic2=0
                    do l=1,banda(iii)

                      ie_ot=con_dual(l,iii)
                      loop_cerca_node_equiv_3: do k1=1,nn
                        do j1=1,nn
                          do i1=1,nn
                            iii1=con_spx((k1-1)*nn2+(j1-1)*nn+i1,ie_ot)
                            if (iii1==iii) exit loop_cerca_node_equiv_3
                          enddo
                         enddo
                      enddo loop_cerca_node_equiv_3

                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((k1-1)*nn2+(q-1)*nn+p,ie_ot)
                          if (jjj.ne.iii)  then
                            iii2=point_no_Dir(jjj+nnod)
                            if (iii2>0)  then
                              ic2=ic2+1
                              serv2(ic2)=iii2
                            endif
                          endif
                        enddo
                      enddo
                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((q-1)*nn2+(j1-1)*nn+p,ie_ot)
                          if (jjj.ne.iii)  then
                            iii2=point_no_Dir(jjj+nnod)
                            if (iii2>0)  then
                              ic2=ic2+1
                              serv2(ic2)=iii2
                            endif
                          endif
                        enddo
                      enddo
                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((q-1)*nn2+(p-1)*nn+i1,ie_ot)
                          if (jjj.ne.iii)  then
                            iii2=point_no_Dir(jjj+nnod)
                            if (iii2>0)  then
                              ic2=ic2+1
                              serv2(ic2)=iii2
                            endif
                          endif
                        enddo
                      enddo

                    enddo    ! loop over elements adjacent to node iii

                    call mio_shell(nserv,serv2)

                    loop_cerca_non_zero_23: do l=1,nserv
                      if (serv2(l).ne.0) exit loop_cerca_non_zero_23
                    enddo loop_cerca_non_zero_23
                    ic2=1
                    do jjj=l+1,nserv
                      if (serv2(jjj).ne.serv2(jjj-1)) ic2=ic2+1
                    enddo
                    bw(iii+nnod)=ic2
                    deallocate (serv2)

                  endif
                
                endif
              endif

            enddo
          enddo
        enddo
      enddo

!     bw takes into account all non-diagonal entries, 
!     including the secondary diagonal

      do i=1,nnod
        if (point_no_Dir(i)>0) then
          if (point_no_Dir(i+nnod)>0) bw(i)=bw(i)+1
        endif
      enddo
      do i=nnod+1,2*nnod
        if (point_no_Dir(i)>0) then
          if (point_no_Dir(i-nnod)>0) bw(i)=bw(i)+1
        endif
      enddo

      nnz_tot=0
      do i=1,2*nnod
        if (point_no_Dir(i)>0) nnz_tot=nnz_tot+bw(i)+1
      enddo
      return

      end subroutine make_bw


!     --------------------------------------------------------------------------


      subroutine make_CSR_indices(nn,nnod,nelem,nrig,nnz,con_spx,point_no_Dir, &
                                  bw,row_ptr,col_idx)

!     index arrays row_ptr and col_idx of the global marix in C-style CSR format
!     are generated

      implicit none

      integer(4) :: nn, nn2, nnod, nelem, nnz, bw_max, stencil, nrig
      integer(4) :: i, j, k, ie, iii, l, ie_ot, i1, j1, k1, irig1, irig2
      integer(4) :: iii1, iii2, jjj, p, q, nserv, ic, ic1, ic2
      integer(4) :: con_spx(nn**3,nelem), bw(2*nnod), point_no_Dir(2*nnod)
      integer(4) :: row_ptr(0:nrig), col_idx(0:nnz-1)
      integer(4), allocatable :: banda(:), posizione(:), con_dual(:,:)
      integer(4), allocatable :: serv1(:), serv2(:)
             
      nn2=nn*nn
      stencil=3*nn2
      
      row_ptr=-1
      row_ptr(0)=0
      do i=1,2*nnod
        iii=point_no_Dir(i)
        if (iii>0) row_ptr(iii)=row_ptr(iii-1)+bw(i)+1
      enddo

      allocate (banda(nnod)); banda=0
      do ie=1,nelem
        do k=1,nn
          do j=1,nn
            do i=1,nn
              iii=con_spx((k-1)*nn2+(j-1)*nn+i,ie)
              banda(iii)=banda(iii)+1
            enddo
          enddo
        enddo
      enddo
      bw_max=maxval(banda)

      allocate (con_dual(bw_max,nnod)); con_dual=0
      allocate (posizione(nnod)); posizione=0      
      do ie=1,nelem
        do k=1,nn
          do j=1,nn
            do i=1,nn
              iii=con_spx((k-1)*nn2+(j-1)*nn+i,ie)
              if (point_no_Dir(iii)>0.or.point_no_Dir(iii+nnod)>0) then
                posizione(iii)=posizione(iii)+1
                if (posizione(iii)>bw_max)                                     &
                  stop "Error! Insufficient allocation in make_spectral_connect&
ivities!"
                con_dual(posizione(iii),iii)=ie
              endif
            enddo
          enddo
        enddo
      enddo
      deallocate(posizione)

      do ie=1,nelem
        do k=1,nn
          do j=1,nn
            do i=1,nn
              iii=con_spx((k-1)*nn2+(j-1)*nn+i,ie)
              irig1=point_no_Dir(iii); irig2=point_no_Dir(iii+nnod)

              if (irig1>0.or.irig2>0) then

                if (banda(iii)>0) then
                  nserv=banda(iii)*stencil

                  if (irig1>0.and.irig2>0) then

                    allocate(serv1(nserv+1),serv2(nserv+1)); serv1=0; serv2=0                            
                    ic1=0; ic2=0
                    do l=1,banda(iii)

                      ie_ot=con_dual(l,iii)
                      loop_cerca_node_equiv_1: do k1=1,nn
                        do j1=1,nn
                          do i1=1,nn
                            iii1=con_spx((k1-1)*nn2+(j1-1)*nn+i1,ie_ot)
                            if (iii1==iii) exit loop_cerca_node_equiv_1
                          enddo
                         enddo
                      enddo loop_cerca_node_equiv_1

                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((k1-1)*nn2+(q-1)*nn+p,ie_ot)
                          iii1=point_no_Dir(jjj); iii2=point_no_Dir(jjj+nnod)
                          if (iii1>0)  then
                            ic1=ic1+1
                            serv1(ic1)=iii1
                          endif
                          if (iii2>0)  then
                            ic2=ic2+1
                            serv2(ic2)=iii2
                          endif
                        enddo
                      enddo
                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((q-1)*nn2+(j1-1)*nn+p,ie_ot)
                          iii1=point_no_Dir(jjj); iii2=point_no_Dir(jjj+nnod)
                          if (iii1>0)  then
                            ic1=ic1+1
                            serv1(ic1)=iii1
                          endif
                          if (iii2>0)  then
                            ic2=ic2+1
                            serv2(ic2)=iii2
                          endif
                        enddo
                      enddo
                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((q-1)*nn2+(p-1)*nn+i1,ie_ot)
                          iii1=point_no_Dir(jjj); iii2=point_no_Dir(jjj+nnod)
                          if (iii1>0)  then
                            ic1=ic1+1
                            serv1(ic1)=iii1
                          endif
                          if (iii2>0)  then
                            ic2=ic2+1
                            serv2(ic2)=iii2
                          endif
                        enddo
                      enddo

                    enddo    ! loop over elements adjacent to node iii

!                   secondary diagonal element:
                    ic1=ic1+1
                    serv1(ic1)=irig2
                    ic2=ic2+1
                    serv2(ic2)=irig1

                    call mio_shell(nserv+1,serv1)
                    call mio_shell(nserv+1,serv2)

                    loop_cerca_non_zero_11: do l=1,nserv+1
                      if (serv1(l).ne.0) exit loop_cerca_non_zero_11
                    enddo loop_cerca_non_zero_11
                    ic1=row_ptr(irig1-1)

                    col_idx(ic1)=serv1(l)-1
                    do jjj=l+1,nserv+1
                      if (serv1(jjj).ne.serv1(jjj-1)) then
                        ic1=ic1+1
                        col_idx(ic1)=serv1(jjj)-1                       
                      endif
                    enddo
                    loop_cerca_non_zero_21: do l=1,nserv+1
                      if (serv2(l).ne.0) exit loop_cerca_non_zero_21
                    enddo loop_cerca_non_zero_21
                    ic2=row_ptr(irig2-1) 
                    col_idx(ic2)=serv2(l)-1
                    do jjj=l+1,nserv+1
                      if (serv2(jjj).ne.serv2(jjj-1)) then
                        ic2=ic2+1
                        col_idx(ic2)=serv2(jjj)-1
                      endif
                    enddo
                    deallocate (serv1, serv2)

                  else if (irig2==0) then

                    allocate(serv1(nserv)); serv1=0                            
                    ic1=0
                    do l=1,banda(iii)

                      ie_ot=con_dual(l,iii)
                      loop_cerca_node_equiv_2: do k1=1,nn
                        do j1=1,nn
                          do i1=1,nn
                            iii1=con_spx((k1-1)*nn2+(j1-1)*nn+i1,ie_ot)
                            if (iii1==iii) exit loop_cerca_node_equiv_2
                          enddo
                         enddo
                      enddo loop_cerca_node_equiv_2

                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((k1-1)*nn2+(q-1)*nn+p,ie_ot)
                          iii1=point_no_Dir(jjj)
                          if (iii1>0)  then
                            ic1=ic1+1
                            serv1(ic1)=iii1
                          endif
                        enddo
                      enddo
                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((q-1)*nn2+(j1-1)*nn+p,ie_ot)
                          iii1=point_no_Dir(jjj)
                          if (iii1>0)  then
                            ic1=ic1+1
                            serv1(ic1)=iii1
                          endif
                        enddo
                      enddo
                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((q-1)*nn2+(p-1)*nn+i1,ie_ot)
                          iii1=point_no_Dir(jjj)
                          if (iii1>0)  then
                            i1=ic1+1
                            serv1(ic1)=iii1
                          endif
                        enddo
                      enddo

                    enddo    ! loop over elements adjacent to node iii

                    call mio_shell(nserv,serv1)

                    loop_cerca_non_zero_12: do l=1,nserv
                      if (serv1(l).ne.0) exit loop_cerca_non_zero_12
                    enddo loop_cerca_non_zero_12

                    ic1=row_ptr(irig1-1)
                    col_idx(ic1)=serv1(l)-1
                    do jjj=l+1,nserv
                      if (serv1(jjj).ne.serv1(jjj-1)) then
                        ic1=ic1+1
                        col_idx(ic1)=serv1(jjj)-1                       
                      endif
                    enddo
                    deallocate (serv1)

                  else if (irig1==0) then

                    allocate(serv2(nserv)); serv2=0                            
                    ic2=0
                    do l=1,banda(iii)

                      ie_ot=con_dual(l,iii)
                      loop_cerca_node_equiv_3: do k1=1,nn
                        do j1=1,nn
                          do i1=1,nn
                            iii1=con_spx((k1-1)*nn2+(j1-1)*nn+i1,ie_ot)
                            if (iii1==iii) exit loop_cerca_node_equiv_3
                          enddo
                         enddo
                      enddo loop_cerca_node_equiv_3

                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((k1-1)*nn2+(q-1)*nn+p,ie_ot)
                          iii2=point_no_Dir(jjj+nnod)
                          if (iii2>0)  then
                            ic2=ic2+1
                            serv2(ic2)=iii2
                          endif
                        enddo
                      enddo
                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((q-1)*nn2+(j1-1)*nn+p,ie_ot)
                          iii2=point_no_Dir(jjj+nnod)
                          if (iii2>0)  then
                            ic2=ic2+1
                            serv2(ic2)=iii2
                          endif
                        enddo
                      enddo
                      do p=1,nn
                        do q=1,nn
                          jjj=con_spx((q-1)*nn2+(p-1)*nn+i1,ie_ot)
                          iii2=point_no_Dir(jjj+nnod)
                          if (iii2>0)  then
                            ic2=ic2+1
                            serv2(ic2)=iii2
                          endif
                        enddo
                      enddo

                    enddo    ! loop over elements adjacent to node iii

                    call mio_shell(nserv,serv2)

                    loop_cerca_non_zero_23: do l=1,nserv
                      if (serv2(l).ne.0) exit loop_cerca_non_zero_23
                    enddo loop_cerca_non_zero_23
                    ic2=row_ptr(irig2-1)   
                    col_idx(ic2)=serv2(l)-1
                    do jjj=l+1,nserv
                      if (serv2(jjj).ne.serv2(jjj-1)) then
                        ic2=ic2+1
                        col_idx(ic2)=serv2(jjj)-1
                      endif
                    enddo
                    deallocate (serv2)

                  endif
                
                endif
                banda(iii)=0
              endif

            enddo
          enddo
        enddo
      enddo

      return

      end subroutine make_CSR_indices


!     --------------------------------------------------------------------------


      subroutine time_loop_CSR(myid,master,nn,nnod,nelem,nrig,nnz_tot,nrec,xxx,&
                               yyy,zzz,con_spx,index_rec,t0,tfin,ttt,          &
                               deltat_ODE,deltat_PDE,omega,effe_Dir,           &
                               point_no_Dir,effe_Neu,row_ptr,col_idx,val,uuu,  &
                               Dir_not_empty,Neu_not_empty,write_it_hist,      &
                               write_vtk,write_xdmf,it_snap,nsnap,niter_ave,   &
                               cg_time1,cg_time2)

      implicit none

      include "Luo_Rudy_I_1991_f90.inc"

      integer(4) :: myid,master
      integer(4) :: nn, nn2, nn3, nnod, nelem, nnz_tot, nts_ODE, iter_max, nrec
      integer(4) :: i,j,k,l,m,ie,iii,jjj,ic,its,nts_frac(3),fa,niter,nrig, nsnap
      integer(4) :: con_spx(nn**3,nelem), index_rec(nrec), it_snap(0:nsnap-1)
      integer(4) :: deltat_PDE_divided_ODE, nts_PDE, ODE_step
      integer(4) :: i0, i1, i2, i3, i4, j1, k1, irec, isnap, llll
      integer(4) :: point_no_Dir(2*nnod), row_ptr(0:nrig), col_idx(0:nnz_tot-1)
      real(8) :: ttt, ttt0, t0, tfin, deltat_ODE, deltat_PDE, zero=0.d0, vol_om
      real(8) :: tempo1, tempo2, cg_time1, cg_time2, cost, term_i, term_e
      real(8) :: area, norm_const, term
      real(8) :: xxx(nnod), yyy(nnod), zzz(nnod), uuu(nrig)
      real(8) :: effe(nrig), c_m_M_delta_t(nnod), effe_Dir(nrig), effe_Neu(nrig)
      real(8) :: val(0:nnz_tot-1), vvv(nnod), u_i(nnod), u_e(nnod), omega(nnod)
      real(8) :: serv0, serv1, serv2, serv3, xx, yy, zz, eps=1.d-12,psi_i,psi_e
      real(8), external :: c_m_per_vol, mcap_i, mcap_e, ggg_i, ggg_e, hhh
      real(8) :: amp_I_app_i, amp_I_app_e
      real(8) :: time_ODE_begin, time_ODE_end, tempo_ODE, toll, time_term
      real(8) :: time_PDE_begin, time_PDE_end, tempo_PDE, ttt_PDE, niter_ave
      real(8) :: beg1,end1,temp1,beg2,end2,temp2,term1,term2
      real(8) :: x1, x2, y1, y2, z1, z2
      character :: fichier(nrec)*512, formato*80, TITOLO*512
      character :: riga512*512, solpar*512
      logical :: write_it_hist, Dir_not_empty, Neu_not_empty     
      logical :: write_vtk, write_xdmf, exists
!     LR1 model:
!     y_CellMl are the unknowns of the ODE in the LR1 model
      real(8) :: y_CellMl(8,nnod), y_CellMl_point(8), ion_curr, dy(8)
      real(8), external :: I_i_app, I_e_app
      real(8) :: Chi, C_m, D_i, D_e

      common / domain_box / x1, x2, y1, y2, z1, z2
      common / applied_Currents / amp_I_app_i, amp_I_app_e
      common / membrane / Chi, C_m, D_i, D_e
      common /Y/y_CellMl_point
      common /DY/dy

      nn2=nn*nn; nn3=nn**3

      term=dabs(dfloat(idint(deltat_PDE/deltat_ODE))-deltat_PDE/deltat_ODE)    &
          /(deltat_PDE/deltat_ODE)
      if (term>eps) then
        stop "Error! deltat_PDE should be an exact multiple of deltat_ODE!"
      else
        deltat_PDE_divided_ODE=idint(deltat_PDE/deltat_ODE)
      endif

      nts_ODE=idnint((tfin-t0)/deltat_ODE); nts_PDE=0

      fichier="rec_000.d"
      do irec=1,nrec
        if (nrec<10) then
          write(fichier(irec)(7:7),'(i1)') irec
        else if (nrec<100) then
          write(fichier(irec)(6:7),'(i2)') irec
        else if (nrec<1000) then
          write(fichier(irec)(5:7),'(i3)') irec
        else 
          stop "Error! Too many receivers!"
        endif
        open(100+irec,file=fichier(irec))
        write(100+irec,201)
        write(100+irec,202) xxx(index_rec(irec)), yyy(index_rec(irec)),        &
                            zzz(index_rec(irec))
      enddo
201   format(t1,"# time-history of the transmembrane potential evaluated at ")
202   format(t1,"# point (",g9.3,",",g9.3,",",g9.3,")") 

      if (myid==master) open(33,file="stimulus.d")

      do i=1,nnod
        c_m_M_delta_t(i)=c_m_per_vol(xxx(i),yyy(i),zzz(i))*omega(i)/deltat_PDE 
      enddo

!     Initial conditions are set s.t. v = u_i-u_e = y_CellMl_point,
!     i.e. v(t=0) from the LR1 model is set *everywhere*:
      call init
      do i=1,nnod
         u_i(i)=y_CellMl_point(5)
         u_e(i)=zero
         vvv(i)=u_i(i)-u_e(i)
      enddo
       do i=1,nnod
         iii=point_no_Dir(i)
         if (iii>0) uuu(iii)=y_CellMl_point(5)    
         iii=point_no_Dir(i+nnod)
         if (iii>0) uuu(iii)=zero    
       enddo

      do i=1,nnod              
        do j=1,8
          y_CellMl(j,i)=y_CellMl_point(j)
        enddo
      enddo

      tempo_ODE=0.d0; tempo_PDE=0.d0; ODE_step=0; temp1=zero; temp2=zero
      ttt0=t0; ttt_PDE=ttt0; ic=1; niter=0; cg_time1=zero; cg_time2=zero
      isnap=0; llll=0

      if (nts_ODE.ge.4) then
        do i=1,3
          nts_frac(i)=idnint(dfloat(i)*0.25d0*dfloat(nts_ODE))
        enddo
      endif

      open(23,file='lin_solv.input')
      read(23,*)             ! algebraic_solver
      k=0
      loop_0: do
        read(23,'(a)') riga512
        loop_1: do j=1,512
          if (riga512(j:j).ne.' ') exit loop_1
        enddo loop_1
        if (riga512(j:j).ne.'!') then
          backspace(23)
          exit loop_0
        endif
      enddo loop_0
      read(23,*) iter_max
      read(23,*) toll
      close(23)

      if (write_it_hist.and.myid==master)                                      &
          open(23,file='iterations_info.output',position='append')

      do its=1,nts_ODE

        ttt=ttt0+deltat_ODE
        ODE_step=ODE_step+1

        if (write_it_hist.and.myid==master) then
          write(23,*)
          write(23,231) its, ttt
        endif

!       the membrane variables are updated numerically:
        call CPU_TIME(time_ODE_begin)
        do i=1,nnod
          if (point_no_Dir(i)>0.or.point_no_Dir(i+nnod)>0) then
            y_CellMl_point(:)=y_CellMl(:,i)
!           before the call to compute_imp, y_CellMl_point(5) is the 
!           transmembrane potential [mV]
            y_CellMl_point(5)=vvv(i)
            call compute_I(ttt,deltat_ODE)
            y_CellMl(:,i)=y_CellMl_point(:)
          endif
        enddo
        call CPU_TIME(time_ODE_end)
        tempo_ODE=tempo_ODE+time_ODE_end-time_ODE_begin

!       PDE's solution is updated, if needed
        if (ODE_step==deltat_PDE_divided_ODE) then
          call CPU_TIME(time_PDE_begin)
          ttt_PDE=ttt
          nts_PDE=nts_PDE+1
          ODE_step=0
          time_term=hhh(ttt_PDE)
          if (myid==master) write(33,*) ttt_PDE, time_term

!         the RHS is built
          call CPU_TIME(beg1)
          effe=0.d0

!         term due to ionic current (membrane model)
          do i=1,nnod
            iii=point_no_Dir(i)
            jjj=point_no_Dir(i+nnod)
            if (iii>0.or.jjj>0) ion_curr=Chi*y_CellMl(5,i)*omega(i)
!           after the call to compute_I, y_CellMl_point(5) is the 
!           ionic current i_ion. We need I_ion = Chi*i_ion [microA/cm3]
            if (iii>0) effe(iii)=effe(iii)-ion_curr
            if (jjj>0) effe(jjj)=effe(jjj)+ion_curr
          enddo    

!         term due to time-marching
          do i=1,nnod
            iii=point_no_Dir(i)
            jjj=point_no_Dir(i+nnod)
            if (iii>0) effe(iii)=effe(iii)+c_m_M_delta_t(i)*vvv(i)
            if (jjj>0) effe(jjj)=effe(jjj)-c_m_M_delta_t(i)*vvv(i)
          enddo

!         term due to applied currents
          do i=1,nnod
            iii=point_no_Dir(i)
            jjj=point_no_Dir(i+nnod)
            if (iii>0) then
              term1=I_i_app(xxx(i),yyy(i),zzz(i))*omega(i)*time_term
              effe(iii)=effe(iii)+term1
            endif
            if (jjj>0) then
              term2=I_e_app(xxx(i),yyy(i),zzz(i))*omega(i)*time_term 
              effe(jjj)=effe(jjj)+term2
            endif
          enddo
780       format(t1,2(g12.6,3x),g12.6)

!         term due to Dirichlet boundary condition
          if (Dir_not_empty) then
            where (effe_Dir.ne.zero) effe=effe+effe_Dir*time_term
          endif

!         term due to Neumann boundary condition
          if (Neu_not_empty) then
            where (effe_Neu.ne.zero) effe=effe+effe_Neu*time_term
          endif

          call CPU_TIME(end1)
          temp1=temp1+end1-beg1

          call CPU_TIME(beg2)
          call solve_with_CG_CSR(myid,master,nrig,nnz_tot,row_ptr,col_idx,val, &
                                 uuu,effe,write_it_hist,niter,tempo1,tempo2)
          call CPU_TIME(end2)
          temp2=temp2+end2-beg2

!         extension to u_i(nnod) and u_e(nnod)
          do i=1,nnod
            iii=point_no_Dir(i)
            if (iii>0) then
              u_i(i)=uuu(iii)
            else
              u_i(i)=ggg_i(xxx(i),yyy(i),zzz(i))*hhh(ttt)
            endif
            jjj=point_no_Dir(i+nnod)
            if (jjj>0) then
              u_e(i)=uuu(jjj)
            else
              u_e(i)=ggg_e(xxx(i),yyy(i),zzz(i))*hhh(ttt)
            endif
            vvv(i)=u_i(i)-u_e(i)
          enddo

          if (nrec>0) then
            do irec=1,nrec
              write(100+irec,402) ttt_PDE, vvv(index_rec(irec))
            enddo
          endif
402       format(t1,g12.6,2x,g12.6)

          if (write_vtk.or.write_xdmf) then
            if (nsnap>0) then
!             draw the computed solution in vtk or xdmf format 
              if (nts_PDE==it_snap(isnap)) then
                isnap=isnap+1
                llll=llll+1
                if (write_vtk) then
                 solpar='V_xxx.vtk'
                 if (llll<10) then
                   write(solpar(3:4),'(a2)')'00'
                   write(solpar(5:5),'(i1)')llll
                 else if (llll<99) then
                   write(solpar(3:3),'(a1)')'0'
                   write(solpar(4:5),'(i2)')llll
                 else if (llll<999) then
                   write(solpar(3:5),'(i3)')llll
                 endif
                 write(titolo,*) "transmembrane potential at t = ",ttt_PDE,' ms'
                 call clean_multiple_blank_string(512,titolo)
                 inquire (file=solpar(1:len_trim(solpar)), exist=exists)
                 if (exists) call system ('rm '//solpar(1:len_trim(solpar)))
                 call draw_vtk(solpar,titolo,nn,nnod,nelem,xxx,yyy,zzz,con_spx,&
                               vvv)
                 exists=.false.
                 inquire (file=solpar(1:len_trim(solpar))//'.gz', exist=exists)
                 if (exists)                                                   &
                   call system ('rm '//solpar(1:len_trim(solpar))//'.gz')
                 call system('gzip '//solpar(1:len_trim(solpar)))
                else if (write_xdmf) then
!                tsnap=dfloat(nts_PDE)*deltat_PDE
!                call XDMFWRITE("V_t=", llll, tsnap, nnod, nelem, XYZ, con, vvv)
                endif
              endif
            endif
          endif
        
        endif
!       PDE's solution has been updated

        call CPU_TIME(time_PDE_end)
        tempo_PDE=tempo_PDE+time_PDE_end-time_PDE_begin

!       time upgrade
        ttt0=ttt

        if (ic<4.and.its==nts_frac(ic)) then
          write(*,'(i2,"% - ",$)') 25*ic
          ic=ic+1
        endif

      enddo            ! end of the time-loop

      if (myid==master) close(33)

      niter_ave=dfloat(niter)/dfloat(nts_PDE)

231   format(t1,'time-step: ',i8,5x,'t = ',g9.3,' s')
      if (write_it_hist.and.myid==master) close(23)
831   format(t1,i10,2x,4(2x,g12.6))
      write(*,'("100%            done")') 
!     write(*,*)

!     if (tempo_PDE.le.999999.d0) then
!       write(*,52) idint(tempo_PDE)
!     else
!       write(*,*) "execution-time for solving the PDEs: ",idint(tempo_PDE)," s"
!     endif
!     write(*,*) 'execution times (s) for building RHS and solving Ax=b: ',    &
!                temp1,temp2
52    format (t1,"CPU-time for solving the PDEs: ",i6," s")

      return

      end subroutine time_loop_CSR


!     --------------------------------------------------------------------------


      real(8) function c_m_per_vol(xxx,yyy,zzz)
                       
!     membrane capacity per unit volume = Chi * C_m
!     it is expressed in mF/cm^3
!     Chi and C_m are read from the file "membrane.input"

      implicit none

      real(8) :: xxx,yyy,zzz
      real(8) :: Chi, C_m, D_i, D_e
      common / membrane / Chi, C_m, D_i, D_e

      c_m_per_vol=Chi*C_m

      return

      end function c_m_per_vol


!     --------------------------------------------------------------------------


      subroutine solve_with_CG_CSR(myid,master,nrig,nnz,row_ptr,col_idx,       &
                                   val_CSR,xxx,bbb,write_it_hist,niter,        &
                                   tempo1,tempo2)

      implicit none

      integer(4) :: i,j,k,iter,myid,master
      integer(4) :: nrig, nnz, nmax, iprec, niter
      integer(4) :: row_ptr(0:nrig), col_idx(0:nnz-1)
      real(8) :: toll, err, bnrm2, alpha, beta, rho, zero=0.d0, uno=1.d0
      real(8) :: beg1,end1,tempo1,beg2,end2,tempo2, val_CSR(0:nnz-1)
      real(8), dimension(:), allocatable :: erre,zeta,ppp,qqq,prec
      real(8), dimension(nrig) :: xxx, bbb
      character :: riga512*512
      logical :: write_it_hist

      allocate(erre(nrig),zeta(nrig),ppp(nrig),qqq(nrig),prec(nrig))
      zeta=0.d0
      iter=0

      open(39,file='lin_solv.input')
      read(39,*)             ! algebraic_solver
      k=0
      loop_0: do
        read(39,'(a)') riga512
        loop_1: do j=1,512
          if (riga512(j:j).ne.' ') exit loop_1
        enddo loop_1
        if (riga512(j:j).ne.'!') then
          backspace(39)
          exit loop_0
        endif
      enddo loop_0
      read(39,*) nmax
      read(39,*) toll
      read(39,*)             !    solver
      read(39,*) iprec
      close(39)

!     Diagonal preconditioner:
      if (iprec==0) then
        do i=1,nrig
          prec(i)=uno
        enddo
      else if (iprec==1) then

        do i=1,nrig
          if (row_ptr(i)>row_ptr(i-1)) then
            loop_cerca_el_diag: do j=row_ptr(i-1),row_ptr(i)-1
              if (col_idx(j)==i-1) exit loop_cerca_el_diag
            enddo loop_cerca_el_diag
            prec(i)=val_CSR(j)
          endif
        enddo
      endif

      tempo1=0.d0
      tempo2=0.d0


      call CPU_time(beg1)
      bnrm2=0.d0
      do i=1,nrig
        bnrm2=bnrm2+bbb(i)*bbb(i)
      enddo
      bnrm2=dsqrt(bnrm2)
      if (bnrm2==0.d0) stop "Error ! rhs=0 !"
      call CPU_time(end1)
      tempo1=tempo1+end1-beg1
    
      call CPU_time(beg2)
      do i=1,nrig
        erre(i)=bbb(i)
        if (row_ptr(i)>row_ptr(i-1)) then
          do j=row_ptr(i-1),row_ptr(i)-1
            erre(i)=erre(i)-val_CSR(j)*xxx(col_idx(j)+1)
          enddo
        endif
        zeta(i)=erre(i)/prec(i)
        ppp(i)=zeta(i)
        err=err+erre(i)*erre(i)
      enddo
      call CPU_time(end2)
      tempo2=tempo2+end2-beg2
      err=dsqrt(err)/bnrm2
      if (err<toll) then
        if (write_it_hist.and.myid==master)                                    &
            write(23,*)'iter. 0:         res = ',err
        goto 101
      endif

      do iter=1,nmax
        
        call CPU_time(beg2)
        qqq=zero
        do i=1,nrig
          if (row_ptr(i)>row_ptr(i-1)) then
            do j=row_ptr(i-1),row_ptr(i)-1
              qqq(i)=qqq(i)+val_CSR(j)*ppp(col_idx(j)+1)
            enddo
          endif
        enddo
        call CPU_time(end2)
        tempo2=tempo2+end2-beg2
        call CPU_time(beg1)
        rho=0.d0; alpha=0.d0
        do i=1,nrig
          alpha=alpha+ppp(i)*erre(i)
          rho=rho+ppp(i)*qqq(i)
        enddo
        alpha=alpha/rho
        do i=1,nrig
          xxx(i)=xxx(i)+alpha*ppp(i)
          erre(i)=erre(i)-alpha*qqq(i)
        enddo      
        do i=1,nrig
          zeta(i)=erre(i)/prec(i)
        enddo
        beta=0.d0
        do i=1,nrig
          beta=beta+qqq(i)*zeta(i)
        enddo
        beta=beta/rho
        do i=1,nrig              
          ppp(i)=zeta(i)-beta*ppp(i)
        enddo

        err=0.d0    
        do i=1,nrig
          err=err+erre(i)*erre(i)
        enddo
        call CPU_time(end1)
        tempo1=tempo1+end1-beg1
        err=dsqrt(err)/bnrm2
        if (write_it_hist.and.myid==master) write(23,*)'iter. ',iter,          &
                                            ':         res = ',err
        if (err<toll) then
          niter=niter+iter
          goto 101
        endif

!       if (mod(iter,10)==0) write(*,*)'iter. ',iter,':         res = ',err

      enddo

      if (err>toll) &
        write(*,*) "Warning! maximum number of iterations done without convergence!"

101   continue
      do i=1,nrig
        bbb(i)=xxx(i)
      enddo

      return

      end subroutine solve_with_CG_CSR


!     --------------------------------------------------------------------------


      subroutine clean_multiple_blank_string(n,stringa)

!     removes multiple blank spaces from string stringa

      implicit none
      integer(4) :: i, i0, ic, n
      character :: stringa*(n), serv*(n)

      serv=' '
      loop_find_beginning: do i=1,n
        if (stringa(i:i).ne.' ') exit loop_find_beginning
      enddo loop_find_beginning
      ic=1; i0=i
      serv(1:1)=stringa(i0:i0)
      if (len_trim(stringa)>i0) then
        do i=i0+1,len_trim(stringa)
          if (stringa(i:i).ne.' ') then
            ic=ic+1
            serv(ic:ic)=stringa(i:i)
          else
            if (stringa(i-1:i-1).ne.' ') then
              ic=ic+1
              serv(ic:ic)=stringa(i:i)
            endif
          endif
        enddo
      endif

      stringa=' '
      stringa(1:len_trim(serv))=serv(1:len_trim(serv))

      return

      end subroutine clean_multiple_blank_string


!     --------------------------------------------------------------------------


      subroutine make_spectral_connectivities(nnod_macro,nn,nelem,con,         &
                                              con_spx,nnod,ncol_max,err)

!     1) builds up spectral connectivities con_spx, s.t. con_spx(ie,k) is the
!        index of k-th spectral node of ie-th hexahedron in the global numbering
!     2) evaluates nnod, total number of spectral nodes (with NO duplicates)
!     3) evaluates ncol_max, the max bandwidth of the stiffness matrix

!     local numbering of MACRO nodes (read from the grid file) is shown below

!               k
!               .                        
!               .
!              5.________ 8
!              /.       /|
!             / .      / |
!           6/--------/7 |
!            | 1.....|....4.......j 
!            | .     |  /
!            |.      | /
!            |_______|/
!           2         3
!          . 
!         .
!        i
   
      implicit none

      integer(4) :: nnod_macro, nn, nn2, nn3, nelem, nnod, ncol_max, nelem_comm
      integer(4) :: con(8,nelem), con_spx(nn**3,nelem), maschera(nn**3), err
      integer(4) :: ie,ie_ot,ia,ib,iinc,ja,jb,jinc,ka,kb,kinc,neigh_comm(4,2)
      integer(4) :: ia_ot,ib_ot,iinc_ot,ja_ot,jb_ot,jinc_ot,ka_ot,kb_ot,kinc_ot
      integer(4) :: i, j, k, l, l_ot, iii, iii_ot, ic, bw_max, icont, icont0
      integer(4) :: int_arr_aux(1000), serv_edge(nn), serv_n3(nn*nn*nn)    
      integer(4), allocatable :: con_dual(:,:), nnz_max(:)
      character*3 :: faccia, faccia_ot

      err=0

      if (nn==2) then
        do ie=1,nelem
          con_spx(1,ie)=con(1,ie)
          con_spx(2,ie)=con(2,ie)
          con_spx(3,ie)=con(4,ie)
          con_spx(4,ie)=con(3,ie)
          con_spx(5,ie)=con(5,ie)
          con_spx(6,ie)=con(6,ie)
          con_spx(7,ie)=con(8,ie)
          con_spx(8,ie)=con(7,ie)
        enddo
        return
      endif 
      
      nn2=nn*nn; nn3=nn2*nn

!     dual connectivity is built
      bw_max=30
      allocate (con_dual(bw_max,nnod_macro)); con_dual=0
      do ie=1,nelem
        do i=1,8
          iii=con(i,ie)
          if (con_dual(bw_max,iii).ne.0)    &
              stop "Error in make_spectral_connectivities!"
          loop_001: do j=1,bw_max
            if (con_dual(j,iii)==0) exit loop_001
          enddo loop_001
          con_dual(j,iii)=ie
        enddo
      enddo

      nnod=0

      do ie=1,nelem

        maschera=0; int_arr_aux=0

        ic=0
        do i=1,8
          iii=con(i,ie)
          loop_002: do j=1,bw_max
            if (con_dual(j,iii)==0) exit loop_002
            ic=ic+1
            if (con_dual(j,iii)<ie) int_arr_aux(ic)=con_dual(j,iii)
          enddo loop_002
        enddo
        call mio_shell(1000,int_arr_aux)
!       if there are no neighbours with lower index, goto 101
!       and define the spectral gridpoints on hexa ie:

        if  (int_arr_aux(1000)==0) goto 101

        nelem_comm=1
        loop_trova_start: do i=1,1000
          if (int_arr_aux(i).ne.0) exit loop_trova_start
        enddo loop_trova_start
        icont0=i

        loop_003: do icont=icont0,1000
          if (icont<1000.and.int_arr_aux(icont)==int_arr_aux(icont+1)) then
            nelem_comm=nelem_comm+1
          else
            ie_ot=int_arr_aux(icont)

            neigh_comm=0
            iii=0
            loop_004: do l=1,8
              do l_ot=1,8
                if (con(l,ie)==con(l_ot,ie_ot)) then
                  iii=iii+1
                  neigh_comm(iii,1)=l; neigh_comm(iii,2)=l_ot
                  if (iii==nelem_comm) exit loop_004
                endif
              enddo
            enddo loop_004

            if (nelem_comm==1) then                ! vertex

              if (neigh_comm(1,1)==1) then
                iii=1
              else if (neigh_comm(1,1)==2) then
                iii=nn
              else if (neigh_comm(1,1)==3) then
                iii=nn2
              else if (neigh_comm(1,1)==4) then
                iii=nn2-nn+1
              else if (neigh_comm(1,1)==5) then
                iii=nn3-nn2+1
              else if (neigh_comm(1,1)==6) then
                iii=nn3-nn2+nn
              else if (neigh_comm(1,1)==7) then
                iii=nn3
              else if (neigh_comm(1,1)==8) then
                iii=nn3-nn+1
              endif

              if (neigh_comm(1,2)==1) then
                iii_ot=1
              else if (neigh_comm(1,2)==2) then
                iii_ot=nn
              else if (neigh_comm(1,2)==3) then
                iii_ot=nn2
              else if (neigh_comm(1,2)==4) then
                iii_ot=nn2-nn+1
              else if (neigh_comm(1,2)==5) then
                iii_ot=nn3-nn2+1
              else if (neigh_comm(1,2)==6) then
                iii_ot=nn3-nn2+nn
              else if (neigh_comm(1,2)==7) then
                iii_ot=nn3
              else if (neigh_comm(1,2)==8) then
                iii_ot=nn3-nn+1
              endif

              if (maschera(iii)==0) then
                con_spx(iii,ie)=con_spx(iii_ot,ie_ot)
                maschera(iii)=1
              endif

            else if (nelem_comm==2) then        ! edge

              call find_common_bound_edge(nelem_comm,ie,ie_ot,nn,neigh_comm,   &
                        ia,ib,iinc,ja,jb,jinc,ka,kb,kinc,ia_ot,ib_ot,iinc_ot,  &
                        ja_ot,jb_ot,jinc_ot,ka_ot,kb_ot,kinc_ot)

              iii=0; serv_edge=0
              do k=ka_ot,kb_ot,kinc_ot
                do j=ja_ot,jb_ot,jinc_ot
                  do i=ia_ot,ib_ot,iinc_ot
                    iii=iii+1
                    serv_edge(iii)=con_spx((k-1)*nn2+(j-1)*nn+i,ie_ot)
                  enddo
                enddo
              enddo
              iii=0
              do k=ka,kb,kinc
                do j=ja,jb,jinc
                  do i=ia,ib,iinc
                    ic=(k-1)*nn2+(j-1)*nn+i
                    iii=iii+1
                    if (maschera(ic)==0) then
                      con_spx(ic,ie)=serv_edge(iii)
                      maschera(ic)=1
                    endif
                  enddo
                enddo
              enddo
                  
            else if (nelem_comm==4) then        ! face

              call set_common_face(ie,ie_ot,nn,nelem,con,con_spx,neigh_comm,   &
                                   maschera)
   
            endif
            if (int_arr_aux(icont)==0) exit loop_003
            nelem_comm=1
          endif

        enddo loop_003

101     continue

        do k=1,nn
          do j=1,nn
            do i=1,nn
              iii=(k-1)*nn2+(j-1)*nn+i
              if (maschera(iii)==0) then
                nnod=nnod+1
                con_spx(iii,ie)=nnod
              endif
            enddo
          enddo
        enddo      
      enddo
 
      allocate (nnz_max(nnod)); nnz_max=0
      do ie=1,nelem
        do k=1,nn
          do j=1,nn
            do i=1,nn
              iii=(k-1)*nn2+(j-1)*nn+i
              nnz_max(iii)=nnz_max(iii)+1
            enddo
          enddo
        enddo
      enddo
      nnz_max=3*nn*nnz_max
      ncol_max=0
      do i=1,nnod
        if (nnz_max(i)>ncol_max) ncol_max=nnz_max(i)
      enddo

!     check for consistency (no repeated vertices in hexahedron)
      do ie=1,nelem
        serv_n3=con_spx(:,ie)
        call mio_shell(nn3,serv_n3)
        do i=2,nn3
          if (serv_n3(i)==serv_n3(i-1)) then
            if (ie==0) write(*,*)
            write(*,*) 'Error in the connectivities of element ',ie,'!'
            write(87,*) 'Error in the connectivities of element ',ie,'!'
            err=ie
          endif
        enddo
      enddo
!     if (ie.ne.0) write(*,*)

      return 

      end subroutine make_spectral_connectivities


!     --------------------------------------------------------------------------


      subroutine find_common_bound_edge(dim,ie,ie_ot,nn,neigh_comm,            &
                        ia,ib,iinc,ja,jb,jinc,ka,kb,kinc,ia_ot,ib_ot,iinc_ot,  &
                        ja_ot,jb_ot,jinc_ot,ka_ot,kb_ot,kinc_ot)

!     finds index range [ia,ib],[ja,jb],[ka,kb] for hexa ie and
!     index range [ia_ot,ib_ot],[ja_ot,jb_ot],[ka_ot,kb_ot] for hexa ie_ot
!     such ranges define common edge between ie and ie_ot
!     inc=+1 ---> range [-1,+1], inc=-1 ---> range [+1,-1]

      implicit none

      integer(4) :: dim,ie,ie_ot,nn,neigh_comm(4,2),ia,ib,ja,jb,ka,kb
      integer(4) :: ia_ot,ib_ot,ja_ot,jb_ot,ka_ot,kb_ot
      integer(4) :: iinc,jinc,kinc,iinc_ot,jinc_ot,kinc_ot

      if (dim==2) then            !     edges:    

!        hexa ie
         iinc=1; jinc=1; kinc=1
         if (neigh_comm(1,1)==1) then
           if (neigh_comm(2,1)==2) then
             ia=1; ib=nn; ja=1; jb=1; ka=1; kb=1
           else if (neigh_comm(2,1)==4) then
             ia=1; ib=1; ja=1; jb=nn; ka=1; kb=1
           else if (neigh_comm(2,1)==5) then
             ia=1; ib=1; ja=1; jb=1; ka=1; kb=nn
           endif
         else if (neigh_comm(1,1)==2) then
           if (neigh_comm(2,1)==3) then
             ia=nn; ib=nn; ja=1; jb=nn; ka=1; kb=1
           else if (neigh_comm(2,1)==6) then
             ia=nn; ib=nn; ja=1; jb=1; ka=1; kb=nn
           endif
         else if (neigh_comm(1,1)==3) then
           if (neigh_comm(2,1)==4) then
             ia=nn; ib=1; ja=nn; jb=nn; ka=1; kb=1; iinc=-1
           else if (neigh_comm(2,1)==7) then
             ia=nn; ib=nn; ja=nn; jb=nn; ka=1; kb=nn
           endif
         else if (neigh_comm(1,1)==4) then
           if (neigh_comm(2,1)==8) then
             ia=1; ib=1; ja=nn; jb=nn; ka=1; kb=nn
           endif
         else if (neigh_comm(1,1)==5) then
           if (neigh_comm(2,1)==6) then
             ia=1; ib=nn; ja=1; jb=1; ka=nn; kb=nn
           else if (neigh_comm(2,1)==8) then
             ia=1; ib=1; ja=1; jb=nn; ka=nn; kb=nn
           endif
         else if (neigh_comm(1,1)==6) then
           if (neigh_comm(2,1)==7) then
             ia=nn; ib=nn; ja=1; jb=nn; ka=nn; kb=nn
           endif
         else if (neigh_comm(1,1)==7) then
           if (neigh_comm(2,1)==8) then
             ia=nn; ib=1; ja=nn; jb=nn; ka=nn; kb=nn; iinc=-1
           endif
         endif

!        hexa ie_ot
         iinc_ot=1; jinc_ot=1; kinc_ot=1
         if (neigh_comm(1,2)==1) then
           if (neigh_comm(2,2)==2) then
             ia_ot=1; ib_ot=nn; ja_ot=1; jb_ot=1; ka_ot=1; kb_ot=1
           else if (neigh_comm(2,2)==4) then
             ia_ot=1; ib_ot=1; ja_ot=1; jb_ot=nn; ka_ot=1; kb_ot=1
           else if (neigh_comm(2,2)==5) then
             ia_ot=1; ib_ot=1; ja_ot=1; jb_ot=1; ka_ot=1; kb_ot=nn
           endif
         else if (neigh_comm(1,2)==2) then
           if (neigh_comm(2,2)==1) then
             ia_ot=nn; ib_ot=1; ja_ot=1; jb_ot=1; ka_ot=1; kb_ot=1; iinc_ot=-1
           else if (neigh_comm(2,2)==3) then
             ia_ot=nn; ib_ot=nn; ja_ot=1; jb_ot=nn; ka_ot=1; kb_ot=1
           else if (neigh_comm(2,2)==6) then
             ia_ot=nn; ib_ot=nn; ja_ot=1; jb_ot=1; ka_ot=1; kb_ot=nn
           endif
         else if (neigh_comm(1,2)==3) then
           if (neigh_comm(2,2)==2) then
             ia_ot=nn; ib_ot=nn; ja_ot=nn; jb_ot=1; ka_ot=1; kb_ot=1; jinc_ot=-1
           else if (neigh_comm(2,2)==4) then
             ia_ot=nn; ib_ot=1; ja_ot=nn; jb_ot=nn; ka_ot=1; kb_ot=1; iinc_ot=-1
           else if (neigh_comm(2,2)==7) then
             ia_ot=nn; ib_ot=nn; ja_ot=nn; jb_ot=nn; ka_ot=1; kb_ot=nn
           endif
         else if (neigh_comm(1,2)==4) then
           if (neigh_comm(2,2)==1) then
             ia_ot=1; ib_ot=1; ja_ot=nn; jb_ot=1; ka_ot=1; kb_ot=1; jinc_ot=-1
           else if (neigh_comm(2,2)==3) then
             ia_ot=1; ib_ot=nn; ja_ot=nn; jb_ot=nn; ka_ot=1; kb_ot=1
           else if (neigh_comm(2,2)==8) then
             ia_ot=1; ib_ot=1; ja_ot=nn; jb_ot=nn; ka_ot=1; kb_ot=nn
           endif
         else if (neigh_comm(1,2)==5) then
           if (neigh_comm(2,2)==1) then
             ia_ot=1; ib_ot=1; ja_ot=1; jb_ot=1; ka_ot=nn; kb_ot=1; kinc_ot=-1
           else if (neigh_comm(2,2)==6) then
             ia_ot=1; ib_ot=nn; ja_ot=1; jb_ot=1; ka_ot=nn; kb_ot=nn
           else if (neigh_comm(2,2)==8) then
             ia_ot=1; ib_ot=1; ja_ot=1; jb_ot=nn; ka_ot=nn; kb_ot=nn
           endif
         else if (neigh_comm(1,2)==6) then
           if (neigh_comm(2,2)==2) then
             ia_ot=nn; ib_ot=nn; ja_ot=1; jb_ot=1; ka_ot=nn; kb_ot=1; kinc_ot=-1
           else if (neigh_comm(2,2)==5) then
             ia_ot=nn; ib_ot=1; ja_ot=1; jb_ot=1; ka_ot=nn; kb_ot=nn; iinc_ot=-1
           else if (neigh_comm(2,2)==7) then
             ia_ot=nn; ib_ot=nn; ja_ot=1; jb_ot=nn; ka_ot=nn; kb_ot=nn
           endif
         else  if (neigh_comm(1,2)==7) then
           if (neigh_comm(2,2)==3) then
             ia_ot=nn;ib_ot=nn; ja_ot=nn;jb_ot=nn; ka_ot=nn;kb_ot=1; kinc_ot=-1
           else if (neigh_comm(2,2)==6) then
             ia_ot=nn;ib_ot=nn; ja_ot=nn;jb_ot=1; ka_ot=nn;kb_ot=nn; jinc_ot=-1
           else if (neigh_comm(2,2)==8) then
             ia_ot=nn;ib_ot=1; ja_ot=nn;jb_ot=nn; ka_ot=nn;kb_ot=nn; iinc_ot=-1
           endif
         else  if (neigh_comm(1,2)==8) then
           if (neigh_comm(2,2)==4) then
             ia_ot=1; ib_ot=1; ja_ot=nn; jb_ot=nn; ka_ot=nn; kb_ot=1; kinc_ot=-1
           else if (neigh_comm(2,2)==5) then
             ia_ot=1; ib_ot=1; ja_ot=nn; jb_ot=1; ka_ot=nn; kb_ot=nn; jinc_ot=-1
           else if (neigh_comm(2,2)==7) then
             ia_ot=1; ib_ot=nn; ja_ot=nn; jb_ot=nn; ka_ot=nn; kb_ot=nn
           endif
         endif

      else 

         stop "Error! Wrong value of dim in find_common_bound_edge"

      endif         

      return
       
      end subroutine find_common_bound_edge


!     --------------------------------------------------------------------------

      subroutine set_common_face(ie,ie_ot,nn,nelem,con,con_spx,neigh_comm,     &
                                 maschera)

!     sets global indices con_spx on common face between ie and ie_ot
!     finds index range [ia,ib],[ja,jb],[ka,kb] for hexa ie and
!     index range [ia_ot,ib_ot],[ja_ot,jb_ot],[ka_ot,kb_ot] for hexa ie_ot
!     such ranges define common face between ie and ie_ot

      implicit none

      integer(4) :: ie,ie_ot,nn,neigh_comm(4,2),ia,ib,ja,jb,ka,kb,iii,iii_ot
      integer(4) :: ia_ot,ib_ot,ja_ot,jb_ot,ka_ot,kb_ot,serv_4(4),face_macro(4)
      integer(4) :: nelem,con(8,nelem),con_spx(nn**3,nelem),nn2,serv_face(nn*nn)
      integer(4) :: i,j,k,ic,fa,fa_ot,maschera(nn**3),face_ot_macro(4),deg
      integer(4) :: pointer_ie(nn*nn),pointer_ie_ot(nn*nn)
      character  :: or_ie*3, or_ie_ot*3

      nn2=nn*nn; pointer_ie=0; pointer_ie_ot=0
!     face convention: bac=1; fro=2; lef=3; rig=4; bot=5; top=6

!     hexa ie ------------------------------------------------------------------
      serv_4=neigh_comm(:,1)
      call mio_shell(4,serv_4)
!     face convention: bac=1; fro=2; lef=3; rig=4; bot=5; top=6
      if (serv_4(1)==1.and.serv_4(2)==4.and.serv_4(3)==5.and.serv_4(4)==8) then
        face_macro(1)=1; face_macro(2)=4; face_macro(3)=8; face_macro(4)=5; fa=1
        ia=1; ib=1; ja=1; jb=nn; ka=1; kb=nn; or_ie="ant"
      else if (serv_4(1)==2.and.serv_4(2)==3.and.serv_4(3)==6.and.serv_4(4)==7)&
        then
        face_macro(1)=2; face_macro(2)=3; face_macro(3)=7; face_macro(4)=6; fa=2
        ia=nn; ib=nn; ja=1; jb=nn; ka=1; kb=nn; or_ie="ora"
      else if (serv_4(1)==1.and.serv_4(2)==2.and.serv_4(3)==5.and.serv_4(4)==6)&
        then
        face_macro(1)=1; face_macro(2)=2; face_macro(3)=6; face_macro(4)=5; fa=3
        ia=1; ib=nn; ja=1; jb=1; ka=1; kb=nn; or_ie="ora"
      else if (serv_4(1)==3.and.serv_4(2)==4.and.serv_4(3)==7.and.serv_4(4)==8)&
         then
         face_macro(1)=4; face_macro(2)=3; face_macro(3)=7; face_macro(4)=8
         fa=4
         ia=1; ib=nn; ja=nn; jb=nn; ka=1; kb=nn; or_ie="ant"
      else if (serv_4(1)==1.and.serv_4(2)==2.and.serv_4(3)==3.and.serv_4(4)==4)&
         then
         face_macro(1)=1; face_macro(2)=2; face_macro(3)=3; face_macro(4)=4
         fa=5
         ia=1; ib=nn; ja=1; jb=nn; ka=1; kb=1; or_ie="ant"
      else if (serv_4(1)==5.and.serv_4(2)==6.and.serv_4(3)==7.and.serv_4(4)==8)&
        then  
        face_macro(1)=5; face_macro(2)=6; face_macro(3)=7; face_macro(4)=8; fa=6
        ia=1; ib=nn; ja=1; jb=nn; ka=nn; kb=nn; or_ie="ora" 
      else
         stop "Error in node numbering (subroutine set_common_face I)!"
      endif
      ic=0
      do k=ka,kb
        do j=ja,jb
          do i=ia,ib
            iii=(k-1)*nn2+(j-1)*nn+i
            ic=ic+1
            pointer_ie(ic)=iii
          enddo
        enddo
      enddo
!     --------------------------------------------------------------------------

!     hexa ie_ot ---------------------------------------------------------------
!     face convention: bac=1; fro=2; lef=3; rig=4; bot=5; top=6
      do i=1,4
        loop_allinea_macro: do j=1,4
          if (face_macro(i)==neigh_comm(j,1)) then
            face_ot_macro(i)=neigh_comm(j,2)
            exit loop_allinea_macro
          endif
        enddo loop_allinea_macro
      enddo
      serv_4=neigh_comm(:,2)
      call mio_shell(4,serv_4)
      if (serv_4(1)==1.and.serv_4(2)==4.and.serv_4(3)==5.and.serv_4(4)==8) then
        fa_ot=1
      else if (serv_4(1)==2.and.serv_4(2)==3.and.serv_4(3)==6.and.serv_4(4)==7)&
        then
        fa_ot=2
      else if (serv_4(1)==1.and.serv_4(2)==2.and.serv_4(3)==5.and.serv_4(4)==6)&
        then
        fa_ot=3
      else if (serv_4(1)==3.and.serv_4(2)==4.and.serv_4(3)==7.and.serv_4(4)==8)&
        then
        fa_ot=4
      else if (serv_4(1)==1.and.serv_4(2)==2.and.serv_4(3)==3.and.serv_4(4)==4)&
        then
        fa_ot=5
      else if (serv_4(1)==5.and.serv_4(2)==6.and.serv_4(3)==7.and.serv_4(4)==8)&
        then  
        fa_ot=6  
      else
        stop "Error in node numbering (subroutine set_common_face II)!"
      endif
      call make_pointer_ie_ot(nn,fa_ot,face_ot_macro,pointer_ie_ot)

      do ic=1,nn2
        iii=pointer_ie(ic)
        iii_ot=pointer_ie_ot(ic)
        if (maschera(iii)==0) then
          con_spx(iii,ie)=con_spx(iii_ot,ie_ot)
          maschera(iii)=1
        endif
      enddo

      return
       
      end subroutine set_common_face


!     --------------------------------------------------------------------------


      subroutine make_spectral_grid(nn,nnod,nelem,nnod_macro,x_macro,y_macro,  &
                                    z_macro,con,con_spx,x_spx,y_spx,z_spx)

!     generates the spectral grid x_spx(nnod),y_spx(nnod),z_spx(nnod)
!     local numbering is as follows: iii=(k-1)*nn2+(j-1)*nn+i
!     where i,j,k run over the x,y,z direction, respectively
!     the Jacobian jac(nelem) for each hexa is also computed

      implicit none

      integer(4) :: ie, i, j, k, ic, iii, nn, nn2
      integer(4) :: nnod,nelem,nnod_macro,con(8,nelem),con_spx(nn**3,nelem)
      real(8) :: x_macro(nnod_macro),y_macro(nnod_macro),z_macro(nnod_macro)
      real(8) :: x_spx(nnod),y_spx(nnod),z_spx(nnod),xcor,ycor,zcor
      real(8) :: ax,bx,cx,dx,ex,fx,gx,hx,ay,by,cy,dy,ey,fy,gy,hy
      real(8) :: az,bz,cz,dz,ez,fz,gz,hz,jac(nelem)
      real(8) :: alf1,alf2,bet1,bet2,gamm1,gamm2,j_check
      real(8) :: xxx,yyy,zzz,ct(nn),ww(nn),dd(nn,nn)
      real(8), dimension (nelem) :: alfa1,beta1,gamma1,delta1
      real(8), dimension (nelem) :: alfa2,beta2,gamma2,delta2
      real(8), dimension (nelem) :: alfa3,beta3,gamma3,delta3
      real(8), dimension (nelem) :: lambda1,mu1,nu1,sigma1
      real(8), dimension (nelem) :: lambda2,mu2,nu2,sigma2
      real(8), dimension (nelem) :: lambda3,mu3,nu3,sigma3

      if (nn==2) then
        do ie=1,nelem
          i=con(1,ie); iii=con_spx(1,ie)
          x_spx(iii)=x_macro(i); y_spx(iii)=y_macro(i); z_spx(iii)=z_macro(i)
          i=con(2,ie); iii=con_spx(2,ie)
          x_spx(iii)=x_macro(i); y_spx(iii)=y_macro(i); z_spx(iii)=z_macro(i)
          i=con(4,ie); iii=con_spx(3,ie)
          x_spx(iii)=x_macro(i); y_spx(iii)=y_macro(i); z_spx(iii)=z_macro(i)
          i=con(3,ie); iii=con_spx(4,ie)
          x_spx(iii)=x_macro(i); y_spx(iii)=y_macro(i); z_spx(iii)=z_macro(i)
          i=con(5,ie); iii=con_spx(5,ie)
          x_spx(iii)=x_macro(i); y_spx(iii)=y_macro(i); z_spx(iii)=z_macro(i)
          i=con(6,ie); iii=con_spx(6,ie)
          x_spx(iii)=x_macro(i); y_spx(iii)=y_macro(i); z_spx(iii)=z_macro(i)
          i=con(8,ie); iii=con_spx(7,ie)
          x_spx(iii)=x_macro(i); y_spx(iii)=y_macro(i); z_spx(iii)=z_macro(i)
          i=con(7,ie); iii=con_spx(8,ie)
          x_spx(iii)=x_macro(i); y_spx(iii)=y_macro(i); z_spx(iii)=z_macro(i)
        enddo
        return
      endif 

      nn2=nn*nn
      ct=0.d0;ww=0.d0;dd=0.d0
      x_spx=0.d0; y_spx=0.d0; z_spx=0.d0
      call lgl(nn,ct,ww,dd)

      ElemLoop1: do ie=1,nelem

        ax=x_macro(con(1,ie)); ay=y_macro(con(1,ie))
        az=z_macro(con(1,ie))
        bx=x_macro(con(2,ie)); by=y_macro(con(2,ie))
        bz=z_macro(con(2,ie))
        cx=x_macro(con(3,ie)); cy=y_macro(con(3,ie))
        cz=z_macro(con(3,ie))
        dx=x_macro(con(4,ie)); dy=y_macro(con(4,ie))
        dz=z_macro(con(4,ie))
        ex=x_macro(con(5,ie)); ey=y_macro(con(5,ie))
        ez=z_macro(con(5,ie))
        fx=x_macro(con(6,ie)); fy=y_macro(con(6,ie))
        fz=z_macro(con(6,ie))
        gx=x_macro(con(7,ie)); gy=y_macro(con(7,ie))
        gz=z_macro(con(7,ie))
        hx=x_macro(con(8,ie)); hy=y_macro(con(8,ie))
        hz=z_macro(con(8,ie))
          
        !  mapping
         
        alfa1(ie)=(-ax+bx+cx-dx-ex+fx+gx-hx)/8.d0
        beta1(ie)=(-ax-bx+cx+dx-ex-fx+gx+hx)/8.d0
        gamma1(ie)=(-ax-bx-cx-dx+ex+fx+gx+hx)/8.d0
        lambda1(ie)=(ax-bx+cx-dx+ex-fx+gx-hx)/8.d0
        mu1(ie)=(+ax-bx-cx+dx-ex+fx+gx-hx)/8.d0
        nu1(ie)=(+ax+bx-cx-dx-ex-fx+gx+hx)/8.d0
        sigma1(ie)=(-ax+bx-cx+dx+ex-fx+gx-hx)/8.d0
        delta1(ie)=(ax+bx+cx+dx+ex+fx+gx+hx)/8.d0
         
        alfa2(ie)=(-ay+by+cy-dy-ey+fy+gy-hy)/8.d0
        beta2(ie)=(-ay-by+cy+dy-ey-fy+gy+hy)/8.d0
        gamma2(ie)=(-ay-by-cy-dy+ey+fy+gy+hy)/8.d0
        lambda2(ie)=(ay-by+cy-dy+ey-fy+gy-hy)/8.d0
        mu2(ie)=(+ay-by-cy+dy-ey+fy+gy-hy)/8.d0
        nu2(ie)=(+ay+by-cy-dy-ey-fy+gy+hy)/8.d0
        sigma2(ie)=(-ay+by-cy+dy+ey-fy+gy-hy)/8.d0
        delta2(ie)=(ay+by+cy+dy+ey+fy+gy+hy)/8.d0
            
        alfa3(ie)=(-az+bz+cz-dz-ez+fz+gz-hz)/8.d0
        beta3(ie)=(-az-bz+cz+dz-ez-fz+gz+hz)/8.d0
        gamma3(ie)=(-az-bz-cz-dz+ez+fz+gz+hz)/8.d0
        lambda3(ie)=(az-bz+cz-dz+ez-fz+gz-hz)/8.d0
        mu3(ie)=(+az-bz-cz+dz-ez+fz+gz-hz)/8.d0
        nu3(ie)=(+az+bz-cz-dz-ez-fz+gz+hz)/8.d0
        sigma3(ie)=(-az+bz-cz+dz+ez-fz+gz-hz)/8.d0
        delta3(ie)=(az+bz+cz+dz+ez+fz+gz+hz)/8.d0

       ! Jacobian in the node (x,y,z)=(+1,+1,+1)

        jac(ie)=alfa1(ie)*(beta2(ie)*gamma3(ie)-gamma2(ie)*beta3(ie))          &
               -alfa2(ie)*(beta1(ie)*gamma3(ie)-gamma1(ie)*beta3(ie))          &
               +alfa3(ie)*(beta1(ie)*gamma2(ie)-gamma1(ie)*beta2(ie))
  
        if (jac(ie).lt.0.d0) then
           write(10,*)'Error ! Orientation of elements is non-conforming !'
           write(10,*)'(element ',ie,' is clockwise oriented)'
           stop
        else if  (jac(ie)==0.d0) then
           write(10,*)'Error ! Orientation of elements is non-conforming !'
           write(10,*)'(element ',ie,' degenerates at a point)'
           stop
        endif

      enddo ElemLoop1

      do ie=1,nelem
        do k=1,nn  
          do j=1,nn
            do i=1,nn
              ic=(k-1)*nn2+(j-1)*nn+i
              iii=con_spx(ic,ie)
              if (x_spx(iii)==0.d0.and.y_spx(iii)==0.d0.and.z_spx(iii)==0.d0)  &
                then
                x_spx(iii)=alfa1(ie)*ct(i)+beta1(ie)*ct(j)+gamma1(ie)*ct(k) +  &
                           lambda1(ie)*ct(i)*ct(j)+mu1(ie)*ct(i)*ct(k) +       &
                           nu1(ie)*ct(j)*ct(k)+sigma1(ie)*ct(i)*ct(j)*ct(k) +  &
                           delta1(ie)
                y_spx(iii)=alfa2(ie)*ct(i)+beta2(ie)*ct(j)+gamma2(ie)*ct(k) +  &
                           lambda2(ie)*ct(i)*ct(j)+mu2(ie)*ct(i)*ct(k) +       &
                           nu2(ie)*ct(j)*ct(k)+sigma2(ie)*ct(i)*ct(j)*ct(k) +  &
                           delta2(ie)
                z_spx(iii)=alfa3(ie)*ct(i)+ beta3(ie)*ct(j)+gamma3(ie)*ct(k) + &
                           lambda3(ie)*ct(i)*ct(j)+mu3(ie)*ct(i)*ct(k) +       &
                           nu3(ie)*ct(j)*ct(k)+sigma3(ie)*ct(i)*ct(j)*ct(k) +  &
                           delta3(ie)
              endif
            enddo
          enddo
        enddo
      enddo

      return

      end subroutine make_spectral_grid


!     --------------------------------------------------------------------------


      subroutine legendref (p2,p2der,p1,p1der,n,x)

!     COMPUTES THE LEGENDRE POLYNOMIAL (p2) AND ITS DERIVATIVE (p2der)
!     OF DEGREE n AT x AND
!     THE LEGENDRE POLYNOMIAL (p1) AND ITS DERIVATIVE (p1der)
!     OF DEGREE n-1 AT x                                              
                                                                     
      implicit real*8 (a-h,o-z)                            

      p2=1.d0                                                    
      p2der=0.d0                                                    
      if (n.eq.0) return                                          
      p1=p2
      p2=x
      p1der=p2der
      p2der=1.d0
      if (n.eq.1) return                               
      do k=1,n-1
        p0=p1
        p1=p2
        p0der=p1der
        p1der=p2der
        dk=dfloat(k)                                         
        a1=(2.d0*dk+1.d0)/(dk+1.d0)
        a2=-dk/(dk+1.d0)
        p2=a1*p1*x+a2*p0
        p2der=a1*p1+a1*p1der*x+a2*p0der
      enddo
      return                    
      end                       


!     --------------------------------------------------------------------------


      subroutine lgl(np,ct,ww,dd)

      implicit real*8 (a-h,o-z)

      parameter (nstep=1000,acc=1.d-15)

      dimension ct(*),ww(*),dd(np,*)

      n=np-1
      ct(1)=-1.d0
      ct(np)=1.d0
      n2=idint(0.5d0*np)
      dx=2.d0/dfloat(nstep)
      x=-1.d0
      iroot=1
      call legendref (p2,a1,p1,p1der,n,x)
      do while (iroot.lt.n2)
        x=x+dx
        call legendref (p2,a2,p1,p1der,n,x)
        if (dabs(a2).le.acc) then
          iroot=iroot+1
          ct(iroot)=a2
        else
          aa=a1*a2
          if (aa.lt.0.d0) then
            iroot=iroot+1
            ct(iroot)=rtbis(x-dx,x,acc,n)
          endif
        endif
        a1=a2
      enddo

      n_filt=2*n2
      if (n_filt.ne.np) then         ! np odd
        ct(n2+1)=0.d0
        do i=1,n2-1
          ct(np-i)=-ct(i+1)
        enddo
      else                           ! np even
        do i=1,n2-1
          ct(np-i)=-ct(i+1)
        enddo
      endif

      xn=dfloat(n)
      acost=2.d0/(xn*(xn+1.d0))

      do j=1,np
        call legendref (p2,p2der,p1,p1der,n,ct(j))
        den=p2*p2
        ww(j)=acost/den
        do i=1,np
           if (i.ne.j) then
              call legendref (pnum,p2der,p1,p1der,n,ct(i))
              den=p2*(ct(i)-ct(j))
              dd(i,j)=pnum/den
           endif
        enddo
      enddo
      do j=2,np-1
        dd(j,j)=0.d0
      enddo
      dd(1,1)=-0.25d0*xn*(xn+1.d0)
      dd(np,np)=0.25d0*xn*(xn+1.d0)

      return

      end


!     --------------------------------------------------------------------------


      real*8 function rtbis(x1,x2,xacc,n)

      implicit real*8(a-h,o-z)

      parameter (jmax=1000)

      call legendref(p2,fmid,p1,p1der,n,x2)
      call legendref(p2,f,p1,p1der,n,x1)
      if (f*fmid.ge.0.d0) then
        write(10,*) 'root non-bracketed !'
        stop
      endif
      if (f.lt.0.d0) then
        rtbis=x1
        dx=x2-x1
      else
        rtbis=x2
        dx=x1-x2
      endif

      do j=1,jmax
        dx=0.5d0*dx
        xmid=rtbis+dx
        call legendref(p2,fmid,p1,p1der,n,xmid)
        if (fmid.le.0.d0) rtbis=xmid
        if (dabs(dx).lt.xacc.and.dabs(fmid).le.xacc) then
          rtbis=xmid
          return
        endif
      enddo

      end


!     --------------------------------------------------------------------------


      subroutine make_mapping_parameters(nn,nnod,nelem,xxx,yyy,zzz,con,        &
                 alfa1,beta1,gamma1,alfa2,beta2,gamma2,alfa3,beta3,gamma3,     &
                 lambda1,mu1,nu1,sigma1,lambda2,mu2,nu2,sigma2,                &
                 lambda3,mu3,nu3,sigma3)

      integer(4) :: nn,nn2,nn3,nnod,nelem,con(nn**3,nelem)
      integer(4) :: ie,i1,i2,i3,i4,i5,i6,i7,i8
      real(8), dimension(nnod) :: xxx,yyy,zzz
      real(8) :: ax,bx,cx,dx,ex,fx,gx,hx,ay,by,cy,dy,ey,fy,gy,hy,              &
                 az,bz,cz,dz,ez,fz,gz,hz
      real(8), dimension(nelem) :: alfa1,beta1,gamma1,alfa2,beta2,gamma2,      &
                 alfa3,beta3,gamma3,lambda1,mu1,nu1,sigma1,lambda2,mu2,nu2,    &
                 sigma2,lambda3,mu3,nu3,sigma3
      real(8), dimension(nelem) :: delta1,delta2,delta3

      nn2=nn*nn; nn3=nn2*nn

      do ie=1,nelem

        i1=con(1,ie); i2=con(nn,ie); i3=con(nn2,ie)
        i4=con(nn2-nn+1,ie); i5=con(nn3-nn2+1,ie); i6=con(nn3-nn2+nn,ie)
        i7=con(nn3,ie); i8=con(nn3-nn+1,ie)

        ax=xxx(i1); bx=xxx(i2); cx=xxx(i3); dx=xxx(i4)
        ex=xxx(i5); fx=xxx(i6); gx=xxx(i7); hx=xxx(i8)
        ay=yyy(i1); by=yyy(i2); cy=yyy(i3); dy=yyy(i4)
        ey=yyy(i5); fy=yyy(i6); gy=yyy(i7); hy=yyy(i8)
        az=zzz(i1); bz=zzz(i2); cz=zzz(i3); dz=zzz(i4)
        ez=zzz(i5); fz=zzz(i6); gz=zzz(i7); hz=zzz(i8)

        !  mapping
         
        alfa1(ie)=(-ax+bx+cx-dx-ex+fx+gx-hx)/8.d0
        beta1(ie)=(-ax-bx+cx+dx-ex-fx+gx+hx)/8.d0
        gamma1(ie)=(-ax-bx-cx-dx+ex+fx+gx+hx)/8.d0
        lambda1(ie)=(ax-bx+cx-dx+ex-fx+gx-hx)/8.d0
        mu1(ie)=(ax-bx-cx+dx-ex+fx+gx-hx)/8.d0
        nu1(ie)=(ax+bx-cx-dx-ex-fx+gx+hx)/8.d0
        sigma1(ie)=(-ax+bx-cx+dx+ex-fx+gx-hx)/8.d0
         
        alfa2(ie)=(-ay+by+cy-dy-ey+fy+gy-hy)/8.d0
        beta2(ie)=(-ay-by+cy+dy-ey-fy+gy+hy)/8.d0
        gamma2(ie)=(-ay-by-cy-dy+ey+fy+gy+hy)/8.d0
        lambda2(ie)=(ay-by+cy-dy+ey-fy+gy-hy)/8.d0
        mu2(ie)=(ay-by-cy+dy-ey+fy+gy-hy)/8.d0
        nu2(ie)=(ay+by-cy-dy-ey-fy+gy+hy)/8.d0
        sigma2(ie)=(-ay+by-cy+dy+ey-fy+gy-hy)/8.d0
            
        alfa3(ie)=(-az+bz+cz-dz-ez+fz+gz-hz)/8.d0
        beta3(ie)=(-az-bz+cz+dz-ez-fz+gz+hz)/8.d0
        gamma3(ie)=(-az-bz-cz-dz+ez+fz+gz+hz)/8.d0
        lambda3(ie)=(az-bz+cz-dz+ez-fz+gz-hz)/8.d0
        mu3(ie)=(az-bz-cz+dz-ez+fz+gz-hz)/8.d0
        nu3(ie)=(az+bz-cz-dz-ez-fz+gz+hz)/8.d0
        sigma3(ie)=(-az+bz-cz+dz+ez-fz+gz-hz)/8.d0

      enddo

      return

      end subroutine make_mapping_parameters


!     --------------------------------------------------------------------------


      real(8) function ggg_i(xxx,yyy,zzz)

      implicit none
      
      real(8) :: xxx, yyy, zzz, amp_Dir_i, amp_Dir_e
      common / Dirichlet_stimulus / amp_Dir_i, amp_Dir_e

      ggg_i=amp_Dir_i

      return

      end function ggg_i


!     --------------------------------------------------------------------------


      real(8) function ggg_e(xxx,yyy,zzz)

      implicit none
      
      real(8) :: xxx, yyy, zzz, amp_Dir_i, amp_Dir_e
      common / Dirichlet_stimulus / amp_Dir_i, amp_Dir_e

      ggg_e=amp_Dir_e

      return

      end function ggg_e


!     --------------------------------------------------------------------------


      real(8) function hhh(ttt)

!     time-history is evaluated through parameters provided in "bidomain.input"

      implicit none
      
      real(8) :: ttt, stim_start, stim_end, stim_duration, stim_period
      common / stim_parameters / stim_start, stim_end, stim_duration,          &
                                 stim_period

      hhh=0.d0
      if (ttt.ge.stim_start.and.ttt.le.stim_end) then
        if (ttt-dfloat(idint((ttt-stim_start)/stim_period))*stim_period.le.    &
            stim_duration) hhh=1.d0
      endif

      return

      end function hhh


!     --------------------------------------------------------------------------


      subroutine make_omega2_enne_BC(nn,nnod,nelem,nBC,con_BC,x_spx,y_spx,     &
                                     z_spx,con_spx,omega2_BC,enne)

      implicit none

      integer(4) :: nn,nnod,nelem,nBC,iBC
      integer(4) :: con_BC(2,nBC), con_spx(nn**3,nelem)
      integer(4) ie,i1,j1,k1,fa,iii
      real(8) :: x_spx(nnod),y_spx(nnod),z_spx(nnod),omega2_BC(nn*nn,nBC)
      real(8) :: enne(3,nn*nn,nBC), den
      real(8), dimension(nelem) :: alfa1,beta1,gamma1,alfa2,beta2,gamma2,alfa3,&
               beta3,gamma3,lambda1,mu1,nu1,sigma1,lambda2,mu2,nu2,sigma2,     &
               lambda3,mu3,nu3,sigma3
      real(8), dimension(nn,nn) :: dexdex, dexdey, dexdez, deydex, deydey,     &
                                   deydez, dezdex, dezdey, dezdez
      real(8) :: ct(nn), ww(nn), dd(nn,nn), zero=0.d0, term, term0

      call make_mapping_parameters(nn,nnod,nelem,x_spx,y_spx,z_spx,con_spx,    &
           alfa1,beta1,gamma1,alfa2,beta2,gamma2,alfa3,beta3,gamma3,lambda1,   &
           mu1,nu1,sigma1,lambda2,mu2,nu2,sigma2,lambda3,mu3,nu3,sigma3)
      
      ct=zero; ww=zero; dd=zero
      call lgl(nn,ct,ww,dd)

      do iBC=1,nBC

        ie=con_BC(1,iBC); fa=con_BC(2,iBC)

        do k1=1,nn
          do i1=1,nn
            dexdey(i1,k1)=beta1(ie)+lambda1(ie)*ct(i1)+nu1(ie)*ct(k1)+         &
                          sigma1(ie)*ct(i1)*ct(k1)
            deydey(i1,k1)=beta2(ie)+lambda2(ie)*ct(i1)+nu2(ie)*ct(k1)+         &
                          sigma2(ie)*ct(i1)*ct(k1)
            dezdey(i1,k1)=beta3(ie)+lambda3(ie)*ct(i1)+nu3(ie)*ct(k1)+         &
                          sigma3(ie)*ct(i1)*ct(k1)
          enddo
          do j1=1,nn
            dexdex(j1,k1)=alfa1(ie)+lambda1(ie)*ct(j1)+mu1(ie)*ct(k1)+         &
                          sigma1(ie)*ct(j1)*ct(k1)
            deydex(j1,k1)=alfa2(ie)+lambda2(ie)*ct(j1)+mu2(ie)*ct(k1)+         &
                          sigma2(ie)*ct(j1)*ct(k1)
            dezdex(j1,k1)=alfa3(ie)+lambda3(ie)*ct(j1)+mu3(ie)*ct(k1)+         &
                          sigma3(ie)*ct(j1)*ct(k1)
          enddo
        enddo
        do j1=1,nn
          do i1=1,nn
            dexdez(i1,j1)=gamma1(ie)+mu1(ie)*ct(i1)+nu1(ie)*ct(j1)+            &
                          sigma1(ie)*ct(i1)*ct(j1)
            deydez(i1,j1)=gamma2(ie)+mu2(ie)*ct(i1)+nu2(ie)*ct(j1)+            &
                          sigma2(ie)*ct(i1)*ct(j1)
            dezdez(i1,j1)=gamma3(ie)+mu3(ie)*ct(i1)+nu3(ie)*ct(j1)+            &
                          sigma3(ie)*ct(i1)*ct(j1)
          enddo
        enddo

        if (fa==1) then
          i1=1
          iii=0
          do k1=1,nn
            do j1=1,nn
              iii=iii+1
              term0=deydey(i1,k1)*dezdez(i1,j1)-dezdey(i1,k1)*deydez(i1,j1)
              term=term0*term0
              term0=dezdey(i1,k1)*dexdez(i1,j1)-dexdey(i1,k1)*dezdez(i1,j1)
              term=term+term0*term0
              term0=dexdey(i1,k1)*deydez(i1,j1)-deydey(i1,k1)*dexdez(i1,j1)
              term=term+term0*term0
              omega2_BC(iii,iBC)=dsqrt(term)*ww(j1)*ww(k1)
              enne(1,iii,iBC)=-deydey(i1,k1)*dezdez(i1,j1)+dezdey(i1,k1)*      &
                              deydez(i1,j1)
              enne(2,iii,iBC)=-dezdey(i1,k1)*dexdez(i1,j1)+dexdey(i1,k1)*      &
                              dezdez(i1,j1)
              enne(3,iii,iBC)=-dexdey(i1,k1)*deydez(i1,j1)+deydey(i1,k1)*      &
                              dexdez(i1,j1)
            enddo
          enddo
        else if (fa==2) then
          i1=nn
          iii=0
          do k1=1,nn
            do j1=1,nn
              iii=iii+1
              term0=deydey(i1,k1)*dezdez(i1,j1)-dezdey(i1,k1)*deydez(i1,j1)
              term=term0*term0
              term0=dezdey(i1,k1)*dexdez(i1,j1)-dexdey(i1,k1)*dezdez(i1,j1)
              term=term+term0*term0
              term0=dexdey(i1,k1)*deydez(i1,j1)-deydey(i1,k1)*dexdez(i1,j1)
              term=term+term0*term0
              omega2_BC(iii,iBC)=dsqrt(term)*ww(j1)*ww(k1)
              enne(1,iii,iBC)=deydey(i1,k1)*dezdez(i1,j1)-dezdey(i1,k1)*       &
                              deydez(i1,j1)
              enne(2,iii,iBC)=dezdey(i1,k1)*dexdez(i1,j1)-dexdey(i1,k1)*       &
                              dezdez(i1,j1)
              enne(3,iii,iBC)=dexdey(i1,k1)*deydez(i1,j1)-deydey(i1,k1)*       &
                              dexdez(i1,j1)
            enddo
          enddo
        else if (fa==3) then
          j1=1
          iii=0
          do k1=1,nn
            do i1=1,nn
              iii=iii+1
              term0=deydex(j1,k1)*dezdez(i1,j1)-dezdex(j1,k1)*deydez(i1,j1)
              term=term0*term0
              term0=dezdex(j1,k1)*dexdez(i1,j1)-dexdex(j1,k1)*dezdez(i1,j1)
              term=term+term0*term0
              term0=dexdex(j1,k1)*deydez(i1,j1)-deydex(j1,k1)*dexdez(i1,j1)
              term=term+term0*term0
              omega2_BC(iii,iBC)=dsqrt(term)*ww(i1)*ww(k1)
              enne(1,iii,iBC)=deydex(j1,k1)*dezdez(i1,j1)-dezdex(j1,k1)*       &
                              deydez(i1,j1)
              enne(2,iii,iBC)=dezdex(j1,k1)*dexdez(i1,j1)-dexdex(j1,k1)*       &
                              dezdez(i1,j1)
              enne(3,iii,iBC)=dexdex(j1,k1)*deydez(i1,j1)-deydex(j1,k1)*       &
                              dexdez(i1,j1)
            enddo
          enddo
        else if (fa==4) then
          j1=nn
          iii=0
          do k1=1,nn
            do i1=1,nn
              iii=iii+1
              term0=deydex(j1,k1)*dezdez(i1,j1)-dezdex(j1,k1)*deydez(i1,j1)
              term=term0*term0
              term0=dezdex(j1,k1)*dexdez(i1,j1)-dexdex(j1,k1)*dezdez(i1,j1)
              term=term+term0*term0
              term0=dexdex(j1,k1)*deydez(i1,j1)-deydex(j1,k1)*dexdez(i1,j1)
              term=term+term0*term0
              omega2_BC(iii,iBC)=dsqrt(term)*ww(i1)*ww(k1)
              enne(1,iii,iBC)=-deydex(j1,k1)*dezdez(i1,j1)+dezdex(j1,k1)*      &
                              deydez(i1,j1)
              enne(2,iii,iBC)=-dezdex(j1,k1)*dexdez(i1,j1)+dexdex(j1,k1)*      &
                              dezdez(i1,j1)
              enne(3,iii,iBC)=-dexdex(j1,k1)*deydez(i1,j1)+deydex(j1,k1)*      &
                              dexdez(i1,j1)
            enddo
          enddo
        else if (fa==5) then
          k1=1
          iii=0
          do j1=1,nn
            do i1=1,nn
              iii=iii+1
              term0=deydex(j1,k1)*dezdey(i1,k1)-dezdex(j1,k1)*deydey(i1,k1)
              term=term0*term0
              term0=dezdex(j1,k1)*dexdey(i1,k1)-dexdex(j1,k1)*dezdey(i1,k1)
              term=term+term0*term0
              term0=dexdex(j1,k1)*deydey(i1,k1)-deydex(j1,k1)*dexdey(i1,k1)
              term=term+term0*term0
              omega2_BC(iii,iBC)=dsqrt(term)*ww(i1)*ww(j1)
              enne(1,iii,iBC)=-deydex(j1,k1)*dezdey(i1,k1)+dezdex(j1,k1)*      &
                              deydey(i1,k1)
              enne(2,iii,iBC)=-dezdex(j1,k1)*dexdey(i1,k1)+dexdex(j1,k1)*      &
                              dezdey(i1,k1)
              enne(3,iii,iBC)=-dexdex(j1,k1)*deydey(i1,k1)+deydex(j1,k1)*      &
                              dexdey(i1,k1)
            enddo
          enddo
        else if (fa==6) then
          k1=nn
          iii=0
          do j1=1,nn
            do i1=1,nn
              iii=iii+1
              term0=deydex(j1,k1)*dezdey(i1,k1)-dezdex(j1,k1)*deydey(i1,k1)
              term=term0*term0
              term0=dezdex(j1,k1)*dexdey(i1,k1)-dexdex(j1,k1)*dezdey(i1,k1)
              term=term+term0*term0
              term0=dexdex(j1,k1)*deydey(i1,k1)-deydex(j1,k1)*dexdey(i1,k1)
              term=term+term0*term0
              omega2_BC(iii,iBC)=dsqrt(term)*ww(i1)*ww(j1)
              enne(1,iii,iBC)=deydex(j1,k1)*dezdey(i1,k1)-dezdex(j1,k1)*       &
                              deydey(i1,k1)
              enne(2,iii,iBC)=dezdex(j1,k1)*dexdey(i1,k1)-dexdex(j1,k1)*       &
                              dezdey(i1,k1)
              enne(3,iii,iBC)=dexdex(j1,k1)*deydey(i1,k1)-deydex(j1,k1)*       &
                              dexdey(i1,k1)
            enddo
          enddo
        endif

      enddo        ! loop over B.C. faces

      do iBC=1,nBC
        do iii=1,nn*nn
          den=dsqrt(enne(1,iii,iBC)*enne(1,iii,iBC)+enne(2,iii,iBC)*           &
                    enne(2,iii,iBC)+enne(3,iii,iBC)*enne(3,iii,iBC))
          enne(1,iii,iBC)=enne(1,iii,iBC)/den
          enne(2,iii,iBC)=enne(2,iii,iBC)/den
          enne(3,iii,iBC)=enne(3,iii,iBC)/den
        enddo
      enddo
          
      return
      
      end subroutine make_omega2_enne_BC      


!     --------------------------------------------------------------------------


      subroutine evaluate_DOF_Dir(nn,nelem,nnod,ndir_i,ndir_e,con_dir_i,       &
                                  con_dir_e,con_spx,DOF_no_Dir_i,DOF_no_Dir_e, &
                                  point_no_Dir)

!     evaluate DOF_Dir, number of DOF lying on the Dirichlet boundary
!     and point_no_Dir, pointer to non-Dirichlet DOFs

      implicit none

      integer(4) :: nn,nn2,nnod,nelem,ndir_i,ndir_e,DOF_no_Dir_i,DOF_no_Dir_e
      integer(4) :: con_spx(nn**3,nelem), point_no_Dir(2*nnod)
      integer(4) :: con_dir_i(2,ndir_i), con_dir_e(2,ndir_e)
      integer(4) :: idir,iii,ie,fa,i1,j1,k1,ia,ib,ja,jb,ka,kb,DOF_Dir
     
      nn2=nn*nn
      do iii=1,2*nnod
        point_no_Dir(iii)=iii
      enddo

      if (ndir_i>0) then
        do idir=1,ndir_i
          ie=con_dir_i(1,idir); fa=con_dir_i(2,idir)

          ia=1; ib=nn; ja=1; jb=nn; ka=1; kb=nn
          if (fa==1) then
            ib=1
          else if (fa==2) then
            ia=nn
          else if (fa==3) then
            jb=1
          else if (fa==4) then
            ja=nn
          else if (fa==5) then
            kb=1
          else if (fa==6) then
            ka=nn
          endif

          do k1=ka,kb
            do j1=ja,jb
              do i1=ia,ib
                iii=con_spx((k1-1)*nn2+(j1-1)*nn+i1,ie)
                point_no_Dir(iii)=0
              enddo
            enddo
          enddo
        enddo
      endif
      if (ndir_e>0) then
        do idir=1,ndir_e
          ie=con_dir_e(1,idir); fa=con_dir_e(2,idir)

          ia=1; ib=nn; ja=1; jb=nn; ka=1; kb=nn
          if (fa==1) then
            ib=1
          else if (fa==2) then
            ia=nn
          else if (fa==3) then
            jb=1
          else if (fa==4) then
            ja=nn
          else if (fa==5) then
            kb=1
          else if (fa==6) then
            ka=nn
          endif

          do k1=ka,kb
            do j1=ja,jb
              do i1=ia,ib
                iii=con_spx((k1-1)*nn2+(j1-1)*nn+i1,ie)
                point_no_Dir(iii+nnod)=0
              enddo
            enddo
          enddo
        enddo
      endif

      DOF_no_Dir_i=0; DOF_no_Dir_e=0; DOF_Dir=0
      do iii=1,nnod
        if (point_no_Dir(iii)==0) then
          DOF_Dir=DOF_Dir+1
        else
          point_no_Dir(iii)=point_no_Dir(iii)-DOF_Dir
          DOF_no_Dir_i=DOF_no_Dir_i+1
        endif
      enddo
      do iii=nnod+1,2*nnod
        if (point_no_Dir(iii)==0) then
          DOF_Dir=DOF_Dir+1
        else
          point_no_Dir(iii)=point_no_Dir(iii)-DOF_Dir
          DOF_no_Dir_e=DOF_no_Dir_e+1
        endif
      enddo

      return

      end subroutine evaluate_DOF_Dir      


!     --------------------------------------------------------------------------


      subroutine make_pointer_ie_ot(nn,fa_ot,face_ot_macro,pointer_ot)

!     fills the pointer pointer_ie_ot

      implicit none

      integer(4) :: i,j,k,ic,iii
      integer(4) :: nn, nn2, fa_ot, face_ot_macro(4), pointer_ot(nn*nn*nn)

      nn2=nn*nn

      if (fa_ot==1) then
        i=1; ic=0
        if (face_ot_macro(1)==1) then
          if (face_ot_macro(2)==4) then
            do k=1,nn
              do j=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==5) then
            do j=1,nn 
              do k=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==4) then
          if (face_ot_macro(2)==8) then
            do j=nn,1,-1
              do k=1,nn 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==1) then
            do k=1,nn 
              do j=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==8) then
          if (face_ot_macro(2)==4) then
            do j=nn,1,-1
              do k=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==5) then
            do k=nn,1,-1 
              do j=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==5) then
          if (face_ot_macro(2)==8) then
            do k=nn,1,-1
              do j=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==1) then
            do j=1,nn
              do k=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        endif
      
      else if (fa_ot==2) then
        i=nn; ic=0
        if (face_ot_macro(1)==2) then
          if (face_ot_macro(2)==3) then
            do k=1,nn
              do j=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==6) then
            do j=1,nn 
              do k=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==3) then
          if (face_ot_macro(2)==7) then
            do j=nn,1,-1
              do k=1,nn 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==2) then
            do k=1,nn 
              do j=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==7) then
          if (face_ot_macro(2)==3) then
            do j=nn,1,-1
              do k=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==6) then
            do k=nn,1,-1 
              do j=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==6) then
          if (face_ot_macro(2)==7) then
            do k=nn,1,-1
              do j=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==2) then
            do j=1,nn
              do k=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        endif
      
      else if (fa_ot==3) then
        j=1; ic=0
        if (face_ot_macro(1)==1) then
          if (face_ot_macro(2)==2) then
            do k=1,nn
              do i=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==5) then
            do i=1,nn 
              do k=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==2) then
          if (face_ot_macro(2)==1) then
            do k=1,nn
              do i=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==6) then
            do i=nn,1,-1 
              do k=1,nn 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==6) then
          if (face_ot_macro(2)==2) then
            do i=nn,1,-1
              do k=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==5) then
            do k=nn,1,-1 
              do i=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==5) then
          if (face_ot_macro(2)==6) then
            do k=nn,1,-1
              do i=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==1) then
            do i=1,nn
              do k=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        endif

      else if (fa_ot==4) then
        j=nn; ic=0
        if (face_ot_macro(1)==4) then
          if (face_ot_macro(2)==3) then
            do k=1,nn
              do i=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==8) then
            do i=1,nn 
              do k=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==3) then
          if (face_ot_macro(2)==4) then
            do k=1,nn
              do i=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==7) then
            do i=nn,1,-1 
              do k=1,nn 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==7) then
          if (face_ot_macro(2)==3) then
            do i=nn,1,-1
              do k=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==8) then
            do k=nn,1,-1 
              do i=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==8) then
          if (face_ot_macro(2)==7) then
            do k=nn,1,-1
              do i=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==4) then
            do i=1,nn
              do k=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        endif

      else if (fa_ot==5) then
        k=1; ic=0
        if (face_ot_macro(1)==1) then
          if (face_ot_macro(2)==2) then
            do j=1,nn
              do i=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==4) then
            do i=1,nn 
              do j=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==2) then
          if (face_ot_macro(2)==1) then
            do j=1,nn
              do i=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==3) then
            do i=nn,1,-1 
              do j=1,nn 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==3) then
          if (face_ot_macro(2)==2) then
            do i=nn,1,-1
              do j=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==4) then
            do j=nn,1,-1 
              do i=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==4) then
          if (face_ot_macro(2)==3) then
            do j=nn,1,-1
              do i=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==1) then
            do i=1,nn
              do j=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        endif

      else if (fa_ot==6) then
        k=nn; ic=0
        if (face_ot_macro(1)==5) then
          if (face_ot_macro(2)==6) then
            do j=1,nn
              do i=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==8) then
            do i=1,nn 
              do j=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==6) then
          if (face_ot_macro(2)==5) then
            do j=1,nn
              do i=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==7) then
            do i=nn,1,-1 
              do j=1,nn 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==7) then
          if (face_ot_macro(2)==6) then
            do i=nn,1,-1
              do j=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==8) then
            do j=nn,1,-1 
              do i=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        else if (face_ot_macro(1)==8) then
          if (face_ot_macro(2)==7) then
            do j=nn,1,-1
              do i=1,nn
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          else if (face_ot_macro(2)==5) then
            do i=1,nn
              do j=nn,1,-1 
                iii=(k-1)*nn2+(j-1)*nn+i
                ic=ic+1
                pointer_ot(ic)=iii
              enddo
            enddo
          endif
        endif

      endif

      return

      end subroutine make_pointer_ie_ot      


!     --------------------------------------------------------------------------


      subroutine check_zero_in_stiffness_mat_CSR(nnz,val,nzero)

      implicit none

      integer(4) :: nnz, nzero, i
      real(8) :: val(0:nnz-1)

      nzero=0
      do i=0,nnz-1
        if (val(i)==0) nzero=nzero+1
      enddo

      return

      end subroutine check_zero_in_stiffness_mat_CSR      


!     --------------------------------------------------------------------------


      subroutine clean_file(fichier)

      implicit none

      integer(4) :: i, i_comm
      character :: fichier*(*)

      i_comm=index(fichier,'!')
      if (i_comm==0) then
        i_comm=len_trim(fichier)
      else
        i_comm=i_comm-1
      endif
      loop_1: do i=i_comm,1,-1
        if (fichier(i:i).ne.' '.and.fichier(i:i).ne.char(9)) exit loop_1
      enddo loop_1
      i_comm=i+1
      fichier(i_comm:len_trim(fichier))=' '

      return

      end


!     --------------------------------------------------------------------------


      real(8) function psi_i(xxx,yyy,zzz)

!     The Neumann BC Psi_i = n^T M_i grad u_i n on       G_N_i x (0,T)
!     is supposed to be in the form Psi_i(x,y,z,t)=psi_i(x,y,z)*hhh(t)

      implicit none
      real(8) :: xxx, yyy, zzz, amp_Neu_i, amp_Neu_e
      common / Neumann_stimulus / amp_Neu_i, amp_Neu_e

      psi_i=amp_Neu_i

      return

      end function psi_i


!     --------------------------------------------------------------------------


      real(8) function psi_e(xxx,yyy,zzz)

!     The Neumann BC Psi_e = n^T M_e grad u_e n       on G_N_e x (0,T)
!     is supposed to be in the form Psi_e(x,y,z,t)=psi_e(x,y,z)*hhh(t)

      implicit none
      real(8) :: xxx, yyy, zzz, amp_Neu_i, amp_Neu_e
      common / Neumann_stimulus / amp_Neu_i, amp_Neu_e

      psi_e=amp_Neu_e

      return

      end function psi_e


!     --------------------------------------------------------------------------


      subroutine make_effe_Neu(nn,nnod,nrig,nelem,con_spx,nneu_i,nneu_e,       &
                               con_neu_i,con_neu_e,xxx,yyy,zzz,point_no_Dir,   &
                               effe_Neu)

      implicit none

      integer(4) :: nnod, nelem, nneu_i, nneu_e, nrig, point_no_Dir(2*nnod)
      integer(4) :: con_neu_i(2,nneu_i), con_neu_e(2,nneu_e), nn, nn2, nn3
      integer(4) :: con_spx(nn*nn*nn,nelem), ineu, i, j, iii, jjj, ie, fa
      integer(4) :: a1, a2, b1, b2, c1, c2, i1, j1, k1
      real(8) :: xxx(nnod), yyy(nnod), zzz(nnod), effe_Neu(nrig)
      real(8) :: omega2_neu_i(nn*nn,nneu_i), omega2_neu_e(nn*nn,nneu_e)
      real(8) :: enne_i(3,nn*nn,nneu_i), enne_e(3,nn*nn,nneu_e)
      real(8), external :: psi_i, psi_e

      nn2=nn*nn; nn3=nn2*nn
      effe_Neu=0.d0

      if (nneu_i>0) then
        call make_omega2_enne_BC(nn,nnod,nelem,nneu_i,con_neu_i,xxx,yyy,zzz,   &
                                 con_spx,omega2_neu_i,enne_i)

        do ineu=1,nneu_i
          ie=con_neu_i(1,ineu); fa=con_neu_i(2,ineu)

          a1=1; a2=nn; b1=1; b2=nn; c1=1; c2=nn
          if (fa==1) then
            a2=1
          else if (fa==2) then
            a1=nn
          else if (fa==3) then
            b2=1
          else if (fa==4) then
            b1=nn
          else if (fa==5) then
            c2=1
          else if (fa==6) then
            c1=nn
          endif

          jjj=0 
          do k1=c1,c2
            do j1=b1,b2
              do i1=a1,a2
                jjj=jjj+1
                i=con_spx((k1-1)*nn2+(j1-1)*nn+i1,ie)
                iii=point_no_Dir(i)
                if (iii>0) effe_Neu(iii)=effe_Neu(iii)+psi_i(xxx(i),yyy(i),    &
                                         zzz(i))*omega2_neu_i(jjj,ineu)
              enddo
            enddo
          enddo

        enddo        ! loop over Neumann faces (do ineu=1,nneu)
      endif

      if (nneu_e>0) then
        call make_omega2_enne_BC(nn,nnod,nelem,nneu_e,con_neu_e,xxx,yyy,zzz,   &
                                 con_spx,omega2_neu_e,enne_e)

        do ineu=1,nneu_e
          ie=con_neu_e(1,ineu); fa=con_neu_e(2,ineu)

          a1=1; a2=nn; b1=1; b2=nn; c1=1; c2=nn
          if (fa==1) then
            a2=1
          else if (fa==2) then
            a1=nn
          else if (fa==3) then
            b2=1
          else if (fa==4) then
            b1=nn
          else if (fa==5) then
            c2=1
          else if (fa==6) then
            c1=nn
          endif

          jjj=0 
          do k1=c1,c2
            do j1=b1,b2
              do i1=a1,a2
                jjj=jjj+1
                i=con_spx((k1-1)*nn2+(j1-1)*nn+i1,ie)
                iii=point_no_Dir(i+nnod)
                if (iii>0) effe_Neu(iii)=effe_Neu(iii)+psi_e(xxx(i),yyy(i),    &
                                         zzz(i))*omega2_neu_e(jjj,ineu)
              enddo
            enddo
          enddo

        enddo        ! loop over Neumann faces (do ineu=1,nneu)
      endif

      return

      end subroutine make_effe_Neu


!     --------------------------------------------------------------------------


      subroutine evaluate_index_rec(nnod,nrec,xxx,yyy,zzz,xs,ys,zs,index_rec)

!     evaluation of indices of FE grids closest to receivers

      implicit none

      integer(4) :: nnod,nrec,index_rec(nrec),i,irec
      real(8) :: xxx(nnod),yyy(nnod),zzz(nnod),xs(nrec),ys(nrec),zs(nrec)
      real(8) :: term, dist

      do irec=1,nrec
        dist=1.d90
        do i=1,nnod
          term=(xxx(i)-xs(irec))*(xxx(i)-xs(irec))+(yyy(i)-ys(irec))           &
              *(yyy(i)-ys(irec))+(zzz(i)-zs(irec))*(zzz(i)-zs(irec))
          if (term<dist) then
            dist=term
            index_rec(irec)=i
          endif
        enddo
      enddo

      return

      end subroutine evaluate_index_rec


!     --------------------------------------------------------------------------


      real(8) function I_i_app(xxx,yyy,zzz)

      implicit none
      real(8) :: xxx, yyy, zzz, amp_I_app_i, amp_I_app_e
      real(8) :: ax,bx,ay,by,az,bz
      common / applied_Currents / amp_I_app_i, amp_I_app_e
      common /app_curr_location/ ax,bx,ay,by,az,bz

      I_i_app=0.d0
      if (ax<xxx.and.xxx<bx) then
        if (ay<yyy.and.yyy<by) then
          if (az<zzz.and.zzz<bz) then
            I_i_app=amp_I_app_i
          endif
        endif
      endif

      return

      end function I_i_app


!     --------------------------------------------------------------------------


      real(8) function I_e_app(xxx,yyy,zzz)

      implicit none
      real(8) :: xxx, yyy, zzz, amp_I_app_i, amp_I_app_e
      real(8) :: ax,bx,ay,by,az,bz
      common / applied_Currents / amp_I_app_i, amp_I_app_e
      common /app_curr_location/ ax,bx,ay,by,az,bz

      I_e_app=0.d0
      if (ax<xxx.and.xxx<bx) then
        if (ay<yyy.and.yyy<by) then
          if (az<zzz.and.zzz<bz) then
            I_e_app=amp_I_app_e
          endif
        endif
      endif

      end function I_e_app


!     --------------------------------------------------------------------------


      subroutine make_histogram_CSR(nrig,nnz,row_ptr,fichier)

!     histogram of # non-zero elements per matrix row is saved in file fichier

      implicit none

      integer(4) :: i, bw, bw_max
      integer(4) :: nrig, nnz, row_ptr(0:nrig)
      integer(4), allocatable :: histo(:)
      character :: fichier*512

      bw_max=0
      do i=1,nrig
        bw=row_ptr(i)-row_ptr(i-1)
        if (bw>bw_max) bw_max=bw
      enddo
      allocate(histo(0:bw_max)); histo=0
      do i=1,nrig
        bw=row_ptr(i)-row_ptr(i-1)
        histo(bw)=histo(bw)+1
      enddo

      open(53,file=fichier)
      do i=0,bw_max
        write(53,*) i, histo(i)
      enddo
      close(53)

      return

      end subroutine make_histogram_CSR


!     --------------------------------------------------------------------------


      subroutine time_loop_PETSc(myid, nprocs, master,                         &
                                 nn,nnod,nelem,nrig,nnz_tot,nrec,xxx,yyy,zzz,  &
                                 con_spx,index_rec,t0,tfin,ttt,deltat_ODE,     &
                                 deltat_PDE,omega,effe_Dir,point_no_Dir,       &
                                 effe_Neu,row_ptr,col_idx,val,uuu,             &
                                 Dir_not_empty,Neu_not_empty,write_it_hist,    &
                                 write_vtk,write_xdmf,it_snap,nsnap,           &
                                 niter_ave,cg_time1,cg_time2)

      implicit none

      include "Luo_Rudy_I_1991_f90.inc"

#include "finclude/petscsys.h"
#include "finclude/petscvec.h"
#include "finclude/petscmat.h"
#include "finclude/petscksp.h"
#include "finclude/petscpc.h"

      integer(4) :: nn, nn2, nn3, nnod, nelem, nnz_tot, nts_ODE, iter_max, nrec
      integer(4) :: i,j,k,l,m,ie,iii,jjj,ic,its,nts_frac(3),fa,niter,nrig
      integer(4) :: con_spx(nn**3,nelem), index_rec(nrec), its_PETSc=-1
      integer(4) :: deltat_PDE_divided_ODE, nts_PDE, ODE_step, isnap, llll
      integer(4) :: id, tag, nnn, status(MPI_STATUS_SIZE), nnod_PETSc
      integer(4) :: i0, i1, i2, i3, i4, j1, k1, irec, maxits, nrig_loc_max
      integer(4) :: point_no_Dir(2*nnod), row_ptr(0:nrig), col_idx(0:nnz_tot-1)
      integer(4) :: nsnap, it_snap(0:nsnap-1), dim_null_space
      integer(4), allocatable :: ix(:), istarts(:), node_local_PETSc(:)
      integer(4), allocatable :: row_ptr_loc(:), col_idx_loc(:)
      real(8) :: ttt, ttt0, t0, tfin, deltat_ODE, deltat_PDE, zero=0.d0, vol_om
      real(8) :: tempo1, tempo2, cg_time1, cg_time2, cost, term_i, term_e
      real(8) :: area, norm_const, term
      real(8) :: xxx(nnod), yyy(nnod), zzz(nnod), uuu(nrig)
      real(8) :: effe(nrig), c_m_M_delta_t(nnod), effe_Dir(nrig), effe_Neu(nrig)
      real(8) :: val(0:nnz_tot-1), vvv(nnod), omega(nnod)
      real(8) :: serv0, serv1, serv2, serv3, xx, yy, zz, eps=1.d-12,psi_i,psi_e
      real(8), external :: c_m_per_vol, mcap_i, mcap_e, ggg_i, ggg_e, hhh
      real(8) :: amp_I_app_i, amp_I_app_e
      real(8) :: time_ODE_begin, time_ODE_end, tempo_ODE, toll, time_term
      real(8) :: time_PDE_begin, time_PDE_end, tempo_PDE, ttt_PDE, niter_ave
      real(8) :: beg1,end1,temp1,beg2,end2,temp2,term1,term2,beg3,end3
      real(8) :: x1, x2, y1, y2, z1, z2
      real(8) :: rtol,abstol,dtol
      real(8), allocatable :: uuu_loc(:), fff_loc(:), val_loc(:)
      real(8), allocatable :: serv_g(:), uuu_glo(:)
      character :: fichier(nrec)*512, formato*80, TITOLO*512
      character :: riga512*512, solpar*512
      logical :: write_it_hist, write_iteration_file=.true.
      logical :: Dir_not_empty, Neu_not_empty, singular_system
      logical :: write_vtk, write_xdmf, exists
!     LR1 model:
!     y_CellMl are the unknowns of the ODE in the LR1 model
      real(8) :: y_CellMl(8,nnod), y_CellMl_point(8), ion_curr, dy(8)
      real(8), external :: I_i_app, I_e_app
      real(8) :: Chi, C_m, D_i, D_e

      Vec              uuu_PETSc,fff_PETSc, uuu_PETSc_seq, basis_null_space
      VecScatter       ctx
      Mat              A
      KSP              ksp
      PC               pc
      PetscReal        tol
      PetscErrorCode   ierr
      PetscInt istart, iend, istart_mat, iend_mat, nrig_loc, nnz_loc
      PetscMPIInt nprocs, myid, master
      PetscScalar uuu_array(1), alpha
      PetscOffset i_uuu
      PetscViewer iteration_file
      KSPConvergedReason reason
      PetscViewer viewer
      MatNullSpace null_space
      PetscTruth hasconstants

      common / domain_box / x1, x2, y1, y2, z1, z2
      common / applied_Currents / amp_I_app_i, amp_I_app_e
      common / membrane / Chi, C_m, D_i, D_e
      common /Y/y_CellMl_point
      common /DY/dy

      nn2=nn*nn; nn3=nn**3
      if (Dir_not_empty) then
        singular_system=.false.
      else
        singular_system=.true.
      endif

      open(73,file='lin_solv.input')
      k=0
      loop_0: do
        read(73,'(a)') riga512
        loop_1: do j=1,512
          if (riga512(j:j).ne.' ') exit loop_1
        enddo loop_1
        if (riga512(j:j).ne.'!') k=k+1
        if (k==6) exit loop_0
      enddo loop_0
      backspace(73)
      read(73,*) rtol
      read(73,*) abstol
      read(73,*) dtol
      read(73,*) maxits
      close(73)

      term=dabs(dfloat(idint(deltat_PDE/deltat_ODE))-deltat_PDE/deltat_ODE)    &
          /(deltat_PDE/deltat_ODE)
      if (term>eps) then
        stop "Error! deltat_PDE should be an exact multiple of deltat_ODE!"
      else
        deltat_PDE_divided_ODE=idint(deltat_PDE/deltat_ODE)
      endif
      nts_ODE=idnint((tfin-t0)/deltat_ODE); nts_PDE=0

!     Create PETSc vectors:
      call VecCreateMPI(PETSC_COMM_WORLD, PETSC_DECIDE, nrig, uuu_PETSc, ierr)
      call VecDuplicate(uuu_PETSc, fff_PETSc, ierr)
      call VecGetOwnershipRange(uuu_PETSc, istart, iend, ierr)
      
      nrig_loc=iend-istart
      nnz_loc=row_ptr(iend)-row_ptr(istart)

      allocate (row_ptr_loc(0:nrig_loc), col_idx_loc(0:nnz_loc-1))
      allocate (val_loc(0:nnz_loc-1))
      call define_local_matrices(istart,iend,nrig,nnz_tot,row_ptr,col_idx,val, &
                                 nrig_loc,nnz_loc,row_ptr_loc,col_idx_loc,     &
                                 val_loc)

      call eval_nnod_PETSc(nnod,point_no_Dir,istart,iend,nnod_PETSc)
      allocate (node_local_PETSc(nnod_PETSc))
      call eval_node_local_PETSc(nnod,point_no_Dir,istart,iend,nnod_PETSc,     &
                                 node_local_PETSc)

      call MatCreateMPIAIJWithArrays(PETSC_COMM_WORLD,nrig_loc,PETSC_DECIDE,   &
                                     nrig,nrig,row_ptr_loc,col_idx_loc,val_loc,&
                                     A,ierr)

      call MatGetOwnershipRange(A, istart_mat, iend_mat, ierr)

!  Create linear solver context

      call KSPCreate(PETSC_COMM_WORLD,ksp,ierr)

!  Set operators. Here the matrix that defines the linear system
!  also serves as the preconditioning matrix.

      call KSPSetOperators(ksp,A,A,SAME_PRECONDITIONER,ierr)
!     If the system is singular, the null space is created
      if (singular_system) then
        call VecCreateMPI(PETSC_COMM_WORLD, PETSC_DECIDE, nrig,                &
                          basis_null_space, ierr)
!       dim_null_space=0
!       call MatNullSpaceCreate(PETSC_COMM_WORLD, PETSC_TRUE, dim_null_space,  &
!                               PETSC_NULL_DOUBLE_PRECISION, null_space, ierr)
        dim_null_space=1
        alpha=1.d0/dsqrt(dfloat(nrig))
        call VecSet(basis_null_space, alpha, ierr)
        call VecAssemblyBegin(basis_null_space, ierr)
        call VecAssemblyEnd(basis_null_space, ierr)
        call MatNullSpaceCreate(PETSC_COMM_WORLD, PETSC_FALSE, dim_null_space, &
                                basis_null_space, null_space, ierr)
        call KSPSetNullSpace(ksp, null_space, ierr)
      endif

      call KSPSetTolerances(ksp,rtol,abstol,dtol,maxits,ierr)
      call KSPSetFromOptions(ksp,ierr)

!     PETSc vectors:
      allocate (ix(0:nrig_loc-1)); ix=-1
      do i=0,nrig_loc-1
        ix(i)=istart+i
      enddo

      if (myid==master) then
        fichier="rec_000.d"
        do irec=1,nrec
          if (nrec<10) then
            write(fichier(irec)(7:7),'(i1)') irec
          else if (nrec<100) then
            write(fichier(irec)(6:7),'(i2)') irec
          else if (nrec<1000) then
            write(fichier(irec)(5:7),'(i3)') irec
          else 
            stop "Error! Too many receivers!"
          endif
          open(100+irec,file=fichier(irec))
          write(100+irec,201)
          write(100+irec,202) xxx(index_rec(irec)), yyy(index_rec(irec)),      &
                              zzz(index_rec(irec))
        enddo
      endif
201   format(t1,"# time-history of the transmembrane potential evaluated at ")
202   format(t1,"# point (",g9.3,",",g9.3,",",g9.3,")")

      if (myid==master) open(33,file="stimulus.d")

      do i=1,nnod
        c_m_M_delta_t(i)=c_m_per_vol(xxx(i),yyy(i),zzz(i))*omega(i)/deltat_PDE 
      enddo

!     Initial conditions are set s.t. v = y_CellMl_point,
!     i.e. v(t=0) from the LR1 model is set *everywhere*:
      call init
      do i=1,nnod
         vvv(i)=y_CellMl_point(5)
      enddo
      do i=1,nnod
        iii=point_no_Dir(i)
        if (iii>0) uuu(iii)=y_CellMl_point(5)    
        iii=point_no_Dir(i+nnod)
        if (iii>0) uuu(iii)=zero    
      enddo
      do i=1,nnod              
        do j=1,8
          y_CellMl(j,i)=y_CellMl_point(j)
        enddo
      enddo

      allocate (uuu_loc(0:nrig_loc-1), fff_loc(0:nrig_loc-1))
      do i=0,nrig_loc-1
        uuu_loc(i)=uuu(ix(i)+1)
      enddo

      call VecSetValues(uuu_PETSc, nrig_loc, ix, uuu_loc, INSERT_VALUES, ierr)
      call VecAssemblyBegin(uuu_PETSc, ierr)
      call VecAssemblyEnd(uuu_PETSc, ierr)

      allocate (istarts(0:nprocs)); istarts=-1
      call VecGetOwnershipRanges(uuu_PETSc, istarts, ierr)

      tempo_ODE=0.d0; tempo_PDE=0.d0; ODE_step=0; temp1=zero; temp2=zero
      ttt0=t0; ttt_PDE=ttt0; ic=1; niter=0; cg_time1=zero; cg_time2=zero
      isnap=0; llll=0

      if (nts_ODE.ge.4) then
        do i=1,3
          nts_frac(i)=idnint(dfloat(i)*0.25d0*dfloat(nts_ODE))
        enddo
      endif

      allocate (uuu_glo(0:nrig-1))
      if (myid==master) then
        nrig_loc_max=0
        do id=0,nprocs-1
          if (id.ne.myid) then
            if (nrig_loc_max<(istarts(id+1)-istarts(id)))                    &
              nrig_loc_max=istarts(id+1)-istarts(id)
          endif
        enddo
        allocate (serv_g(0:nrig_loc_max-1))
      endif

      if (write_iteration_file) then
        call PetscViewerASCIIOpen(PETSC_COMM_WORLD,"iterations_info.output",   &
                                  iteration_file, ierr)
        call KSPView(ksp,iteration_file,ierr)
        call PetscViewerDestroy(iteration_file,ierr)
        if (myid==master) open(23,file='iterations_info.output',               &
                               position='append')
      endif

      do its=1,nts_ODE

        ttt=ttt0+deltat_ODE
        ODE_step=ODE_step+1

!       the membrane variables are updated numerically:
        call CPU_TIME(time_ODE_begin)
        do i=1,nnod
          if (point_no_Dir(i)>0.or.point_no_Dir(i+nnod)>0) then
            y_CellMl_point(:)=y_CellMl(:,i)
!           before the call to compute_imp, y_CellMl_point(5) is the 
!           transmembrane potential [mV]
            y_CellMl_point(5)=vvv(i)
            call compute_I(ttt,deltat_ODE)
            y_CellMl(:,i)=y_CellMl_point(:)
          endif
        enddo
        call CPU_TIME(time_ODE_end)
        tempo_ODE=tempo_ODE+time_ODE_end-time_ODE_begin

!       PDE's solution is updated, if needed
        if (ODE_step==deltat_PDE_divided_ODE) then
          call CPU_TIME(time_PDE_begin)
          ttt_PDE=ttt
          nts_PDE=nts_PDE+1
          ODE_step=0
          time_term=hhh(ttt_PDE)
          if (myid==master) write(33,*) ttt_PDE, time_term

!         the RHS is built
          call CPU_TIME(beg1)
          effe=0.d0

!         term due to ionic current (membrane model)
          do i=1,nnod
            iii=point_no_Dir(i)
            jjj=point_no_Dir(i+nnod)
            if (iii>0.or.jjj>0) ion_curr=Chi*y_CellMl(5,i)*omega(i)
!           after the call to compute_I, y_CellMl_point(5) is the 
!           ionic current i_ion. We need I_ion = Chi*i_ion [microA/cm3]
            if (iii>0) effe(iii)=effe(iii)-ion_curr
            if (jjj>0) effe(jjj)=effe(jjj)+ion_curr
          enddo    

!         term due to time-marching
          do i=1,nnod
            iii=point_no_Dir(i)
            jjj=point_no_Dir(i+nnod)
            if (iii>0) effe(iii)=effe(iii)+c_m_M_delta_t(i)*vvv(i)
            if (jjj>0) effe(jjj)=effe(jjj)-c_m_M_delta_t(i)*vvv(i)
          enddo

!         term due to applied currents
          do i=1,nnod
            iii=point_no_Dir(i)
            jjj=point_no_Dir(i+nnod)
            if (iii>0) then
              term1=I_i_app(xxx(i),yyy(i),zzz(i))*omega(i)*time_term
              effe(iii)=effe(iii)+term1
            endif
            if (jjj>0) then
              term2=I_e_app(xxx(i),yyy(i),zzz(i))*omega(i)*time_term 
              effe(jjj)=effe(jjj)+term2
            endif
          enddo
780       format(t1,2(g12.6,3x),g12.6)

!         term due to Dirichlet boundary condition
          if (Dir_not_empty) then
            where (effe_Dir.ne.zero) effe=effe+effe_Dir*time_term
          endif

!         term due to Neumann boundary condition
          if (Neu_not_empty) then
            where (effe_Neu.ne.zero) effe=effe+effe_Neu*time_term
          endif

          do i=0,nrig_loc-1
            fff_loc(i)=effe(ix(i)+1)
          enddo
          call VecSetValues(fff_PETSc, nrig_loc, ix, fff_loc, INSERT_VALUES, &
                            ierr)
          call VecAssemblyBegin(fff_PETSc, ierr)
          call VecAssemblyEnd(fff_PETSc, ierr)
          call CPU_TIME(end1)
          temp1=temp1+end1-beg1

          call CPU_TIME(beg2)
          call KSPSetInitialGuessNonzero(ksp,PETSC_TRUE,ierr)
!         call KSPSetInitialGuessNonzero(ksp,PETSC_FALSE,ierr)

!***********************************************************************
          call KSPSolve(ksp,fff_PETSc,uuu_PETSc,ierr)
!***********************************************************************

          call CPU_TIME(end2)

          call KSPGetIterationNumber(ksp, its_PETSc, ierr)
          niter=niter+its_PETSc
          call  KSPGetConvergedReason(ksp, reason, ierr);
          call write_convergence_reason(myid,master,ttt_PDE,its_PETSc,reason)

          uuu_loc=0.d0
          call VecAssemblyBegin(uuu_PETSc, ierr)
          call VecAssemblyEnd(uuu_PETSc, ierr)
          call VecGetValues(uuu_PETSc, nrig_loc, ix, uuu_loc, ierr)

          tag=111
          uuu_glo=zero
          if (myid==master) then
            serv_g=zero
            do id=0,nprocs-1
              if (id.ne.master) then
                 nnn=istarts(id+1)-istarts(id)
                 call MPI_RECV(serv_g(0:nnn-1), nnn, MPI_DOUBLE_PRECISION,     &
                               id, tag, PETSC_COMM_WORLD, status, ierr)
                 uuu_glo(istarts(id):istarts(id+1)-1)=serv_g(0:nnn-1)
              endif
            enddo
            uuu_glo(istart:iend-1)=uuu_loc(0:nrig_loc-1)
          else
            call MPI_SEND(uuu_loc, nrig_loc, MPI_DOUBLE_PRECISION, master,   &
                          tag, PETSC_COMM_WORLD, ierr)
          endif

          call MPI_Bcast (uuu_glo, nrig, MPI_DOUBLE_PRECISION, master,         &
                          PETSC_COMM_WORLD, ierr)

          call CPU_TIME(end2)
          temp2=temp2+end2-beg2

!         extension to vvv(nnod)
          do i=1,nnod
            iii=point_no_Dir(i)
            if (iii>0) then
              vvv(i)=uuu_glo(iii-1)
            else
              vvv(i)=ggg_i(xxx(i),yyy(i),zzz(i))*hhh(ttt)
            endif
            jjj=point_no_Dir(i+nnod)
            if (jjj>0) then
              vvv(i)=vvv(i)-uuu_glo(jjj-1)
            else
              vvv(i)=vvv(i)-ggg_e(xxx(i),yyy(i),zzz(i))*hhh(ttt)
            endif
          enddo

          if (myid==master) then
            if (nrec>0) then
              do irec=1,nrec
                write(100+irec,402) ttt_PDE, vvv(index_rec(irec))
              enddo
            endif
          endif
402       format(t1,g12.6,2x,g12.6)

          if ((myid==master).and.(write_vtk.or.write_xdmf)) then
            if (nsnap>0) then
!             draw the computed solution in vtk or xdmf format 
              if (nts_PDE==it_snap(isnap)) then
                isnap=isnap+1
                llll=llll+1
                if (write_vtk) then
                 solpar='V_xx.vtk'
                 if (llll<10) then
                   write(solpar(3:3),'(a1)')'0'
                   write(solpar(4:4),'(i1)')llll
                 else if (llll<99) then
                   write(solpar(3:4),'(i2)')llll
                 endif
                 write(titolo,*) "transmembrane potential at t = ",ttt_PDE,' ms'
                 call clean_multiple_blank_string(512,titolo)
                 inquire (file=solpar(1:len_trim(solpar)), exist=exists)
                 if (exists) call system ('rm '//solpar(1:len_trim(solpar)))
                 call draw_vtk(solpar,titolo,nn,nnod,nelem,xxx,yyy,zzz,con_spx,&
                               vvv)
                 exists=.false.
                 inquire (file=solpar(1:len_trim(solpar))//'.gz', exist=exists)
                 if (exists)                                                   &
                   call system ('rm '//solpar(1:len_trim(solpar))//'.gz')
                 call system('gzip '//solpar(1:len_trim(solpar)))
                else if (write_xdmf) then
!                tsnap=dfloat(nts_PDE)*deltat_PDE
!                call XDMFWRITE("V_t=", llll, tsnap, nnod, nelem, XYZ, con, vvv)
                endif
              endif
            endif
          endif

        endif
!       PDE's solution has been updated

        call CPU_TIME(time_PDE_end)
        tempo_PDE=tempo_PDE+time_PDE_end-time_PDE_begin

!       time upgrade
        ttt0=ttt

        if (myid==master) then
          if (ic<4.and.its==nts_frac(ic)) then
            write(*,'(i2,"% - ",$)') 25*ic
            ic=ic+1
          endif
        endif

      enddo            ! end of the time-loop

      if (myid==master) then
        close(33)
        do irec=1,nrec
          close(100+irec)
        enddo
      endif

      niter_ave=dfloat(niter)/dfloat(nts_PDE)

231   format(t1,'time-step: ',i8,5x,'t = ',g9.3,' s')
      if (myid==master) close(23)
831   format(t1,i10,2x,4(2x,g12.6))
      if (myid==master) then
        write(*,'("100%            done")') 
!       write(*,*)
      endif

!     if (tempo_PDE.le.999999.d0) then
!       write(*,52) idint(tempo_PDE)
!     else
!       write(*,*) "execution-time for solving the PDEs: ",idint(tempo_PDE)," s"
!     endif
!     write(*,*) 'execution times (s) for building RHS and solving Ax=b: ',    &
!                temp1,temp2
52    format (t1,"CPU-time for solving the PDEs: ",i6," s")

      return

      end subroutine time_loop_PETSc


!     --------------------------------------------------------------------------


      subroutine write_convergence_reason(myid,master,ttt,nts,reason)

      implicit none

      integer(4) :: myid,master,reason,nts
      real(8) :: ttt
      character*512 :: riga512=' '

      if (myid==master) then

      if (reason>0) then
        if (reason==2) then
           riga512='KSP_CONVERGED_RTOL'
        else if (reason==3) then      
           riga512='KSP_CONVERGED_ATOL'
        else if (reason==4) then      
           riga512='KSP_CONVERGED_ITS'
        else if (reason==5) then      
           riga512='KSP_CONVERGED_CG_NEG_CURVE'
        else if (reason==6) then      
           riga512='KSP_CONVERGED_CG_CONSTRAINED'
        else if (reason==7) then      
           riga512='KSP_CONVERGED_STEP_LENGTH'
        else if (reason==8) then      
           riga512='KSP_CONVERGED_HAPPY_BREAKDOWN'
        endif
        write(23,*)
        write(23,172) ttt, nts, riga512(1:len_trim(riga512))
      else if (reason==0) then
        riga512='KSP_CONVERGED_ITERATING'
        write(23,172) ttt, nts, riga512(1:len_trim(riga512))
      else if (reason<0) then
        if (reason==-2) then      
           riga512='KSP_DIVERGED_NULL'
        else if (reason==-3) then      
           riga512='KSP_DIVERGED_ITS'
        else if (reason==-4) then      
           riga512='KSP_DIVERGED_DTOL'
        else if (reason==-5) then      
           riga512='KSP_DIVERGED_BREAKDOWN'
        else if (reason==-6) then      
           riga512='KSP_DIVERGED_BREAKDOWN_BICG'
        else if (reason==-7) then      
           riga512='KSP_DIVERGED_NONSYMMETRIC'
        else if (reason==-8) then      
           riga512='KSP_DIVERGED_INDEFINITE_PC'
        else if (reason==-9) then      
           riga512='KSP_DIVERGED_NAN'
        else if (reason==-10) then
          riga512='KSP_DIVERGED_INDEFINITE_MAT'
        endif
        write(23,172) ttt, nts, riga512(1:len_trim(riga512))
        write(23,*) 'ERROR! At time ',ttt,' s KSP did not converge!'
        write(*,*) 'ERROR! At time ',ttt,' s KSP did not converge!'
        write(*,*) '(',riga512(1:len_trim(riga512)),')'
        stop
      endif

      endif

172   format(t1,g12.6,' s:    # iter = ',i6,' ,   ',a)

      close(171)

      return 

      end subroutine write_convergence_reason


!     --------------------------------------------------------------------------


      subroutine curr_amp_location(riga)

      implicit none

      integer(4) :: i, ia, ib, ja, jb, nle, ngr
      real(8) :: ax,bx,ay,by,az,bz,infy,val_lef,val_rig
      character :: riga*512, axys*1

      common /app_curr_location/ ax,bx,ay,by,az,bz

      infy=1.d30; axys=' '
      ax=-infy;bx=+infy;ay=-infy;by=+infy;az=-infy;bz=+infy

      ib=index(riga,'!')
      if (ib>0) riga(ib:512)=' '

      if (len_trim(riga)<3) then
        write(*,*) "ERROR! Inconsistent specification of BC region!"
        stop
      endif

      ib=1
      do while (ib<len_trim(riga))
        if (ib==1) then
          loop_find_begin: do i=ib,len_trim(riga)
            if (riga(i:i).ne.' ') exit loop_find_begin
          enddo loop_find_begin
          ia=i 
        else
          ia=ib+2
        endif
        ib=ia+index(riga(ia:),',')-1
        if (index(riga(ia:),',')==0) then
          ib=len_trim(riga)
        else
          ib=ib-1
        endif
  
        if (index(riga(ia:ib),'x').ne.0.or.index(riga(ia:ib),'X').ne.0) then
          if (axys.ne.'x') then
            axys='x'
          else
            stop "ERROR! Multiple specification (on x) of BC region!"
          endif
        else if (index(riga(ia:ib),'y').ne.0.or.                               &
                 index(riga(ia:ib),'Y').ne.0) then
          if (axys.ne.'y') then
            axys='y'
          else
            stop "ERROR! Multiple specification (on y) of BC region!"
          endif
        else if (index(riga(ia:ib),'z').ne.0.or.                               &
                 index(riga(ia:ib),'Z').ne.0) then
          if (axys.ne.'z') then
            axys='z'
          else
            stop "ERROR! Multiple specification (on z) of BC region!"
          endif
        endif
  
        nle=0; ngr=0
        do i=ia,ib
          if (riga(i:i)=='<') nle=nle+1
          if (riga(i:i)=='>') ngr=ngr+1
        enddo
        if (nle.ne.0.and.nle.ne.1.and.nle.ne.2.) then
          write(*,*) "Error! Inconsistent condition on ",axys," in location of applied currents!"
          stop
        endif
        if (ngr.ne.0.and.ngr.ne.1) then
          write(*,*) "Error! Inconsistent condition on ",axys," in location of applied currents!"
          stop
        endif
  
        if (nle==1) then
          ja=ia+index(riga(ia:ib),'<')-1
          read(riga(ja+1:ib),*) val_rig
          val_lef=-infy
        endif
        if (nle==2) then
          ja=ia+index(riga(ia:ib),'<')-1
          jb=ja+index(riga(ja+1:ib),'<')
          read(riga(ia:ja-1),*) val_lef
          read(riga(jb+1:ib),*) val_rig
        endif
        if (ngr==1) then
          ja=ia+index(riga(ia:ib),'>')-1
          read(riga(ja+1:ib),*) val_lef
          val_rig=+infy
        endif
  
        check_axys: select case (axys)
          case ('x')
            ax=val_lef
            bx=val_rig
          case ('y')
            ay=val_lef
            by=val_rig
          case ('z')
            az=val_lef
            bz=val_rig
        end select check_axys

      enddo

      return
      end


!     --------------------------------------------------------------------------


      subroutine eval_nnod_PETSc(nnod,point_no_Dir,istart,iend,nnod_PETSc)

      implicit none

      integer(4) :: nnod,point_no_Dir(2*nnod),istart,iend,nnod_PETSc,i
      logical :: belongs_to

      nnod_PETSc=0
      do i=1,nnod
        belongs_to=.false.
        if (point_no_Dir(i).ge.istart.and.point_no_Dir(i)<iend) then
          belongs_to=.true.
          goto 101
        endif
        if (point_no_Dir(i+nnod).ge.istart.and.point_no_Dir(i+nnod)<iend)      &
          belongs_to=.true.
101     continue
        if (belongs_to) nnod_PETSc=nnod_PETSc+1
      enddo

      return

      end subroutine eval_nnod_PETSc


!     --------------------------------------------------------------------------


      subroutine eval_node_local_PETSc(nnod,point_no_Dir,istart,iend,          &
                                       nnod_PETSc,node_local_PETSc)

      implicit none

      integer(4) :: nnod,point_no_Dir(2*nnod),istart,iend,nnod_PETSc,i,ic
      integer(4) :: node_local_PETSc(nnod_PETSc)
      logical :: belongs_to

      node_local_PETSc=-1
      ic=0
      do i=1,nnod
        belongs_to=.false.
        if (point_no_Dir(i).ge.istart.and.point_no_Dir(i)<iend) then
          belongs_to=.true.
          goto 101
        endif
        if (point_no_Dir(i+nnod).ge.istart.and.point_no_Dir(i+nnod)<iend)      &
          belongs_to=.true.
101     continue
        if (belongs_to) then
          ic=ic+1
          node_local_PETSc(ic)=i
        endif
      enddo

      return

      end subroutine eval_node_local_PETSc


!     --------------------------------------------------------------------------


      subroutine define_local_matrices(istart,iend,nrig,nnz_tot,row_ptr,       &
                                       col_idx,val,nrig_loc,nnz_loc,           &
                                       row_ptr_loc,col_idx_loc,val_loc)

      implicit none

      integer(4) :: istart,iend,nrig,nnz_tot,nrig_loc,nnz_loc,i,j
      integer(4) :: row_ptr(0:nrig),col_idx(0:nnz_tot-1)
      integer(4) :: row_ptr_loc(0:nrig_loc),col_idx_loc(0:nnz_loc-1)
      real(8) :: val(0:nnz_tot-1),val_loc(0:nnz_loc-1)

      row_ptr_loc=-1; col_idx_loc=-1
      row_ptr_loc(0)=0; i=0
      do j=istart,iend-1
        i=i+1
        row_ptr_loc(i)=row_ptr_loc(i-1)+row_ptr(j+1)-row_ptr(j)
      enddo

      val_loc=0.d0; i=0
      do j=row_ptr(istart),row_ptr(iend)-1
        col_idx_loc(i)=col_idx(j)
        val_loc(i)=val(j)
        i=i+1
      enddo

      end subroutine define_local_matrices


!     --------------------------------------------------------------------------


      subroutine eval_grid_size(nnod_macro,nelem,xxx_macro,yyy_macro,zzz_macro,&
                                con,dx_ave,dy_ave,dz_ave)

      implicit none

      integer(4) :: nnod_macro,nelem,con(8,nelem),ie
      real(8) :: dx_ave,dy_ave,dz_ave
      real(8) :: x1,x2,x3,x4,x5,x6,x7,x8
      real(8) :: y1,y2,y3,y4,y5,y6,y7,y8
      real(8) :: z1,z2,z3,z4,z5,z6,z7,z8
      real(8), dimension(nnod_macro) :: xxx_macro,yyy_macro,zzz_macro

      dx_ave=0.d0; dy_ave=0.d0; dz_ave=0.d0
      do ie=1,nelem
        x1=xxx_macro(con(1,ie)); x2=xxx_macro(con(2,ie))
        x3=xxx_macro(con(3,ie)); x4=xxx_macro(con(4,ie))
        x5=xxx_macro(con(5,ie)); x6=xxx_macro(con(6,ie))
        x7=xxx_macro(con(7,ie)); x8=xxx_macro(con(8,ie))
        dx_ave=dx_ave+dabs(x1-x2)+dabs(x1-x4)+dabs(x3-x2)+dabs(x3-x4)
        dx_ave=dx_ave+dabs(x5-x6)+dabs(x5-x8)+dabs(x7-x6)+dabs(x7-x8)
        dx_ave=dx_ave+dabs(x1-x5)+dabs(x2-x6)+dabs(x3-x7)+dabs(x4-x8)

        y1=yyy_macro(con(1,ie)); y2=yyy_macro(con(2,ie))
        y3=yyy_macro(con(3,ie)); y4=yyy_macro(con(4,ie))
        y5=yyy_macro(con(5,ie)); y6=yyy_macro(con(6,ie))
        y7=yyy_macro(con(7,ie)); y8=yyy_macro(con(8,ie))
        dy_ave=dy_ave+dabs(y1-y2)+dabs(y1-y4)+dabs(y3-y2)+dabs(y3-y4)
        dy_ave=dy_ave+dabs(y5-y6)+dabs(y5-y8)+dabs(y7-y6)+dabs(y7-y8)
        dy_ave=dy_ave+dabs(y1-y5)+dabs(y2-y6)+dabs(y3-y7)+dabs(y4-y8)

        z1=zzz_macro(con(1,ie)); z2=zzz_macro(con(2,ie))
        z3=zzz_macro(con(3,ie)); z4=zzz_macro(con(4,ie))
        z5=zzz_macro(con(5,ie)); z6=zzz_macro(con(6,ie))
        z7=zzz_macro(con(7,ie)); z8=zzz_macro(con(8,ie))
        dz_ave=dz_ave+dabs(z1-z2)+dabs(z1-z4)+dabs(z3-z2)+dabs(z3-z4)
        dz_ave=dz_ave+dabs(z5-z6)+dabs(z5-z8)+dabs(z7-z6)+dabs(z7-z8)
        dz_ave=dz_ave+dabs(z1-z5)+dabs(z2-z6)+dabs(z3-z7)+dabs(z4-z8)
      enddo

      dx_ave=dx_ave/(12.d0*dfloat(nelem))
      dy_ave=dy_ave/(12.d0*dfloat(nelem))
      dz_ave=dz_ave/(12.d0*dfloat(nelem))

      return

      end subroutine eval_grid_size


!     --------------------------------------------------------------------------


      subroutine read_code_BC(riga512,code_BC)

      implicit none

      integer(4) :: a, b
      character :: riga512*512, code_BC*2

      code_BC=' '
      loop_cerca_inizio: do a=1,512
        if (riga512(a:a).ne.' ') exit  loop_cerca_inizio
      enddo loop_cerca_inizio
      loop_cerca_fine: do b=a+1,512
        if (riga512(b:b)==' '.or.riga512(b:b)==char(9)) exit  loop_cerca_fine
      enddo loop_cerca_fine
      b=b-1
      if (b-a+1.ne.2) stop "ERROR! Wrong Boundary condition in model file!"
      code_BC(1:b-a+1)=riga512(a:b)

      return

      end subroutine read_code_BC


!     --------------------------------------------------------------------------


      subroutine count_boundary_faces(nnod_macro,nelem,con,xxx,yyy,zzz,code_BC,&
                                      n_BC)

!     n_BC is evaluated
   
      implicit none

      integer(4) :: nnod_macro, nelem, con(8,nelem), n_BC, bw_max
      integer(4) :: ie, i, j, iii, i1, i2, i3, i4, i5, i6, i7, i8
      integer(4), allocatable :: con_dual(:,:)
      real(8), dimension(nnod_macro) :: xxx, yyy,zzz
      real(8) :: ax,bx,ay,by,az,bz,cmin,cmax
      character :: code_BC*2
      logical :: boundary_face

      common /app_curr_location/ ax,bx,ay,by,az,bz

!     dual connectivity is built
      bw_max=30
      allocate (con_dual(bw_max,nnod_macro)); con_dual=0
      do ie=1,nelem
        do i=1,8
          iii=con(i,ie)
          if (con_dual(bw_max,iii).ne.0)    &
              stop "Error in count_boundary_faces! (bw_max too small)"
          loop_001: do j=1,bw_max
            if (con_dual(j,iii)==0) exit loop_001
          enddo loop_001
          con_dual(j,iii)=ie
        enddo
      enddo

      n_BC=0
      do ie=1,nelem
        i1=con(1,ie); i2=con(2,ie); i3=con(3,ie); i4=con(4,ie)
        i5=con(5,ie); i6=con(6,ie); i7=con(7,ie); i8=con(8,ie)
        cmin=dmin1(xxx(i1),xxx(i2),xxx(i3),xxx(i4),xxx(i5),xxx(i6),xxx(i7),    &
                   xxx(i8))
        if (cmin>bx) goto 101
        cmax=dmax1(xxx(i1),xxx(i2),xxx(i3),xxx(i4),xxx(i5),xxx(i6),xxx(i7),    &
                   xxx(i8))
        if (cmax<ax) goto 101

        cmin=dmin1(yyy(i1),yyy(i2),yyy(i3),yyy(i4),yyy(i5),yyy(i6),yyy(i7),    &
                   yyy(i8))
        if (cmin>by) goto 101
        cmax=dmax1(yyy(i1),yyy(i2),yyy(i3),yyy(i4),yyy(i5),yyy(i6),yyy(i7),    &
                   yyy(i8))
        if (cmax<ay) goto 101

        cmin=dmin1(zzz(i1),zzz(i2),zzz(i3),zzz(i4),zzz(i5),zzz(i6),zzz(i7),    &
                   zzz(i8))
        if (cmin>bz) goto 101
        cmax=dmax1(zzz(i1),zzz(i2),zzz(i3),zzz(i4),zzz(i5),zzz(i6),zzz(i7),    &
                   zzz(i8))
        if (cmax<az) goto 101

!       face 1
        call check_bound_face(i1,i4,i8,i5,nnod_macro,bw_max,con_dual,          &
                              boundary_face)
        if (boundary_face) n_BC=n_BC+1
!       face 2
        call check_bound_face(i2,i3,i7,i6,nnod_macro,bw_max,con_dual,          &
                              boundary_face)
        if (boundary_face) n_BC=n_BC+1
!       face 3
        call check_bound_face(i1,i2,i6,i5,nnod_macro,bw_max,con_dual,          &
                              boundary_face)
        if (boundary_face) n_BC=n_BC+1
!       face 4
        call check_bound_face(i3,i4,i8,i7,nnod_macro,bw_max,con_dual,          &
                              boundary_face)
        if (boundary_face) n_BC=n_BC+1
!       face 5
        call check_bound_face(i1,i2,i3,i4,nnod_macro,bw_max,con_dual,          &
                              boundary_face)
        if (boundary_face) n_BC=n_BC+1
!       face 6
        call check_bound_face(i5,i6,i7,i8,nnod_macro,bw_max,con_dual,          &
                              boundary_face)
        if (boundary_face) n_BC=n_BC+1
101     continue
      enddo

      return 

      end subroutine count_boundary_faces


!     --------------------------------------------------------------------------


      subroutine check_bound_face(i1,i2,i3,i4,nnod_macro,bw_max,con_dual,      &
                                  boundary_face)

      implicit none

      integer(4) :: i1,i2,i3,i4,nnod_macro,bw_max,a,i,rep,how_many_elements
      integer(4) :: con_dual(bw_max,nnod_macro),serv(4*bw_max)
      logical :: boundary_face

      boundary_face=.true.
      serv=0
      serv(1:bw_max)=con_dual(:,i1)
      serv(bw_max+1:2*bw_max)=con_dual(:,i2)
      serv(2*bw_max+1:3*bw_max)=con_dual(:,i3)
      serv(3*bw_max+1:4*bw_max)=con_dual(:,i4)

      call mio_shell(4*bw_max,serv)

      loop_find_start: do a=1,4*bw_max
        if (serv(a).ne.0) exit loop_find_start
      enddo loop_find_start
      rep=1; how_many_elements=0
      loop_count_rep: do i=a+1,4*bw_max
        if (serv(i)==serv(i-1)) then
          rep=rep+1
        else
          rep=1
        endif
        if (rep==4) how_many_elements=how_many_elements+1
        if (how_many_elements==2) then
          boundary_face=.false.
          exit loop_count_rep
        endif
      enddo loop_count_rep

      return

      end subroutine check_bound_face


!     --------------------------------------------------------------------------


      subroutine list_boundary_faces(nnod_macro,nelem,n_BC,con,xxx,yyy,zzz,    &
                                     code_BC,con_BC)

!     con_BC is evaluated
   
      implicit none

      integer(4) :: nnod_macro, nelem, con(8,nelem), con_BC(2,n_BC), n_BC
      integer(4) :: ie, i, j, iii, i1, i2, i3, i4, i5, i6, i7, i8, i_BC, bw_max
      integer(4), allocatable :: con_dual(:,:)
      real(8), dimension(nnod_macro) :: xxx, yyy,zzz
      real(8) :: ax,bx,ay,by,az,bz,cmin,cmax
      character :: code_BC*2
      logical :: boundary_face

      common /app_curr_location/ ax,bx,ay,by,az,bz

!     dual connectivity is built
      bw_max=30
      allocate (con_dual(bw_max,nnod_macro)); con_dual=0
      do ie=1,nelem
        do i=1,8
          iii=con(i,ie)
          if (con_dual(bw_max,iii).ne.0)    &
              stop "Error in count_boundary_faces! (bw_max too small)"
          loop_001: do j=1,bw_max
            if (con_dual(j,iii)==0) exit loop_001
          enddo loop_001
          con_dual(j,iii)=ie
        enddo
      enddo

      i_BC=0
      do ie=1,nelem
        i1=con(1,ie); i2=con(2,ie); i3=con(3,ie); i4=con(4,ie)
        i5=con(5,ie); i6=con(6,ie); i7=con(7,ie); i8=con(8,ie)
        cmin=dmin1(xxx(i1),xxx(i2),xxx(i3),xxx(i4),xxx(i5),xxx(i6),xxx(i7),    &
                   xxx(i8))
        if (cmin>bx) goto 101
        cmax=dmax1(xxx(i1),xxx(i2),xxx(i3),xxx(i4),xxx(i5),xxx(i6),xxx(i7),    &
                   xxx(i8))
        if (cmax<ax) goto 101

        cmin=dmin1(yyy(i1),yyy(i2),yyy(i3),yyy(i4),yyy(i5),yyy(i6),yyy(i7),    &
                   yyy(i8))
        if (cmin>by) goto 101
        cmax=dmax1(yyy(i1),yyy(i2),yyy(i3),yyy(i4),yyy(i5),yyy(i6),yyy(i7),    &
                   yyy(i8))
        if (cmax<ay) goto 101

        cmin=dmin1(zzz(i1),zzz(i2),zzz(i3),zzz(i4),zzz(i5),zzz(i6),zzz(i7),    &
                   zzz(i8))
        if (cmin>bz) goto 101
        cmax=dmax1(zzz(i1),zzz(i2),zzz(i3),zzz(i4),zzz(i5),zzz(i6),zzz(i7),    &
                   zzz(i8))
        if (cmax<az) goto 101

!       face 1
        call check_bound_face(i1,i4,i8,i5,nnod_macro,bw_max,con_dual,          &
                              boundary_face)
        if (boundary_face) then
          i_BC=i_BC+1
          con_BC(1,i_BC)=ie; con_BC(2,i_BC)=1
        endif

!       face 2
        call check_bound_face(i2,i3,i7,i6,nnod_macro,bw_max,con_dual,          &
                              boundary_face)
        if (boundary_face) then
          i_BC=i_BC+1
          con_BC(1,i_BC)=ie; con_BC(2,i_BC)=2
        endif
!       face 3
        call check_bound_face(i1,i2,i6,i5,nnod_macro,bw_max,con_dual,          &
                              boundary_face)
        if (boundary_face) then
          i_BC=i_BC+1
          con_BC(1,i_BC)=ie; con_BC(2,i_BC)=3
        endif
!       face 4
        call check_bound_face(i3,i4,i8,i7,nnod_macro,bw_max,con_dual,          &
                              boundary_face)
        if (boundary_face) then
          i_BC=i_BC+1
          con_BC(1,i_BC)=ie; con_BC(2,i_BC)=4
        endif
!       face 5
        call check_bound_face(i1,i2,i3,i4,nnod_macro,bw_max,con_dual,          &
                              boundary_face)
        if (boundary_face) then
          i_BC=i_BC+1
          con_BC(1,i_BC)=ie; con_BC(2,i_BC)=5
        endif
!       face 6
        call check_bound_face(i5,i6,i7,i8,nnod_macro,bw_max,con_dual,          &
                              boundary_face)
        if (boundary_face) then
          i_BC=i_BC+1
          con_BC(1,i_BC)=ie; con_BC(2,i_BC)=6
        endif

101     continue
      enddo

      return 

      end subroutine list_boundary_faces


!     --------------------------------------------------------------------------


      subroutine find_parent_hex(xr,yr,zr,ns,nnod,nelem,x_SE,y_SE,z_SE,  &
                                 con_spx,hex,x_hat,y_hat,z_hat,dist,inter)

!     given a point (xr,yr,zr), the hexahedron hex containing it is detected.
!     along with point (x_hat,y_hat,z_hat) in [-1,+1]^3 s.t.
!     F(x_hat,y_hat,z_hat)=(xr,yr,zr), where F is the quasi-linear mapping

!     INPUT:
!     ns - spectral degree
!     nnod,nelem - # of SE nodes and elements, respectivley
!     (xr,yr,zr) - coordinates of the given point
!     x_SE,y_SE,z_SE - (x,y,z) coordinates of the SE gridpoints
!     con_spx - SE connectivities

!     OUTPUT:
!     hex = integer, index of the hexahedron containing (xr,yr,zr)
!     (x_hat,y_hat,z_hat) = real triplet in [-1,+1]^3 s.t. 
!                           F(x_hat,y_hat,z_hat)=(xr,yr,zr)
!     dist = distance between (xr,yr,zr) and F(x_hat,y_hat,z_hat)
!     interno = logical, if .true. the hexahedron hex has been found
!                otherwise point (xr,yr,zr) is external to the spectral domain
!                (i.e. the union of all hexahedra)

      implicit none

      integer(4) :: ns,nn,nn2,nn3,nnod,nelem,con_spx((ns+1)**3,nelem),hex
      integer(4) :: i1,i2,i3,i4,i5,i6,i7,i8,ie
      real(8) :: x_SE(nnod),y_SE(nnod),z_SE(nnod)
      real(8) :: xr,yr,zr,x_hat,y_hat,z_hat,x_app,y_app,z_app,dist
      real(8) :: xmin,xmax,ymin,ymax,zmin,zmax
      logical :: inter

      nn=ns+1; nn2=nn*nn; nn3=nn2*nn
      inter=.false.; hex=0

      loop_find_parent_hex: do ie=1,nelem

        i1=con_spx(1,ie); i2=con_spx(nn,ie)
        i3=con_spx(nn2,ie); i4=con_spx(nn2-nn+1,ie)
        i5=con_spx(nn3-nn2+1,ie); i6=con_spx(nn3-nn2+nn,ie)
        i7=con_spx(nn3,ie); i8=con_spx(nn3-nn+1,ie)
        xmin=dmin1(x_SE(i1),x_SE(i2),x_SE(i3),x_SE(i4),x_SE(i5),x_SE(i6),      &
                   x_SE(i7),x_SE(i8)) 
        xmax=dmax1(x_SE(i1),x_SE(i2),x_SE(i3),x_SE(i4),x_SE(i5),x_SE(i6),      &
                   x_SE(i7),x_SE(i8))
        ymin=dmin1(y_SE(i1),y_SE(i2),y_SE(i3),y_SE(i4),y_SE(i5),y_SE(i6),      &
                   y_SE(i7),y_SE(i8))
        ymax=dmax1(y_SE(i1),y_SE(i2),y_SE(i3),y_SE(i4),y_SE(i5),y_SE(i6),      &
                   y_SE(i7),y_SE(i8))
        zmin=dmin1(z_SE(i1),z_SE(i2),z_SE(i3),z_SE(i4),z_SE(i5),z_SE(i6),      &
                   z_SE(i7),z_SE(i8))
        zmax=dmax1(z_SE(i1),z_SE(i2),z_SE(i3),z_SE(i4),z_SE(i5),z_SE(i6),      &
                   z_SE(i7),z_SE(i8))

        if (xr<xmin) goto 101
        if (xr>xmax) goto 101
        if (yr<ymin) goto 101
        if (yr>ymax) goto 101
        if (zr<zmin) goto 101
        if (zr>zmax) goto 101      

        call is_parent_hexa(xr,yr,zr,x_SE(i1),y_SE(i1),z_SE(i1),x_SE(i2),      &
                            y_SE(i2),z_SE(i2),x_SE(i3),y_SE(i3),z_SE(i3),      &
                            x_SE(i4),y_SE(i4),z_SE(i4),x_SE(i5),y_SE(i5),      &
                            z_SE(i5),x_SE(i6),y_SE(i6),z_SE(i6),x_SE(i7),      &
                            y_SE(i7),z_SE(i7),x_SE(i8),y_SE(i8),z_SE(i8),inter)

        if (.not.inter) goto 101
        hex=ie

        call inverse_Q1(ns,xr,yr,zr,x_SE(i1),y_SE(i1),z_SE(i1),x_SE(i2),       &
                        y_SE(i2),z_SE(i2),x_SE(i3),y_SE(i3),z_SE(i3),x_SE(i4), &
                        y_SE(i4),z_SE(i4),x_SE(i5),y_SE(i5),z_SE(i5),x_SE(i6), &
                        y_SE(i6),z_SE(i6),x_SE(i7),y_SE(i7),z_SE(i7),x_SE(i8), &
                        y_SE(i8),z_SE(i8),x_hat,y_hat,z_hat,x_app,y_app,z_app, &
                        dist)

        if (inter) then
          exit loop_find_parent_hex
        else
          write(*,*) "ERROR! inverse_Q1 failed for (xr,yr,zr) = ",xr,yr,zr
          stop
        endif

101     continue

      enddo loop_find_parent_hex

      return

      end subroutine find_parent_hex

      
!     --------------------------------------------------------------------------


      subroutine inverse_Q1(ns,xr,yr,zr,ax,ay,az,bx,by,bz,cx,cy,cz,dx,dy,dz,   &
                            ex,ey,ez,fx,fy,fz,gx,gy,gz,hx,hy,hz,xhat,yhat,zhat,&
                            xapp,yapp,zapp,dist)

!     check is point (xr,yr,zr) is contained in this hexa

      implicit none
      integer(4) :: ns,nx,ny,nz,i,j,k
      real(8) :: xr,yr,zr,deltax,deltay,deltaz,x,y,z,dist,dist_min
      real(8) :: xmin,ymin,zmin,xhat,yhat,zhat,xapp,yapp,zapp
      real(8) :: ax,bx,cx,dx,ex,fx,gx,hx
      real(8) :: ay,by,cy,dy,ey,fy,gy,hy
      real(8) :: az,bz,cz,dz,ez,fz,gz,hz
      real(8) :: alfa1,beta1,gamma1,lambda1,mu1,nu1,sigma1,delta1
      real(8) :: alfa2,beta2,gamma2,lambda2,mu2,nu2,sigma2,delta2
      real(8) :: alfa3,beta3,gamma3,lambda3,mu3,nu3,sigma3,delta3
      real(8), allocatable, dimension(:) :: xx,yy,zz
!     logical :: inter

      alfa1=(-ax+bx+cx-dx-ex+fx+gx-hx)/8.d0
      beta1=(-ax-bx+cx+dx-ex-fx+gx+hx)/8.d0
      gamma1=(-ax-bx-cx-dx+ex+fx+gx+hx)/8.d0
      lambda1=(ax-bx+cx-dx+ex-fx+gx-hx)/8.d0
      mu1=(+ax-bx-cx+dx-ex+fx+gx-hx)/8.d0
      nu1=(+ax+bx-cx-dx-ex-fx+gx+hx)/8.d0
      sigma1=(-ax+bx-cx+dx+ex-fx+gx-hx)/8.d0
      delta1=(ax+bx+cx+dx+ex+fx+gx+hx)/8.d0

      alfa2=(-ay+by+cy-dy-ey+fy+gy-hy)/8.d0
      beta2=(-ay-by+cy+dy-ey-fy+gy+hy)/8.d0
      gamma2=(-ay-by-cy-dy+ey+fy+gy+hy)/8.d0
      lambda2=(ay-by+cy-dy+ey-fy+gy-hy)/8.d0
      mu2=(+ay-by-cy+dy-ey+fy+gy-hy)/8.d0
      nu2=(+ay+by-cy-dy-ey-fy+gy+hy)/8.d0
      sigma2=(-ay+by-cy+dy+ey-fy+gy-hy)/8.d0
      delta2=(ay+by+cy+dy+ey+fy+gy+hy)/8.d0

      alfa3=(-az+bz+cz-dz-ez+fz+gz-hz)/8.d0
      beta3=(-az-bz+cz+dz-ez-fz+gz+hz)/8.d0
      gamma3=(-az-bz-cz-dz+ez+fz+gz+hz)/8.d0
      lambda3=(az-bz+cz-dz+ez-fz+gz-hz)/8.d0
      mu3=(+az-bz-cz+dz-ez+fz+gz-hz)/8.d0
      nu3=(+az+bz-cz-dz-ez-fz+gz+hz)/8.d0
      sigma3=(-az+bz-cz+dz+ez-fz+gz-hz)/8.d0
      delta3=(az+bz+cz+dz+ez+fz+gz+hz)/8.d0

      allocate (xx(0:nx), yy(0:ny), zz(0:nz))
      deltax=2.d0/dfloat(nx); deltay=2.d0/dfloat(ny); deltaz=2.d0/dfloat(nz)
      do i=0,nx
        xx(i)=-1.d0+dfloat(i)*deltax
      enddo
      do j=0,ny
        yy(j)=-1.d0+dfloat(j)*deltay
      enddo
      do k=0,nz
        zz(k)=-1.d0+dfloat(k)*deltaz
      enddo

      dist_min=1.d90
      do i=0,nx
        do j=0,ny
          do k=0,nz
            x=alfa1*xx(i)+beta1*yy(j)+gamma1*zz(k)+lambda1*xx(i)*yy(j)+        &
              mu1*xx(i)*zz(k)+nu1*yy(j)*zz(k)+sigma1*xx(i)*yy(j)*zz(k)+delta1
            y=alfa2*xx(i)+beta2*yy(j)+gamma2*zz(k)+lambda2*xx(i)*yy(j)+        &
              mu2*xx(i)*zz(k)+nu2*yy(j)*zz(k)+sigma2*xx(i)*yy(j)*zz(k)+delta2
            z=alfa3*xx(i)+beta3*yy(j)+gamma3*zz(k)+lambda3*xx(i)*yy(j)+        &
              mu3*xx(i)*zz(k)+nu3*yy(j)*zz(k)+sigma3*xx(i)*yy(j)*zz(k)+delta3
            dist=(xr-x)*(xr-x)+(yr-y)*(yr-y)+(zr-z)*(zr-z)
            if (dist<dist_min) then
              dist_min=dist
              xapp=x
              yapp=y
              zapp=z
              xhat=xx(i)
              yhat=yy(j)
              zhat=zz(k)
            endif
          enddo
        enddo
      enddo

      return

      end subroutine inverse_Q1


!     --------------------------------------------------------------------------


      subroutine is_parent_hexa(xr,yr,zr,ax,ay,az,bx,by,bz,cx,cy,cz,dx,dy,dz,  &
                                ex,ey,ez,fx,fy,fz,gx,gy,gz,hx,hy,hz,inter)

      implicit none

      real(8) :: enne(3,6), term, xr, yr, zr
      real(8) :: ax,bx,cx,dx,ex,fx,gx,hx
      real(8) :: ay,by,cy,dy,ey,fy,gy,hy
      real(8) :: az,bz,cz,dz,ez,fz,gz,hz
      real(8) :: alfa1,beta1,gamma1,delta1,alfa2,beta2,gamma2,delta2,alfa3,beta3
      real(8) :: gamma3,delta3,lambda1,mu1,nu1,sigma1,lambda2,mu2,nu2,sigma2
      real(8) :: lambda3,mu3,nu3,sigma3
      logical :: inter

      inter=.true.

      alfa1=(-ax+bx+cx-dx-ex+fx+gx-hx)/8.d0
      beta1=(-ax-bx+cx+dx-ex-fx+gx+hx)/8.d0
      gamma1=(-ax-bx-cx-dx+ex+fx+gx+hx)/8.d0
      lambda1=(ax-bx+cx-dx+ex-fx+gx-hx)/8.d0
      mu1=(+ax-bx-cx+dx-ex+fx+gx-hx)/8.d0
      nu1=(+ax+bx-cx-dx-ex-fx+gx+hx)/8.d0
      sigma1=(-ax+bx-cx+dx+ex-fx+gx-hx)/8.d0
      delta1=(ax+bx+cx+dx+ex+fx+gx+hx)/8.d0

      alfa2=(-ay+by+cy-dy-ey+fy+gy-hy)/8.d0
      beta2=(-ay-by+cy+dy-ey-fy+gy+hy)/8.d0
      gamma2=(-ay-by-cy-dy+ey+fy+gy+hy)/8.d0
      lambda2=(ay-by+cy-dy+ey-fy+gy-hy)/8.d0
      mu2=(+ay-by-cy+dy-ey+fy+gy-hy)/8.d0
      nu2=(+ay+by-cy-dy-ey-fy+gy+hy)/8.d0
      sigma2=(-ay+by-cy+dy+ey-fy+gy-hy)/8.d0
      delta2=(ay+by+cy+dy+ey+fy+gy+hy)/8.d0

      alfa3=(-az+bz+cz-dz-ez+fz+gz-hz)/8.d0
      beta3=(-az-bz+cz+dz-ez-fz+gz+hz)/8.d0
      gamma3=(-az-bz-cz-dz+ez+fz+gz+hz)/8.d0
      lambda3=(az-bz+cz-dz+ez-fz+gz-hz)/8.d0
      mu3=(+az-bz-cz+dz-ez+fz+gz-hz)/8.d0
      nu3=(+az+bz-cz-dz-ez-fz+gz+hz)/8.d0
      sigma3=(-az+bz-cz+dz+ez-fz+gz-hz)/8.d0
      delta3=(az+bz+cz+dz+ez+fz+gz+hz)/8.d0

!     O = (delta1,delta2,delta3)

!     face 1:
!     C = (-alfa1+delta1,-alfa2+delta2,-alfa3+delta3)
!     C-O = (-alfa1,-alfa2,-alfa3)
!     C-R = (-alfa1+delta1-xr,-alfa2+delta2-yr,-alfa3+delta3-zr)

      enne(1,1)=-(beta2-lambda2)*(gamma3-mu3)+(beta3-lambda3)*(gamma2-mu2)
      enne(2,1)=-(beta3-lambda3)*(gamma1-mu1)+(beta1-lambda1)*(gamma3-mu3)
      enne(3,1)=-(beta1-lambda1)*(gamma2-mu2)+(beta2-lambda2)*(gamma1-mu1)
!     term = (C-O) scalar enne
      term=-alfa1*enne(1,1)-alfa2*enne(2,1)-alfa3*enne(3,1)
      if (term<0.d0) enne(:,1)=-enne(:,1)
!     term = (C-R) scalar enne
      term=(-alfa1+delta1-xr)*enne(1,1)+(-alfa2+delta2-yr)*enne(2,1)+          &
           (-alfa3+delta3-zr)*enne(3,1)
      if (term<0.d0) then
        inter=.false.
        goto 222
      endif

!     face 2:
!     C = (alfa1+delta1,alfa2+delta2,alfa3+delta3)
!     C-O = (alfa1,alfa2,alfa3)
!     C-R = (alfa1+delta1-xr,alfa2+delta2-yr,alfa3+delta3-zr)
      enne(1,2)=(beta2+lambda2)*(gamma3+mu3)-(beta3+lambda3)*(gamma2+mu2)
      enne(2,2)=(beta3+lambda3)*(gamma1+mu1)-(beta1+lambda1)*(gamma3+mu3)
      enne(3,2)=(beta1+lambda1)*(gamma2+mu2)-(beta2+lambda2)*(gamma1+mu1)
!     term = (C-O) scalar enne
      term=alfa1*enne(1,2)+alfa2*enne(2,2)+alfa3*enne(3,2)
      if (term<0.d0) enne(:,2)=-enne(:,2)
!     term = (C-R) scalar enne
      term=(alfa1+delta1-xr)*enne(1,2)+(alfa2+delta2-yr)*enne(2,2)+            &
           (alfa3+delta3-zr)*enne(3,2)
      if (term<0.d0) then
        inter=.false.
        goto 222
      endif

!     face 3:
!     C = (-beta1+delta1,-beta2+delta2,-beta3+delta3)
!     C-O = (-beta1,-beta2,-beta3)
!     C-R = (-beta1+delta1-xr,-beta2+delta2-yr,-beta3+delta3-zr)
      enne(1,3)=(alfa2-lambda2)*(gamma3-nu3)-(alfa3-lambda3)*(gamma2-nu2)
      enne(2,3)=(alfa3-lambda3)*(gamma1-nu1)-(alfa1-lambda1)*(gamma3-nu3)
      enne(3,3)=(alfa1-lambda1)*(gamma2-nu2)-(alfa2-lambda2)*(gamma1-nu1)
!     term = (C-O) scalar enne
      term=-beta1*enne(1,3)-beta2*enne(2,3)-beta3*enne(3,3)
      if (term<0.d0) enne(:,3)=-enne(:,3)
!     term = (C-R) scalar enne
      term=(-beta1+delta1-xr)*enne(1,3)+(-beta2+delta2-yr)*enne(2,3)+          &
           (-beta3+delta3-zr)*enne(3,3)
      if (term<0.d0) then
        inter=.false.
        goto 222
      endif

!     face 4:
!     C = (beta1+delta1,beta2+delta2,beta3+delta3)
!     C-O = (beta1,beta2,beta3)
!     C-R = (beta1+delta1-xr,beta2+delta2-yr,beta3+delta3-zr)
      enne(1,4)=-(alfa2+lambda2)*(gamma3+nu3)+(alfa3+lambda3)*(gamma2+nu2)
      enne(2,4)=-(alfa3+lambda3)*(gamma1+nu1)+(alfa1+lambda1)*(gamma3+nu3)
      enne(3,4)=-(alfa1+lambda1)*(gamma2+nu2)+(alfa2+lambda2)*(gamma1+nu1)
!     term = (C-O) scalar enne
      term=beta1*enne(1,4)+beta2*enne(2,4)+beta3*enne(3,4)
      if (term<0.d0) enne(:,4)=-enne(:,4)
!     term = (C-R) scalar enne
      term=(beta1+delta1-xr)*enne(1,4)+(beta2+delta2-yr)*enne(2,4)+            &
           (beta3+delta3-zr)*enne(3,4)
      if (term<0.d0) then
        inter=.false.
        goto 222
      endif

!     face 5:
!     C = (-gamma1+delta1,-gamma2+delta2,-gamma3+delta3)
!     C-O = (-gamma1,-gamma2,-gamma3)
!     C-R = (-gamma1+delta1-xr,-gamma2+delta2-yr,-gamma3+delta3-zr)
      enne(1,5)=-(alfa2-mu2)*(beta3-nu3)+(alfa3-mu3)*(beta2-nu2)
      enne(2,5)=-(alfa3-mu3)*(beta1-nu1)+(alfa1-mu1)*(beta3-nu3)
      enne(3,5)=-(alfa1-mu1)*(beta2-nu2)+(alfa2-mu2)*(beta1-nu1)
!     term = (C-O) scalar enne
      term=-gamma1*enne(1,5)-gamma2*enne(2,5)-gamma3*enne(3,5)
      if (term<0.d0) enne(:,5)=-enne(:,5)
!     term = (C-R) scalar enne
      term=(-gamma1+delta1-xr)*enne(1,5)+(-gamma2+delta2-yr)*enne(2,5)+        &
           (-gamma3+delta3-zr)*enne(3,5)
      if (term<0.d0) then
        inter=.false.
        goto 222
      endif

!     face 6:
!     C = (gamma1+delta1,gamma2+delta2,gamma3+delta3)
!     C-O = (gamma1,gamma2,gamma3)
!     C-R = (gamma1+delta1-xr,gamma2+delta2-yr,gamma3+delta3-zr)
      enne(1,6)=(alfa2+mu2)*(beta3+nu3)-(alfa3+mu3)*(beta2+nu2)
      enne(2,6)=(alfa3+mu3)*(beta1+nu1)-(alfa1+mu1)*(beta3+nu3)
      enne(3,6)=(alfa1+mu1)*(beta2+nu2)-(alfa2+mu2)*(beta1+nu1)
!     term = (C-O) scalar enne
      term=gamma1*enne(1,6)+gamma2*enne(2,6)+gamma3*enne(3,6)
      if (term<0.d0) enne(:,6)=-enne(:,6)
!     term = (C-R) scalar enne
      term=(gamma1+delta1-xr)*enne(1,6)+(gamma2+delta2-yr)*enne(2,6)+          &
           (gamma3+delta3-zr)*enne(3,6)
      if (term<0.d0) then
        inter=.false.
        goto 222
      endif

222   continue
          
      return
      
      end subroutine is_parent_hexa


!     --------------------------------------------------------------------------
